import logging
import threading

from typing import List, cast, Dict, Any
from typing_extensions import TypeAlias

from eportal import utils
from eportal.utils import BatchProcessor, process_int
from eportal.patchsource import PatchSource
from config import REBOOTS_QUEUE_TIMEOUT, REBOOTS_QUEUE_MAX_BATCH_SIZE, REBOOTS_EXTENDED_STATS

CheckinT: TypeAlias = Dict[str, Any]

logger = logging.getLogger(__name__)


REBOOT_DETECTOR_VERSION = 1
SEND_TIMEOUT = 30.0
INT_FIELDS = [
    'uptime',
    'patch_level',
    'ts',
    'proxy',
    'last_stop',
    'ltimestamp',
    'last',
]


def process_agent_reboots(data: List[CheckinT]) -> None:
    for checkin in data:
        if detect_agent_reboot(checkin):
            agent_reboots_consumer.append(checkin)


def detect_agent_reboot(checkin: CheckinT) -> bool:
    reason = cast(str, checkin.get('reason'))
    uptime = cast(int, checkin.get('uptime'))
    patch_level = cast(int, checkin.get('patch_level'))
    last_stop = cast(int, checkin.get('last_stop'))
    ts = cast(int, checkin.get('ts'))

    try:
        state_ts = int(checkin['state']['ts'])
    except (KeyError, TypeError, ValueError):
        state_ts = 0

    fields = [reason, uptime, patch_level, state_ts, last_stop, ts]

    if not all(fields):
        return False

    first_update_after_reboot_marker = False
    crash_soon_after_update_marker = False
    no_proper_shutdown_marker = False

    if uptime < 300 and patch_level == -1:
        first_update_after_reboot_marker = True
    if (ts - uptime) > state_ts > (ts - uptime - 1800) and reason == 'update':
        crash_soon_after_update_marker = True
    if last_stop < state_ts:
        no_proper_shutdown_marker = True

    markers = [first_update_after_reboot_marker]
    if not REBOOTS_EXTENDED_STATS:
        # send anomaly events only
        markers.extend([crash_soon_after_update_marker, no_proper_shutdown_marker])

    if all(markers):
        logger.info('Agent reboot detected: %s', checkin)
        return True

    return False


CheckinBatchProcessor = BatchProcessor[CheckinT]


@CheckinBatchProcessor(timeout=REBOOTS_QUEUE_TIMEOUT, batch_size=REBOOTS_QUEUE_MAX_BATCH_SIZE)
def agent_reboots_consumer(batch: List[CheckinT]) -> None:
    send_agent_reboots_to_patchserver(batch)


def fields_to_int(checkin: CheckinT) -> CheckinT:
    for field in INT_FIELDS:
        process_int(checkin, field, default=None)

    return checkin


def send_agent_reboots_to_patchserver(batch: List[CheckinT]) -> None:
    batch = [fields_to_int(checkin) for checkin in batch]
    ps = PatchSource()
    url = ps.get_url('/eportal-api/reboots')
    if not ps.is_configured():
        return

    logging.debug('Sending %i reboot events to patchserver', len(batch))

    payload = {
        'reboots': batch,
        'detector_version': REBOOT_DETECTOR_VERSION,
    }
    try:
        with utils.get_session() as s:
            response = s.post(url, json=payload, timeout=SEND_TIMEOUT, **utils.get_req_args(auth=ps.auth))
            response.raise_for_status()
        logger.info('%s reboot events sent to patchserver.', len(batch))
    except Exception as e:
        logger.error('Failed to send reboot events to patchserver Url: %s, error: %s', url, e)


def agent_reboots_queue_processor(stop: threading.Event) -> threading.Thread:
    return agent_reboots_consumer.start(stop, 'Kernel anomaly consumer error')
