import sys
import subprocess
import os.path
import tempfile
import contextlib
import shutil

import pysqlite3 as sqlite3

from typing import Optional, Iterator, Set, List

from eportal import utils
import config


class Restore:
    RSet = Set[str]

    DB = 'db'
    CONFIG = 'config'
    INSTANCE_ID = 'instance_id'

    DATA = DEFAULT = {DB, CONFIG}
    FULL = {DB, CONFIG, INSTANCE_ID}


@contextlib.contextmanager
def temporary_dir(
    suffix: Optional[str] = None, prefix: Optional[str] = None, dir: Optional[str] = None
) -> Iterator[str]:
    tmpdir = tempfile.mkdtemp(suffix=suffix, prefix=prefix, dir=dir)
    try:
        yield tmpdir
    finally:
        if 'EPORTAL_DEBUG' not in os.environ:
            shutil.rmtree(tmpdir)


def bash_cmd() -> List[str]:
    return ['bash', '-x'] if 'EPORTAL_DEBUG' in os.environ else ['bash']


BACKUP = '''\
set -euo pipefail

mkdir -p "$tmpdir"/config
eportal_config=${eportal_config:-/etc/eportal/config}
if [ -f "$eportal_config" ]; then
  mkdir -p "$tmpdir"/config/eportal
  cp "$eportal_config" "$tmpdir"/config/eportal/config
fi

if [ -f "$datadir"/config/config.json ]; then
  cp "$datadir"/config/config.json "$tmpdir"/config/
else
  touch "$tmpdir"/config/config.json
fi

if [ -f "$datadir"/instance-id ]; then
  cp "$datadir"/instance-id "$tmpdir/"
fi

cd "$tmpdir"
find . -type f | xargs sha256sum | grep -v checksums > checksums.sha256

tar -c --auto-compress -f "$dest" -C "$tmpdir" .
'''


def backup(dest: str) -> int:
    dest = os.path.abspath(dest)
    backup_dir = os.path.dirname(dest)
    os.makedirs(backup_dir, exist_ok=True)
    with temporary_dir(prefix='eportal-backup-', dir=backup_dir) as tmpdir:
        tmpdb = os.path.join(tmpdir, 'db')
        os.mkdir(tmpdb)
        backup_databases(None, tmpdb)

        env = os.environ.copy()
        env['tmpdir'] = tmpdir
        env['dest'] = dest
        env['datadir'] = config.BASE_DIR
        return subprocess.run(bash_cmd(), env=env, input=BACKUP.encode()).returncode


RESTORE = '''\
set -euo pipefail

echo Unpack ...
tar --no-same-owner -xf "$archive" -C "$tmpdir"

echo Verify integrity ...
cd "$tmpdir"
sha256sum -c checksums.sha256

cleanupdb() {
  find "$datadir" -name "$(basename "$1")"'*' -maxdepth 1 -delete
}

install_as_owner() {
  local install_owner_opts=$( stat -c '-o %u -g %g' "$datadir" )
  install -v -m 0644 $install_owner_opts -DT "$1" "$2"
}

if [ -n "$restore_config" ]; then
  echo Restore config.json ...
  install_as_owner config/config.json "$datadir/config/config.json"

  eportal_config=${eportal_config:-/etc/eportal/config}
  echo Restore "$eportal_config" ...
  test -f config/eportal/config && cp config/eportal/config "$eportal_config"
fi

if [ -n "$restore_db" ]; then
  echo Restore DB ...
  test -f "$datadb" && install_as_owner "$datadb" "$datadir/olddata/$(basename "$datadb")"
  test -f "$patchesdb" && install_as_owner "$patchesdb" "$datadir/olddata/$(basename "$patchesdb")"
  cleanupdb "$datadb"
  cleanupdb "$patchesdb"
  find ./db -maxdepth 1 -type f | while read db; do
    install_as_owner "$db" "$datadir/$(basename "$db")"
  done
fi

if [ -n "$restore_instance_id" ]; then
  echo Restore instance ID ...
  test -f instance-id && install_as_owner instance-id "$datadir/instance-id"
fi

echo Done.
'''


def restore(archive: str, what: Restore.RSet) -> int:
    if not os.path.isdir(config.BASE_DIR):
        print(f"ePortal base dir ({config.BASE_DIR}) doesn't exist", file=sys.stderr)
        return 6

    if (lock_fd := utils.instance_lock()) is None:
        print('running ePortal instance was detected, it should be stopped during restore', file=sys.stderr)
        return 5

    try:
        with temporary_dir(prefix='eportal-restore-') as tmpdir:
            env = os.environ.copy()
            env['tmpdir'] = tmpdir
            env['archive'] = archive
            env['datadir'] = config.BASE_DIR
            env['datadb'] = get_db_fname_from_uri(config.DATABASE_URI)
            env['patchesdb'] = get_db_fname_from_uri(config.PATCHES_DATA_DATABASE_URI)

            env['restore_config'] = '1' if Restore.CONFIG in what else ''
            env['restore_db'] = '1' if Restore.DB in what else ''
            env['restore_instance_id'] = '1' if Restore.INSTANCE_ID in what else ''
            return subprocess.run(bash_cmd(), env=env, input=RESTORE.encode()).returncode
    finally:
        os.close(lock_fd)


def get_db_fname_from_uri(uri: str) -> str:
    return uri.split('sqlite:///')[-1]


def backup_progress(status, remaining, total) -> None:  # type: ignore[no-untyped-def]
    print(f'Copied {total-remaining} of {total} pages...', file=sys.stderr)


def backup_databases(db_list: Optional[List[str]], backup_dir: str, verbose: bool = False) -> None:
    db_list = db_list or [config.DATABASE_URI, config.PATCHES_DATA_DATABASE_URI]
    os.makedirs(backup_dir, exist_ok=True)
    for db in db_list:
        if verbose:
            print(db, file=sys.stderr)
        db_name = db.split("/")[-1]
        source_db = get_db_fname_from_uri(db)
        backup_db = f'{backup_dir}/{db_name}'
        src_conn = sqlite3.connect(f'file:{source_db}?mode=ro', uri=True)
        dest_conn = sqlite3.connect(backup_db)
        src_conn.backup(dest_conn, pages=16384, progress=backup_progress if verbose else None)
