import re
import logging
from datetime import datetime
from typing import Any, Dict, Optional, List

from eportal import utils, db, dbutils, patches_constants as pc
from eportal.db.models import registered_server_ids
from eportal.patches.api import get_release_by_kernel_level

SQLITE_DATETIME_FMT = '%Y-%m-%d %H:%M:%S.%f'
logger = logging.getLogger(__name__)


def strip_version(name: str) -> Optional[str]:
    if name:
        return re.split(r'([_-]\d|\.so)', name)[0]
    return None  # pragma: no cover TODO


@utils.lru_cache_with_expire(10000, 300)
def get_kernel_release(kernel_id: str, level: int) -> Optional[str]:
    return get_release_by_kernel_level(pc.Stream.KERNEL, kernel_id, level)


@utils.lru_cache_with_expire(10000, 300)
def get_userspace_release(stream_type: str, build_id: str, level: int) -> Optional[str]:
    if stream := pc.Stream.items.get(stream_type):
        return get_release_by_kernel_level(stream, build_id, level)
    return None  # pragma: no cover, return for typechecker


def save_kernel_stat(data: List[Dict[str, Any]]) -> None:
    from statsd.defaults.env import statsd

    cleaned: List[dict] = list(filter(None, (prepare_kernel_checkin(it) for it in data)))
    if not cleaned:
        return

    flist = list(cleaned[0].keys())
    flist.remove('server_id')

    # NOTE: this is parametrized query without possible SQL injection
    set_expr = ', '.join(f'{f}=@{f}' for f in flist)
    q = f'UPDATE kc_servers SET {set_expr} WHERE server_id = @server_id'

    with dbutils.transaction(db.session):
        with dbutils.cursor(db.session) as cur:
            for checkin in cleaned:
                cur.execute(q, checkin)

    statsd.incr('eportal.success_checkins_kernel', len(cleaned))


def prepare_kernel_checkin(data: Dict[str, Any]) -> Optional[Dict[str, Any]]:
    server_id = data.get('server_id')
    if not server_id:
        return None

    return dict(
        server_id=server_id,
        node=data['node'],
        distro_version=data['distro_version'],
        distro=data['distro'],
        machine=data['machine'],
        processor=data['processor'],
        release=data['release'],
        version=data['version'],
        kcare_version=data['kcare_version'],
        virt=data['virt'],
        euname=data['euname'],
        remote_ip=data['ip'],
        patch_level=data['patch_level'],
        kernel_id=data['kernel_id'],
        uptime=data['uptime'],
        patch_type=data['patch_type'],
        ltimestamp=data['ltimestamp'] and datetime.utcfromtimestamp(data['ltimestamp']).strftime(SQLITE_DATETIME_FMT),
        last=datetime.utcfromtimestamp(data['last']).strftime(SQLITE_DATETIME_FMT),
        patchset=get_kernel_release(data['kernel_id'], data['patch_level']),
    )


def save_userspace_stat(data: List[Dict[str, Any]]) -> None:
    from statsd.defaults.env import statsd

    existing_server_ids = registered_server_ids([it['server_id'] for it in data])
    cleaned = [prepare_userspace_checkin(it) for it in data if it['server_id'] in existing_server_ids]
    skipped_count = len(data) - len(cleaned)
    if skipped_count:
        logger.debug('skipped %d userspace stat entries after server_id check', skipped_count)

    if not cleaned:
        return

    flist = list(cleaned[0].keys())

    columns_expr = ', '.join(flist)
    values_expr = ', '.join(f'@{f}' for f in flist)
    sql = f'INSERT OR REPLACE INTO kc_ucheckins ({columns_expr}) VALUES ({values_expr})'

    with dbutils.transaction(db.session):
        with dbutils.cursor(db.session) as cur:
            for checkin in cleaned:
                cur.execute(sql, checkin)

    statsd.incr('eportal.success_checkins_userspace', len(cleaned))


def prepare_userspace_checkin(data: Dict[str, Any]) -> Dict[str, Any]:
    server_id = data.get('server_id')
    return dict(
        server_id=server_id,
        product=pc.Stream.items[data['stream']].product.value,
        libname=strip_version(data['lib']),
        build_id=data['build_id'],
        patch_level=data['patch_level'],
        kcare_version=data['kcare_version'],
        last=datetime.utcfromtimestamp(data['last']).strftime(SQLITE_DATETIME_FMT),
        patchset=get_userspace_release(data['stream'], data['build_id'], data['patch_level']),
    )
