import os
import logging
import socket
import socketserver
import threading
from typing import Optional, Tuple

from dataclasses import dataclass

import config
from eportal import utils, anomalies, patches_constants as pc
from eportal.checkins import (
    start_server_info_processor,
    extract_server_info,
    extract_userspace_info,
    process_kernel_stat,
    process_userspace_stat,
    process_token_regs,
)

logger = logging.getLogger(__name__)

SOCK_PATH = '/run/eportal/checkin.sock'
UDP_RCVBUF_SIZE = 30 * (1024**2)  # 30 MB
STAT_LOG_STEP = 10_000


@dataclass
class ParsedMessage:
    remote_addr: str
    remote_user: str
    http_x_forwarded_for: str
    request_uri: str
    first_url_part: str
    access_token: str


def prepare_server_info(msg: ParsedMessage) -> Tuple[bool, Optional[dict]]:
    if not msg.remote_user:
        return False, None

    ip = utils.parse_proxy_header(msg.http_x_forwarded_for) or msg.remote_addr
    data = extract_server_info(ip=ip, url=msg.request_uri)
    data['server_id'] = msg.remote_user
    data['access_token'] = msg.access_token
    return True, data


def prepare_userspace_info(msg: ParsedMessage) -> Tuple[bool, Optional[dict]]:
    data = extract_userspace_info(url=msg.request_uri)
    if data:
        stream = pc.USER_PREFIX_STREAM_MAP.get(msg.first_url_part)
        if stream:
            data['server_id'] = msg.remote_user
            data['access_token'] = msg.access_token
            data['stream'] = stream.value
            data['ip'] = utils.parse_proxy_header(msg.http_x_forwarded_for) or msg.remote_addr
        else:
            data = None

    if not (data and data['patch_level'] > 0):
        return False, data

    return True, data


def parse_message(raw_msg: str) -> Optional[ParsedMessage]:
    def _handle_dash_value(val: str) -> str:
        return '' if val == '-' else val

    splitted = raw_msg.split('\t')
    if len(splitted) != 6:
        logger.warning('bad syslog message format: %s', raw_msg)
        return None

    _, remote_addr, remote_user, http_x_forwarded_for, request_uri, access_token = splitted
    http_x_forwarded_for = _handle_dash_value(http_x_forwarded_for)
    remote_user = _handle_dash_value(remote_user)

    url_split = request_uri.split('/')
    if len(url_split) < 3:
        logger.warning('bad request_uri format: %s', request_uri)
        return None

    first_url_part = url_split[1]

    return ParsedMessage(
        remote_addr=remote_addr,
        remote_user=remote_user,
        http_x_forwarded_for=http_x_forwarded_for,
        request_uri=request_uri,
        first_url_part=first_url_part,
        access_token=access_token,
    )


def process_message(raw_msg: str) -> int:
    parsed_msg = parse_message(raw_msg)
    if not parsed_msg:
        return 0

    logger.debug('parsed message: %s', parsed_msg)

    if parsed_msg.first_url_part in ('u', 'db', 'qemu'):
        queue = process_userspace_stat
        is_valid, data = prepare_userspace_info(parsed_msg)
    else:
        queue = process_kernel_stat
        is_valid, data = prepare_server_info(parsed_msg)

    if config.RESTORE_REGISTRATIONS_FROM_TOKEN and data and data.get('access_token'):
        process_token_regs.append(data)  # type: ignore[arg-type]

    if is_valid and data:
        logger.debug('enqueueing %s', data)
        queue.append(data)
        return 1

    return 0  # pragma: no cover


class MessageHandler(socketserver.BaseRequestHandler):
    processed = 0

    def handle(self) -> None:  # pragma: no cover TODO
        raw_msg = self.request[0].decode()
        try:
            logger.debug('recieved syslog message: "%s"', raw_msg)
            type(self).processed += process_message(raw_msg)
            if config.COUNT_CHECKIN_PROCESSING and type(self).processed % STAT_LOG_STEP == 0:
                logger.info('enqueued %d checkin messages', type(self).processed)

        except Exception:
            logger.exception('failed to enqueue message, %s', raw_msg)


if __name__ == "__main__":  # pragma: no cover
    os.umask(0)  # for socket permissions
    if os.path.exists(SOCK_PATH):
        os.unlink(SOCK_PATH)

    if config.PUSH_METRICS:  # pragma: no cover
        from eportal import metrics

        stop_statd_pusher = threading.Event()
        metrics.statsd_pusher(stop_statd_pusher)

    with socketserver.UnixDatagramServer(SOCK_PATH, MessageHandler) as server:
        stop_stat_processor = threading.Event()
        start_server_info_processor(stop_stat_processor)
        anomalies.agent_reboots_queue_processor(stop_stat_processor)
        server.socket.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, UDP_RCVBUF_SIZE)
        logger.info('starting UDP server, listening at %s', SOCK_PATH)
        server.serve_forever()
