import ast
import binascii
import logging
import json
import base64
import zlib
import time
import threading
from typing import List, Optional, Callable
from typing_extensions import TypedDict
from urllib.parse import urlparse
from urllib.parse import parse_qs

from werkzeug.wrappers import Request

from eportal import metrics, utils, replicated, patches_constants as pc, token
from eportal.db.models import registered_server_ids
from eportal.utils import BatchProcessor, process_int
from eportal.anomalies import process_agent_reboots, CheckinT

import config

log = logging.getLogger('eportal.checkins')


class TokenCheckin(TypedDict):
    server_id: str
    ip: str
    node: str
    access_token: str


def increase_dropped_checkins_metric(batch: List[CheckinT]) -> None:
    metrics.dropped_checkins += len(batch)


def start_server_info_processor(stop: threading.Event) -> Callable[[int], None]:
    k = process_kernel_stat.start(stop, 'Kernel stat processor error')
    u = process_userspace_stat.start(stop, 'Userspace stat processor error')
    tr = process_token_regs.start(stop, 'Registration token processor error')

    def join(timeout: int) -> None:
        k.join()
        u.join()
        tr.join(timeout)

    return join


CheckinBatchProcessor = BatchProcessor[CheckinT]


@CheckinBatchProcessor(timeout=config.CHECKIN_QUEUE_TIMEOUT, error_callback=increase_dropped_checkins_metric)
def process_kernel_stat(data: List[CheckinT]) -> None:
    replicated.checkin_server_many(data=data)
    process_agent_reboots(data=data)


@CheckinBatchProcessor(timeout=config.CHECKIN_QUEUE_TIMEOUT, error_callback=increase_dropped_checkins_metric)
def process_userspace_stat(data: List[CheckinT]) -> None:
    replicated.checkin_userspace(data=data)


TokenBatchProcessor = BatchProcessor[TokenCheckin]


@TokenBatchProcessor(timeout=config.CHECKIN_QUEUE_TIMEOUT)
def process_token_regs(data: List[TokenCheckin]) -> None:
    existing = registered_server_ids([it['server_id'] for it in data])
    for it in data:
        if it['server_id'] not in existing:
            register_from_checkin(it)


def save_server_info(server_id: Optional[str], request: Request) -> None:
    if server_id:
        data = extract_server_info(ip=utils.extract_ip(request), url=request.full_path)
        data['server_id'] = server_id
        process_kernel_stat.append(data)


def extract_server_info(ip: str, url: str, now: Optional[float] = None) -> CheckinT:
    base_url, _, data = url.partition('?')

    kernel_id = 'unknown'
    # like b16decode, but without expensive re.search check and thus faster
    data_decoded = binascii.unhexlify(data.encode()).decode(encoding='latin1') if data else None

    if base_url:
        url_parts = base_url.replace('//', '/').split('/')
        if len(url_parts) == 3:
            kernel_id = url_parts[1]
        elif len(url_parts) == 4:
            kernel_id = url_parts[2]

    result = parse_data(data_decoded, kernel_id)

    process_int(result, 'patch_level')
    process_int(result, 'uptime')
    process_int(result, 'ltimestamp')
    process_int(result, 'ts', default=None)
    process_int(result, 'last_stop', default=None)

    if result['ltimestamp'] > 0:
        result['ltimestamp'] = result['ltimestamp']
    else:
        result['ltimestamp'] = None

    result['ip'] = ip
    result['virt'] = result['virt'][:19].strip()
    result['last'] = now or time.time()
    return result


def parse_data(line: Optional[str], kernel_id: str) -> CheckinT:
    res = {
        'distro': '',
        'distro_version': '',
        'euname': '',
        'ip': '',
        'kcare_version': '',
        'kernel_id': kernel_id,
        'last': None,
        'ltimestamp': None,
        'machine': '',
        'node': '',
        'patch_level': -1,
        'patch_type': '',
        'processor': '',
        'release': '',
        'server_id': '',
        'uptime': -1,
        'version': '',
        'virt': '',
        'ts': None,
        'last_stop': None,
        'reason': '',
    }

    if line:
        res.update(ast.literal_eval(line))
    return res


def save_userspace_server_info(
    stream: pc.Stream, server_id: str, request: Request, now: Optional[float] = None
) -> None:
    data = extract_userspace_info(request.full_path, now)
    if data and data['patch_level'] > 0:
        data['server_id'] = server_id
        data['stream'] = stream.value
        process_userspace_stat.append(data)


def extract_userspace_info(url: str, now: Optional[float] = None) -> Optional[CheckinT]:
    parsed_url = urlparse(url)
    args = parse_qs(parsed_url.query)
    info = args.get('info', [''])[0]
    parts = parsed_url.path.replace('//', '/').split('/')

    if not info or len(parts) < 5:
        return None

    res: CheckinT = {
        'patch_level': -1,
        'ts': -1,
        'uptime': -1,
        'last': now or time.time(),
    }

    try:
        parsed = json.loads(zlib.decompress(base64.urlsafe_b64decode(info)))
    except Exception:
        return None

    res.update(parsed)
    res['build_id'] = parts[4]
    res['lib'] = parts[3]
    process_int(res, 'ts')
    process_int(res, 'uptime')
    process_int(res, 'patch_level')
    return res


def register_from_checkin(checkin: TokenCheckin) -> None:
    try:
        t = token.verify(checkin['access_token'], audience='agent', required_claims=['key', 'iat', 'server_id'])
    except Exception:
        if getattr(config, 'TEST_CONFIG', False):  # don't swallow exception in test context
            log.exception('Registration token verification failed')
        return

    if t['server_id'] != checkin['server_id']:
        return

    # don't try to register fresh tokens, may be it's a replication lag
    if t['iat'] > time.time() - config.RESTORE_REGISTRATIONS_MIN_TOKEN_AGE:
        return

    replicated.register_server_by_id(
        server_id=checkin['server_id'], ip=checkin['ip'], hostname=checkin['node'], key=t['key']
    )
