import json
import random
import string
from collections import defaultdict
from datetime import datetime, timedelta, timezone
from typing import Any, Dict, Iterable, Iterator, Type, Union, cast, List, Optional, Set, Tuple, Generator, TypeVar

from dataclasses import dataclass

from flask_login import UserMixin
from sqlalchemy import String, Integer, Text, Boolean, DateTime, ForeignKey, Index
from sqlalchemy import func, distinct, or_, select
from sqlalchemy.orm import relationship, Query, object_session
from sqlalchemy.orm.decl_api import Mapped, mapped_column
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.sql.elements import ColumnClause
from sqlalchemy.sql.functions import coalesce
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.ext.compiler import compiles
from sqlbind import WHERE, QueryParams, OR, not_none, WITH, AND_, cond
from typing_extensions import Self
from werkzeug.security import generate_password_hash, check_password_hash

from eportal import utils, perms, audit
from eportal.db import Base, session
from eportal.patches import api as patches_api
from eportal import patches_constants as pc
from eportal.dbutils import transaction, concat

import config
from .view_helper import view

DAY_DELTA = 30
DEFAULT_FEED = 'main'

# TODO KPT-4085 group with enum
DIMENSION_TOTAL = 'total'
DIMENSION_USAGE = 'usage'
DIMENSION_KERNELS = 'kernels'
DIMENSION_VERSIONS = 'versions'
DIMENSION_BUILDS = 'builds'
DIMENSION_DISTROS = 'distros'

_GenericModel = TypeVar('_GenericModel', bound=Base)


class NoSuchServer(Exception):
    pass


def get_unique_id() -> str:
    return ''.join(random.sample(string.ascii_letters + string.digits * 5, 16))


def generate_server_id() -> str:
    server_id = get_unique_id()
    while session.query(KcServer).filter_by(server_id=server_id).first() is not None:
        server_id = get_unique_id()  # pragma: no cover
    return server_id


@transaction(session)
def add_user(username: str, password: str, role: str = 'admin', description: Optional[str] = None) -> "User":
    u = User()
    u.username = username
    u.set_password(password)
    u.active = True
    u.set_role(role)
    u.description = description
    session.add(u)
    return u


@transaction(session)
def delete_user(username: str) -> Optional["User"]:
    u = get_user(username)
    if u:
        session.delete(u)
    return u


@transaction(session)
def toggle_user_readonly(username: str, readonly: Optional[bool] = None) -> "User":
    u = get_user(username)
    if not u:
        raise RuntimeError(f'user "{username}" not found')
    if readonly is None:
        u.readonly = not u.readonly
    else:
        u.readonly = readonly
    return u


def get_all_users() -> List["User"]:
    users = session.query(User).all()
    return users


def get_all_unit_users(unit_id: int) -> List["UnitUser"]:
    return session.query(UnitUser).filter_by(unit_id=unit_id).all()


def get_verified_user(
    username: str, password: str, unit_id: Optional[int] = None
) -> Optional[Union["User", "UnitUser"]]:
    u = get_user(username, unit_id)
    if u and u.verify_password(password):
        return u
    return None


def get_user(username: str, unit_id: Optional[int] = None) -> Optional[Union["User", "UnitUser"]]:
    if unit_id:
        return session.query(UnitUser).filter_by(username=username, unit_id=unit_id).first()
    else:
        return session.query(User).filter_by(username=username).first()


class AuditMixin:
    audit_action: str
    audit_obj_type: str
    audit_delete_desc_field: str

    def audit_delete_msg(self) -> Tuple[str, str]:
        return self.audit_action, f'deleted {self.audit_obj_type} {self.audit_delete_desc()}'

    def audit_delete_desc(self) -> str:
        return cast(str, getattr(self, self.audit_delete_desc_field))

    def audit_change_msg(self) -> Tuple[str, str]:
        data: Dict[str, Any] = self.to_json()  # type: ignore[attr-defined]
        data.pop('id', None)
        data.pop('password_hash', None)
        return self.audit_action, f'changed {self.audit_obj_type} {json.dumps(data, ensure_ascii=False)}'


class User(AuditMixin, UserMixin, Base):  # type: ignore[misc]
    __tablename__ = 'kc_users'

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    username: Mapped[str] = mapped_column(String(64), unique=True, nullable=False, index=True)
    password_hash: Mapped[str] = mapped_column(String(255), nullable=False)
    active: Mapped[bool] = mapped_column(Boolean, nullable=False, default=False)
    description: Mapped[Optional[str]] = mapped_column(Text, nullable=True)
    readonly: Mapped[bool] = mapped_column(Boolean, nullable=False, default=False, server_default='0')

    audit_action = audit.ACTION_MANAGE_USER
    audit_obj_type = 'user'
    audit_delete_desc_field = 'username'

    def set_password(self, password: str) -> None:
        self.password_hash = generate_password_hash(password, method="pbkdf2")

    def verify_password(self, password: str) -> bool:
        return check_password_hash(self.password_hash, password)

    def __repr__(self) -> str:  # pragma: no cover TODO
        return "<User(id='%s', username='%s', description='%s')>" % (self.id, self.username, self.description)

    @property
    def permissions(self) -> perms.PermSet:
        if self.readonly:
            return perms.Roles.VIEWER.permissions
        else:
            return perms.Roles.ADMIN.permissions

    @property
    def role(self) -> str:
        return ['admin', 'readonly'][bool(self.readonly)]

    def set_role(self, role: str) -> None:
        self.readonly = role != 'admin'

    def to_json(self, public: bool = False) -> Dict[str, Any]:
        result = {
            'id': self.id,
            'username': self.username,
            'password_hash': self.password_hash,
            'active': self.active,
            'description': self.description,
            'readonly': self.readonly,
        }

        if public:
            del result['password_hash']
            del result['active']
            return result
        else:
            return result


def get_key(key: str) -> Optional['KcKey']:
    return session.query(KcKey).filter_by(key=key).first()


def get_key_mut(unit_id: int, key: str) -> Optional['KcKey']:
    return session.query(KcKey).filter_by(key=key, unit_id=unit_id).first()


def get_keys(unit_id: int) -> Query['KcKey']:
    return session.query(KcKey).filter_by(unit_id=unit_id)


def get_feed(unit_id: int, name: str) -> Optional['Feed']:
    return session.query(Feed).filter_by(unit_id=unit_id, name=name).first()


def get_feeds(unit_id: int) -> Query['Feed']:
    return session.query(Feed).filter_by(unit_id=unit_id)


class KcKey(AuditMixin, Base):
    __tablename__ = 'kc_keys'

    unit_id: Mapped[int] = mapped_column(Integer, nullable=False, default=0, server_default='0')
    key: Mapped[str] = mapped_column(String(16), primary_key=True)
    enabled: Mapped[Optional[bool]] = mapped_column(Boolean, default=True)
    server_limit: Mapped[int] = mapped_column(Integer, default=0, server_default='0')
    note: Mapped[Optional[str]] = mapped_column(String(256))
    feed: Mapped[Optional[str]] = mapped_column(Text)
    products: Mapped[Optional[str]] = mapped_column(Text, nullable=True)

    audit_action = audit.ACTION_MANAGE_KEY
    audit_obj_type = 'key'
    audit_delete_desc_field = 'key'

    servers: Mapped[List['KcServer']] = relationship(
        'KcServer', back_populates='kc_key_rel', cascade='all, delete-orphan', passive_deletes=True
    )

    def __repr__(self) -> str:  # pragma: no cover TODO
        products_str = ", ".join(self.products_list or [])
        return (
            f'{self.key}\tproducts={products_str}\tfeed={self.public_feed}'
            f'\tserver_limit={self.server_limit}\tnote={self.note}'
        )

    @property
    def public_feed(self) -> str:
        return self.feed or DEFAULT_FEED

    def set_feed(self, feed: Optional[str]) -> None:
        if feed == DEFAULT_FEED:
            feed = None
        self.feed = feed

    @property
    def products_list(self) -> Optional[List[str]]:
        if self.products is None:
            return None
        return cast(List[str], json.loads(self.products))

    def set_products(self, products_raw: Optional[List[Union[str, pc.Product]]]) -> None:
        products = [it.value if isinstance(it, pc.Product) else it for it in products_raw or []]
        unsupported: Set[str] = set(products or []) - set(pc.Product.items)
        if unsupported:
            raise ValueError(f'unsupported products: {sorted(unsupported)}')

        self.products = json.dumps(sorted(products)) if products else None

    def to_json(self, public: bool = False) -> Dict[str, Any]:
        result: Dict[str, Any] = {
            'key': self.key,
            'server_limit': self.server_limit,
            'note': self.note,
            'feed': self.feed,
        }

        if public:
            result['feed'] = self.public_feed
            result['products'] = self.products_list
        else:
            result['unit_id'] = self.unit_id
            result['enabled'] = self.enabled
            result['products'] = self.products  # for replication
        return result

    def server_count(self) -> int:
        return session.query(KcServer).filter_by(kc_key=self.key).count()

    def limit_reached(self) -> bool:
        # server_limit = 0 is considered unlimited
        return 0 < self.server_limit <= self.server_count()

    @staticmethod
    def find_default_feed(unit_id: int) -> bool:
        result = session.query(KcKey).filter_by(unit_id=unit_id, feed=None).count()
        return result != 0

    @staticmethod
    def delete(unit_id: int, key: str) -> None:
        # todo: add cascading delete
        with transaction(session):
            session.query(KcKey).filter_by(unit_id=unit_id, key=key).delete()

    @staticmethod
    def get_ps_product(products_json: Optional[str]) -> pc.PSProduct:
        return pc.PSProduct.KCE

    def validate(self) -> None:
        return None


@transaction(session)
def add_server(ip: str, hostname: str, key: str, server_id: Optional[str] = None) -> "KcServer":
    if config.BUNITS:
        # dedup hosts only for a single key in BU mode (it's good enough for now)
        # it requires major changes to perform dedup for a whole unit
        server = KcServer.get_last_server(remote_ip=ip, node=hostname, kc_key=key)
    else:
        server = KcServer.get_last_server(remote_ip=ip, node=hostname)

    if server:  # delete duplicates
        q = session.query_params()
        sql = f'''\
            DELETE FROM kc_servers
            WHERE remote_ip = {q/ip}
              AND node = {q/hostname}
              AND server_id != {q/server.server_id}
             {AND_(q.kc_key == cond(config.BUNITS)/key)}
        '''
        q.execute(sql)
        server.kc_key = key
        if server_id:
            server.server_id = server_id
    else:
        server = KcServer(kc_key=key, remote_ip=ip, node=hostname, server_id=server_id)
        session.add(server)
    return server


def get_server(server_id: str) -> Optional["KcServer"]:
    kc_server = session.query(KcServer).filter_by(server_id=server_id).first()
    return kc_server


@transaction(session)
def delete_server(server_id: str) -> None:
    KcServer.delete(server_id)


@transaction(session)
def delete_servers(
    server_ids: Optional[List[str]] = None,
    ips: Optional[List[str]] = None,
    hosts: Optional[List[str]] = None,
    last_check_in_age: Optional[int] = None,
) -> int:
    if not any(
        (
            server_ids,
            ips,
            hosts,
            last_check_in_age,
        )
    ):
        raise ValueError('At least one argument should be provided.')

    # NOTE:
    #   "last date" is applicable only for those servers who subjected to check in.
    #   It seems to be higher level business logic, but it handled in this function
    #   in order to prevent mistakes.
    if last_check_in_age:
        last_date = datetime.utcnow() - timedelta(days=last_check_in_age)
        # PERFORMANCE NOTE:
        #   Pre-fetched keys array is 32%...36% faster than single request with JOIN tables:
        #     0.651s...1.573s instead of 1.018s...2.411s
        #     (measured on 500K...1M servers and 10...50 keys)
        server_keys_subjected_to_checkins = [
            si.key for si in iter_kc_key_fields(products_required=pc.PRODUCTS_INDUCING_SERVER_CHECKINS)
        ]
    else:
        last_date = None
        server_keys_subjected_to_checkins = None

    q = session.query_params()
    filters = [
        q.server_id.IN(server_ids),
        q.remote_ip.IN(ips),
        q.node.IN(hosts),
        q.kc_key.IN(server_keys_subjected_to_checkins),
        q.last < not_none / last_date,
    ]

    return q.execute(f'DELETE FROM kc_servers {WHERE(*filters)}').rowcount


def get_servers(key: Optional[str] = None) -> List[tuple]:
    q = session.query_params()
    sql = f'''
        SELECT kc_key, remote_ip, server_id, last, uptime, kcare_version, node FROM kc_servers
        {WHERE(q.kc_key == not_none/key)}
        ORDER BY last DESC
    '''
    return q.execute(sql)


class Tag(Base):
    __tablename__ = 'tag'
    tag_id: Mapped[int] = mapped_column(Integer, nullable=False, primary_key=True, autoincrement=True)
    unit_id: Mapped[int] = mapped_column(Integer, nullable=False, default=0, server_default='0')
    name: Mapped[str] = mapped_column(String, nullable=False)

    @staticmethod
    def get_tags(unit_id: int) -> List[str]:
        """
        Get all tags
        :return: set of tags
        """
        return sorted(set([name for name, in session.query(Tag.name).filter_by(unit_id=unit_id)]))

    @staticmethod
    def tag_to_map(tag: str) -> Dict[str, Any]:
        result: Dict[str, Any] = {}
        lines = tag.split(';')
        for line in lines:
            if ':' in line:
                name, value = line.split(':', 1)
                result[name] = value
            else:
                if line:
                    result[line] = True
        return result


Index('ix_tag_unit_id_name', Tag.unit_id, Tag.name, unique=True)


class KcServerTag(Base):
    __tablename__ = 'kc_servers_tag'
    server_id: Mapped[str] = mapped_column(ForeignKey('kc_servers.server_id', ondelete="CASCADE"), primary_key=True)
    tag_id: Mapped[int] = mapped_column(ForeignKey('tag.tag_id', ondelete="CASCADE"), primary_key=True)
    tag_value: Mapped[str] = mapped_column(String, nullable=True)
    tag = relationship('Tag', lazy='joined')

    @classmethod
    def from_mltag(cls, unit_id: int, mltag: Optional[str]) -> List[Self]:
        '''Creates tag assignments list from mltag representation, unknown tags are implicitely created'''

        if not mltag:
            return []

        tags_by_name = {t.name: t for t in session.query(Tag).filter_by(unit_id=unit_id).all()}
        result: List[Self] = []

        has_new_tags = False
        for item in mltag.split(';'):
            name, *values = item.split(':', 1)
            name = name.strip()
            if not name:
                continue
            value = values[0].strip() if values else None
            if name in tags_by_name:
                tag = tags_by_name[name]
            else:
                tag = Tag(unit_id=unit_id, name=name)
                session.add(tag)
                has_new_tags = True
            result.append(cls(tag=tag, tag_value=value))

        if has_new_tags:
            session.flush()
        return result


class RawSqlPlaceholder(ColumnClause):
    inherit_cache = True


@compiles(RawSqlPlaceholder)  # type: ignore
def compile_as_is(element: ColumnClause, compiler, **kw):
    return element.name


class KcServer(Base):
    __tablename__ = 'kc_servers'

    server_id: Mapped[str] = mapped_column(String(16), primary_key=True, default=generate_server_id)
    created: Mapped[Optional[datetime]] = mapped_column(DateTime, default=datetime.utcnow)
    kc_key: Mapped[Optional[str]] = mapped_column(String(16), ForeignKey('kc_keys.key', ondelete='CASCADE'))
    kc_key_rel: Mapped['KcKey'] = relationship(
        'KcKey',
        back_populates='servers',
        order_by=created,
    )
    remote_ip: Mapped[str] = mapped_column(String(40), nullable=False)
    patch_level: Mapped[Optional[int]] = mapped_column(Integer)
    uptime: Mapped[Optional[int]] = mapped_column(Integer)
    kernel_id: Mapped[Optional[str]] = mapped_column(String(60))
    distro_version: Mapped[Optional[str]] = mapped_column(String(30))
    distro: Mapped[Optional[str]] = mapped_column(String(64))
    node: Mapped[Optional[str]] = mapped_column(String(256))
    machine: Mapped[Optional[str]] = mapped_column(String(10))
    processor: Mapped[Optional[str]] = mapped_column(String(10))
    release: Mapped[Optional[str]] = mapped_column(String(128))
    version: Mapped[Optional[str]] = mapped_column(String(128))
    kcare_version: Mapped[Optional[str]] = mapped_column(String(20))
    virt: Mapped[Optional[str]] = mapped_column(String(20))
    euname: Mapped[Optional[str]] = mapped_column(String(128))
    last: Mapped[Optional[datetime]] = mapped_column(DateTime)
    ltimestamp: Mapped[Optional[datetime]] = mapped_column(DateTime)
    patch_type: Mapped[Optional[str]] = mapped_column(String(20))
    patchset: Mapped[Optional[str]] = mapped_column(Text())

    server_tags = relationship('KcServerTag', lazy='select', cascade="save-update, merge, delete, delete-orphan")
    _mltag_rel = relationship('MltagView', lazy='joined', uselist=False, viewonly=True)

    @hybrid_property
    def mltag(self) -> Optional[str]:
        return self._mltag_rel.mltag if self._mltag_rel else None

    @mltag.inplace.expression
    @classmethod
    def _mltag_expr(cls) -> ColumnClause:
        # Hack to bypass double join or subqueries while searching or sorting by mltag
        return RawSqlPlaceholder('kc_servers_mltag_view_1.mltag')

    @mltag.inplace.setter
    def _set_mltag(self, value: Optional[str]) -> None:
        if value:
            # BU is linked with the server via the registration key
            unit_id = self.kc_key_rel.unit_id if self.kc_key_rel else 0
            self.server_tags = KcServerTag.from_mltag(unit_id, value)
        else:
            self.server_tags = []
        if current_session := object_session(self):
            # mltag view relation will not be auto-reloaded after session flush
            current_session.flush()
            current_session.refresh(self)

    def to_json(self) -> Dict[str, Any]:  # pragma: no cover
        raise NotImplementedError('to_json is not implemented for KcServer')

    @classmethod
    def get_last_server(cls, **kwargs: Any) -> Optional[Self]:
        return session.query(cls).filter_by(**kwargs).order_by(func.coalesce(cls.last, cls.created).desc()).first()

    @staticmethod
    def delete(server_id: str) -> None:
        with transaction(session):
            session.query(KcServer).filter(KcServer.server_id == server_id).delete()

    @staticmethod
    def count() -> int:  # it seems used only in tests
        return session.query(KcServer).count()

    @staticmethod
    def count_grouped_by_key() -> Any:  # CLI
        return (
            session.query(KcKey.key.label('key_id'), func.count(KcServer.server_id).label('servers_count'))
            .outerjoin(KcServer)
            .group_by(KcKey.key)
            .all()
        )


Index('idx_kc_servers_ip_node', KcServer.remote_ip, KcServer.node)


class MltagView(Base):
    _ordered_tags_query = (
        select(KcServerTag.server_id, coalesce(Tag.name + ':' + KcServerTag.tag_value, Tag.name).label('tag'))
        .join_from(KcServerTag, Tag)
        .order_by(Tag.name)
        .subquery()
    )
    _mltags_query = select(
        _ordered_tags_query.c.server_id, func.group_concat(_ordered_tags_query.c.tag, ';').label('mltag')
    ).group_by(_ordered_tags_query.c.server_id)
    __table__ = view('kc_servers_mltag_view', Base.metadata, _mltags_query)


class KcUserspaceCheckins(Base):
    __tablename__ = 'kc_ucheckins'

    server_id: Mapped[str] = mapped_column(Text, primary_key=True)
    product: Mapped[str] = mapped_column(Text, primary_key=True)
    libname: Mapped[str] = mapped_column(Text, primary_key=True)
    build_id: Mapped[Optional[str]] = mapped_column(Text)
    patch_level: Mapped[Optional[int]] = mapped_column(Integer)
    patchset: Mapped[Optional[str]] = mapped_column(Text)
    kcare_version: Mapped[Optional[str]] = mapped_column(Text)
    last: Mapped[Optional[datetime]] = mapped_column(DateTime)


class Feed(AuditMixin, Base):
    __tablename__ = 'feed'

    unit_id: Mapped[int] = mapped_column(Integer, nullable=False, primary_key=True, default=0, server_default='0')
    name: Mapped[str] = mapped_column(Text, primary_key=True)
    auto: Mapped[Optional[bool]] = mapped_column(Boolean, default=False)
    deploy_after: Mapped[Optional[int]] = mapped_column(Integer, default=0)
    channel: Mapped[Optional[str]] = mapped_column(
        Text, default=pc.Channel.DEFAULT.value, server_default=pc.Channel.DEFAULT.value
    )

    audit_action = audit.ACTION_MANAGE_FEED
    audit_obj_type = 'feed'
    audit_delete_desc_field = 'name'

    def to_json(self, public: bool = False) -> Dict[str, Any]:
        result = {
            'name': self.name,
            'auto': self.auto,
            'deploy_after': self.deploy_after,
            'channel': self.channel,
        }
        if not public:
            result['unit_id'] = self.unit_id
        return result

    @staticmethod
    def get_all_names(unit_id: int) -> List[str]:
        return ['main'] + sorted([name for name, in session.query(Feed.name).filter_by(unit_id=unit_id)])

    @staticmethod
    def autoupdated_count() -> int:
        return session.query(Feed.name).filter(Feed.auto.is_(True)).count()

    @staticmethod
    def delayed_count() -> int:
        return session.query(Feed.name).filter(Feed.deploy_after > 0).count()

    @staticmethod
    def fetch_spec(unit_id: int, feed_name: Optional[str]) -> pc.PSFetchSpec:
        feed_name = feed_name or DEFAULT_FEED
        feed_channel = None
        if feed_name == DEFAULT_FEED:
            key = session.query(KcKey).filter(KcKey.feed.is_(None)).order_by('key').first()
        else:
            feed = get_feed(unit_id, feed_name)
            if feed and feed.channel:
                feed_channel = pc.Channel.items.get(feed.channel)  # safe get to avoid db garbage
            key = session.query(KcKey).filter_by(unit_id=unit_id, feed=feed_name).order_by('key').first()

        return pc.PSFetchSpec(KcKey.get_ps_product(key.products if key else None), feed_channel or pc.Channel.DEFAULT)

    @staticmethod
    def get_feeds_by_fetch_spec(unit_id: int) -> Dict[pc.PSFetchSpec, List[str]]:
        all_feeds = Feed.get_all_names(unit_id)

        result: Dict[pc.PSFetchSpec, List[str]] = {}
        for feed in all_feeds:
            result.setdefault(Feed.fetch_spec(unit_id, feed), []).append(feed)
        return result

    @staticmethod
    def get_feeds_by_product(unit_id: int) -> Dict[pc.PSProduct, List[str]]:
        result: Dict[pc.PSProduct, List[str]] = {}
        for spec, feedlist in Feed.get_feeds_by_fetch_spec(unit_id).items():
            result.setdefault(spec.psproduct, []).extend(feedlist)
        return result

    def on_delete(self) -> None:
        session.query(KcKey).filter_by(unit_id=self.unit_id, feed=self.name).update({'feed': None})
        patches_api.remove_feed(self.unit_id, self.name)

    def __repr__(self) -> str:
        return "{0.name}\tauto={0.auto}\tdeploy_after={0.deploy_after}\tchannel={0.channel}".format(self)


class Unit(AuditMixin, Base):
    __tablename__ = 'units'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    slug: Mapped[str] = mapped_column(String, unique=True, index=True)
    title: Mapped[str] = mapped_column(String)

    audit_action = audit.ACTION_MANAGE_UNIT
    audit_obj_type = 'unit'
    audit_delete_desc_field = 'slug'

    @staticmethod
    def by_slug(slug: str) -> Optional['Unit']:
        return session.query(Unit).filter_by(slug=slug).first()

    @staticmethod
    def titles() -> Dict[int, str]:
        q = session.query_params()
        result = {0: 'Default'}
        result.update(q.execute('SELECT id, title FROM units ORDER BY title'))
        return result

    @staticmethod
    def slugs() -> Dict[str, str]:
        q = session.query_params()
        return dict(q.execute('SELECT slug, title FROM units ORDER BY title'))

    @staticmethod
    def unit_ids() -> List[int]:
        return session.query_params().execute('SELECT id FROM units').column()

    def to_json(self) -> Dict[str, Any]:
        return {'id': self.id, 'slug': self.slug, 'title': self.title}


class UnitUser(AuditMixin, UserMixin, Base):  # type: ignore[misc]
    __tablename__ = 'unit_users'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    unit_id: Mapped[int] = mapped_column(Integer, ForeignKey(Unit.id), nullable=False)
    username: Mapped[str] = mapped_column(String, nullable=False)
    password_hash: Mapped[str] = mapped_column(String, nullable=False)
    description: Mapped[Optional[str]] = mapped_column(Text, nullable=True)
    role: Mapped[Optional[str]] = mapped_column(String, nullable=True)

    audit_action = audit.ACTION_MANAGE_USER
    audit_obj_type = 'unit user'
    audit_delete_desc_field = 'username'

    def set_password(self, password: str) -> None:
        self.password_hash = generate_password_hash(password, method="pbkdf2")

    def verify_password(self, password: str) -> bool:
        return check_password_hash(self.password_hash, password)

    @property
    def permissions(self) -> perms.PermSet:
        return perms.UnitRoles.get(self.role).permissions

    def available_units(self) -> List[Tuple[str, str]]:
        q = session.query_params()
        sql = f'''\
            SELECT slug, title
            FROM units u
            INNER JOIN unit_users ON u.id == unit_id
            WHERE username = {q/self.username}
              AND role is NOT NULL
              AND role != {q/perms.UnitRoles.INACTIVE.value}
        '''
        return q.execute(sql)

    def to_json(self, public: bool = False) -> Dict[str, Any]:
        result = {
            'id': self.id,
            'unit_id': self.unit_id,
            'username': self.username,
            'description': self.description,
            'role': self.role,
        }

        if not public:
            result['password_hash'] = self.password_hash

        return result


Index('idx_unit_users_uniq_username', UnitUser.unit_id, UnitUser.username, unique=True)


@transaction(session)
def tag_server(server_id: str, raw_tags: Optional[str]) -> bool:
    rv = (
        session.query(KcServer, KcKey.unit_id)
        .filter(KcServer.server_id == server_id, KcKey.key == KcServer.kc_key)
        .first()
    )
    if not rv:
        raise NoSuchServer

    kc_server, unit_id = rv

    tags = KcServerTag.from_mltag(unit_id, raw_tags)
    kc_server.server_tags = tags
    # we need to reload mltag view relation manually for nested transaction case
    session.flush()
    session.refresh(kc_server)
    return True


def get_statistics(
    dimension: str = '',
    active: bool = False,
    products_required: Tuple[pc.Product, ...] = cast(Tuple[pc.Product, ...], ()),
    products_excluded: Tuple[pc.Product, ...] = cast(Tuple[pc.Product, ...], ()),
) -> utils.Table:
    group_by: tuple
    fields: Optional[tuple] = None

    if dimension == DIMENSION_USAGE:
        group_by = (
            KcServer.patchset,
            KcServer.kernel_id,
        )
    elif dimension == DIMENSION_VERSIONS:
        group_by = (KcServer.kcare_version,)
    elif dimension == DIMENSION_TOTAL:
        group_by = ()
    elif dimension == DIMENSION_DISTROS:
        fields = ((KcServer.distro + '-' + KcServer.distro_version).label('distro'),)
        group_by = (KcServer.distro, KcServer.distro_version)
    else:
        group_by = (KcServer.kernel_id,)

    if not fields:
        fields = group_by

    rows: Iterable[Tuple[Any, ...]]
    if products_required or products_excluded:
        rows = _count_servers_with_products(
            group_by=group_by,
            active=active,
            products_required=products_required,
            products_excluded=products_excluded,
        )
    else:
        rows = _count_servers(KcServer, fields=fields, group_by=group_by, active=active)

    return utils.Table(titles=tuple(f.name for f in fields) + ('count',), rows=tuple(rows))


def _count_servers(
    table: Union[Type[KcServer], Type[KcUserspaceCheckins]], fields: tuple, group_by: tuple, active: bool
) -> List[tuple]:
    q = session.query(*fields, func.count(distinct(cast(ColumnClause, table.server_id))))
    q.order_by(*fields)
    if active:
        # TODO unify now/utcnow usage for filters like this
        limit_date = datetime.utcnow() - timedelta(days=DAY_DELTA)
        q = q.filter(cast(datetime, table.last) > limit_date)  # type: ignore[arg-type]
    return q.group_by(*group_by).all()


def _count_servers_with_products(
    group_by: tuple,
    active: bool,
    products_required: Tuple[pc.Product, ...],
    products_excluded: Tuple[pc.Product, ...],
) -> Generator[tuple, None, None]:
    if not (products_excluded or products_required):  # pragma: no cover
        raise ValueError('Required "product_excluded" and/or "product_required" args, got neither')
    if KcKey.products in group_by:  # pragma: no cover
        raise NotImplementedError('Case when "products" already preset in group_by fields is not required yet')
    group_by_with_products = (KcKey.products, *group_by)
    q = session.query(*group_by_with_products, func.count(KcServer.server_id)).join(KcServer.kc_key_rel)
    if active:
        limit_date = datetime.now() - timedelta(days=DAY_DELTA)
        q = q.filter(KcServer.last > limit_date)

    count__by__grouped_by_values__map: Dict[tuple, int] = defaultdict(int)
    for key_products, *grouped_by_values, count in q.group_by(*group_by_with_products).all():
        try:
            key_products = key_products and json.loads(key_products)
        except Exception:  # pragma: no cover
            key_products = None
        if not all(is_product_allowed(product_required, key_products) for product_required in products_required):
            continue
        if any(is_product_allowed(product_excluded, key_products) for product_excluded in products_excluded):
            continue
        count__by__grouped_by_values__map[tuple(grouped_by_values)] += count

    for grouped_by_values, count in count__by__grouped_by_values__map.items():
        yield (*grouped_by_values, count)


def get_number_of_keys() -> int:
    q = session.query(KcKey.key)
    return q.count()


def get_userspace_statistics(dimension: str, active: bool = False) -> utils.Table:
    T = KcUserspaceCheckins
    fields: Tuple[Union[InstrumentedAttribute[str], InstrumentedAttribute[Optional[str]]], ...]
    if dimension == DIMENSION_USAGE:
        fields = T.product, T.patchset, T.build_id
    else:
        fields = (T.product,)

    rows = _count_servers(T, fields=fields, group_by=fields, active=active)

    return utils.Table(titles=tuple(cast(str, f.name) for f in fields) + ('count',), rows=rows)


@transaction(session)
def cleanup_userspace_checkins() -> None:
    sql = 'DELETE FROM kc_ucheckins WHERE server_id NOT IN (SELECT server_id FROM kc_servers)'
    session.query_params().execute(sql)


def patch_level_statistics_kernel(unit_id: int, active: bool = False) -> patches_api.PatchLevelStat:
    q = session.query_params()
    limit_date = (datetime.now() - timedelta(days=DAY_DELTA)) if active else None
    filters = [
        's.patch_level > 0',
        q.s.last > not_none / limit_date,
        q.unit_id == unit_id,
    ]
    sql = f'''
        SELECT coalesce(feed, 'main'), coalesce(patchset, 'UNKNOWN'), count(server_id)
        FROM kc_servers s
        LEFT JOIN kc_keys ON kc_key=key
        {WHERE(*filters)}
        GROUP BY feed, patchset
        '''
    return q.execute(sql)


def patch_level_statistics_userspace(unit_id: int, product: str, active: bool = False) -> patches_api.PatchLevelStat:
    q = session.query_params()
    limit_date = (datetime.now() - timedelta(days=DAY_DELTA)) if active else None
    filters = [
        's.patch_level > 0',
        q.s.last > not_none / limit_date,
        q.k.unit_id == unit_id,
    ]
    product_filter = q.product == product
    sql = f'''
        SELECT
            coalesce(k.feed, 'main') AS feed,
            coalesce(kcu.patchset, 'UNKNOWN') AS patchset,
            count(kcu.server_id) AS server_count
        FROM (SELECT DISTINCT server_id, product, patchset FROM kc_ucheckins {WHERE(product_filter)}) kcu
        INNER JOIN kc_servers s USING(server_id)
        INNER JOIN kc_keys k ON s.kc_key = k.key
        {WHERE(*filters)}
        GROUP BY k.feed, kcu.patchset
    '''
    return q.execute(sql)


def server_statistics(active: bool = False) -> int:
    q = session.query(func.count(KcServer.server_id))
    if active:
        limit_date = datetime.now() - timedelta(days=DAY_DELTA)
        q = q.filter(KcServer.last > limit_date)
    return cast(int, q.scalar())


def _update_model_fields(obj: Any, attrs: Dict[str, Any], skip_none: bool = True) -> None:
    for k, v in list(attrs.items()):
        if v is None and skip_none:
            continue

        setter = f'set_{k}'
        if hasattr(obj, setter):
            getattr(obj, setter)(v)
        else:
            setattr(obj, k, v)

    if hasattr(obj, 'validate'):
        obj.validate()


@transaction(session)
def create_object(cls: Type[_GenericModel], flt: Dict[str, Any], **attrs: Any) -> Optional[_GenericModel]:
    exc = attrs.pop('exc', None)
    check_record = attrs.pop('check_record', True)

    if hasattr(cls, 'unit_id') and 'unit_id' not in flt:
        flt['unit_id'] = attrs.get('unit_id', 0)

    if check_record and session.query(cls).filter_by(**flt).first():
        if exc:
            raise exc
        return None
    attrs.update(flt)

    obj = cls()
    _update_model_fields(obj, attrs)
    session.add(obj)
    return obj


@transaction(session)
def modify_object(cls: Type[_GenericModel], flt: Dict[str, Any], **attrs: Any) -> Optional[_GenericModel]:
    exc = attrs.pop('exc', None)

    if hasattr(cls, 'unit_id') and 'unit_id' not in flt:
        flt['unit_id'] = 0

    obj: Optional[_GenericModel] = session.query(cls).filter_by(**flt).first()
    if obj is None:
        if exc:
            raise exc
        return None

    _update_model_fields(obj, attrs)
    return obj


@transaction(session)
def ensure_object(cls: Type[_GenericModel], flt: Dict[str, Any], **attrs: Any) -> _GenericModel:
    obj = modify_object(cls, flt, **attrs)
    if not obj:
        obj = create_object(cls, flt, check_record=False, **attrs)
    return cast(_GenericModel, obj)


def age_filter(q_params: QueryParams, name: str, value: Optional[int]) -> str:
    if value is None:
        return ''

    dt = datetime.utcnow() - timedelta(seconds=abs(value))
    if value >= 0:
        return OR(q_params(f'{name} < {{}}', dt), f'{name} is NULL')
    else:
        return q_params(f'{name} >= {{}}', dt)


def fix_server_list_field_name(result: List[Dict[str, Any]]) -> Iterator[Dict[str, Any]]:
    for it in result:
        it['id'] = it.pop('server_id')
        it['key'] = it.pop('kc_key')
        it['feed'] = it.pop('cfeed')
        it['registered'] = it.pop('created')
        it['checkin'] = it.pop('last')
        it['updated'] = it.pop('ltimestamp')
        it['tags'] = it.pop('mltag')
        it['hostname'] = it.pop('node')
        it['ip'] = it.pop('remote_ip')
        yield it


def parse_server_list_datetime_fields(result: Iterable[Dict[str, Any]]) -> Iterator[Dict[str, Any]]:
    for it in result:
        for field in ['registered', 'checkin', 'updated']:
            if date_str := it.get(field):
                it[field] = datetime.fromisoformat(date_str).replace(tzinfo=timezone.utc)

        yield it


def get_server_list(
    unit_id: Optional[int] = None,
    feed: Optional[str] = None,
    key: Optional[str] = None,
    registered_age: Optional[int] = None,
    checkin_age: Optional[int] = None,
    updated_age: Optional[int] = None,
    is_updated: Optional[bool] = None,
    hostname: Optional[str] = None,
    ip: Optional[str] = None,
    server_id: Optional[str] = None,
    limit: int = 10,
    offset: int = 0,
    only_count: Optional[bool] = False,
    tag: Optional[str] = None,
) -> Tuple[int, List[Dict]]:
    """Helper function to back /admin/api/servers endpoint

    See api.md for parameter meaning.

    Function constructs WHERE clause with regard to non-None filter values
    to get list of available servers.

    Most of it are trivial but is_updated is tricky one: we need
    to join server kernel_id with latest available patch for this
    server accounting its feed. To accomplish this we use subquery.
    But here is a catch: patch_level and max result can be NULL
    and NULL compare operations always return FALSE, so we use
    coalesce to compare result.

    Another non-obvious part is feed naming. In kc_keys default feed
    is NULL, in patches it is 'main'. And we have to use coalesce
    to get proper value.

    In the end we perform two queries to get total server count and
    desired server list.

    Table aliases:
        s: kc_servers
        k: kc_keys
        kp: kernel_patches.patches
        kr: kernel_patches.releases
    """
    q = session.query_params()

    if feed == DEFAULT_FEED:
        feed_flt = 'k.feed is NULL'
    else:
        feed_flt = q.k.feed == not_none / feed

    latest_pl_cte = ''
    latest_pl_join = ''
    upd_flt = ''
    if is_updated is not None:
        data = patches_api.get_kernel_level_pairs_for_all_feeds(unit_id or 0, pc.Stream.KERNEL)
        latest_pl_cte = q(
            '''\
            latest_patch_level as (
                SELECT json_extract(value, '$.feed') feed,
                       json_extract(value, '$.kernel_id') kernel_id,
                       json_extract(value, '$.patch_level') patch_level
                FROM json_each({})
            )
            ''',
            json.dumps(data),
        )

        latest_pl_join = (
            "LEFT JOIN latest_patch_level lpl ON lpl.kernel_id = s.kernel_id AND lpl.feed = coalesce(k.feed, 'main')"
        )

        if is_updated:
            upd_flt = 's.patch_level >= lpl.patch_level'
        else:
            upd_flt = 'NOT coalesce(s.patch_level >= lpl.patch_level, 0)'

    filters = [
        feed_flt,
        upd_flt,
        q.k.unit_id == not_none / unit_id,
        q.s.kc_key == not_none / key,
        age_filter(q, 's.created', registered_age),
        age_filter(q, 's.last', checkin_age),
        age_filter(q, 's.ltimestamp', updated_age),
        q.not_none('s.server_id = {}', server_id),
        q.not_none('s.node LIKE {}', hostname),
        q.not_none('s.remote_ip LIKE {}', ip),
        q.LIKE("';' || m.mltag || ';'", '%;{};%', not_none / tag),
    ]

    mltags_join = 'LEFT JOIN kc_servers_mltag_view m ON m.server_id = s.server_id'
    joins = ['INNER JOIN kc_keys k ON k.key = s.kc_key ', latest_pl_join, mltags_join]
    base_sql = f'''\
        {WITH(latest_pl_cte)}
        SELECT {{fields}}
        FROM kc_servers s
        {concat(*joins)}
        {WHERE(*filters)}
    '''

    cnt = q.execute(base_sql.format(fields='count(1)')).scalar()
    if only_count:
        result = []
    else:
        sql = concat(
            base_sql.format(fields="s.*, m.mltag, coalesce(k.feed, 'main') cfeed"),
            'ORDER BY s.server_id',
            q('LIMIT {} OFFSET {}', limit, offset),
        )
        result = list(parse_server_list_datetime_fields(fix_server_list_field_name(q.execute(sql, as_dict=True))))
    return cnt, result


@transaction(session)
def delete_server_by_key_and_ip(key: str, ip: str) -> int:
    cnt = session.query(KcServer).filter_by(kc_key=key, remote_ip=ip).delete()
    return cnt


@dataclass
class ServerInfo:
    unit_id: int
    key: str
    feed: str
    products: Optional[List[str]]


def iter_kc_key_fields(products_required: Optional[List[pc.Product]] = None) -> Generator[ServerInfo, None, None]:
    query = session.query(KcKey)

    products_required = products_required or []
    for product in products_required:
        if product.value in pc.DEFAULT_PRODUCTS:
            query = query.filter(
                or_(
                    KcKey.products.contains(f'"{product}"'),
                    KcKey.products.is_(None),
                )
            )
        else:
            query = query.filter(KcKey.products.contains(f'"{product}"'))

    return (
        ServerInfo(unit_id=kc_key.unit_id, key=kc_key.key, feed=kc_key.public_feed, products=kc_key.products_list)
        for kc_key in query.all()
    )


def get_kc_key_fields_by_server(server_id: Optional[str]) -> Optional[ServerInfo]:
    if not server_id:
        return None

    q = session.query_params()
    sql = f'''\
        SELECT k.unit_id, k.key, coalesce(k.feed, {q/DEFAULT_FEED}), coalesce(k.products, 'null')
        FROM kc_servers s JOIN kc_keys k ON s.kc_key = k.key
        WHERE server_id = {q/server_id}
        LIMIT 1
    '''

    result = q.execute(sql).first()
    if result:
        return ServerInfo(result[0], result[1], result[2], json.loads(result[3]))
    return None


def is_product_allowed(product: pc.Product, key_products: Optional[List[str]]) -> bool:
    products = set(key_products or [])

    # Non patch products handling
    if product == pc.Product.TUXCARE_ESU:
        return product.value in products

    # Don't allow to use patches for keys with TUXCARE_ESU, we want to keep
    # different keys for patches and proxy
    if pc.Product.TUXCARE_ESU.value in products:
        return False

    # Here we try to narrow possible products.
    # If customer specified particular patch product, i.e. KERNEL
    # we should forbid access to LIBCARE patches.
    # If there is no specific patch products we assume all allowed products (patch_products)
    patch_products = set(pc.DEFAULT_PRODUCTS)
    normalized_products = (products & set(pc.DEFAULT_PRODUCTS)) or patch_products

    return product.value in (normalized_products & patch_products)


def registered_server_ids(server_ids: List[str]) -> Set[str]:
    if not server_ids:
        return set()  # pragma: no cover
    q = session.query_params()
    sql = f'SELECT server_id FROM kc_servers WHERE {q.server_id.IN(server_ids)}'
    return set(q.execute(sql).column())
