"""
SQLAlchemy session handling utilities

The conventional session usage approach described in SQAlchemy docs is:

1) create session only for short period while working with db objects
2) for each transactional scope have a separate session
3) deal with autobegin behaviour and late actual transaction start

This approach is largely inadequate for real world codebases especially for
multi-db cases. (1) requires to watch for entrypoints where possible db work could
happen and it's hard to force on language or tooling level. (2) is inherently non-composable.
(3) is simply unacceptable for sqlite and its default lazy begin.

dbutils tries to enforce more convenient workflow:

* There is always a global session.
* By default connection is in AUTOCOMMIT mode and doesn't require needless BEGIN/COMMIT/ROLLBACK for simple
  selects and reduces SQLAlchemy overhead.
* AUTOCOMMIT connections are idle and doesn't consume resources from DB server (PostgreSQL).
* transaction decorator/context manager switches into transactional mode and allows to
  compose and abstract atomic operations. One could wrap atomic operations on lowest level
  and merge multiple atomic ops on higher level without source code refactoring.
"""

import logging
import os
import pysqlite3 as sqlite3
import threading
import time
from contextlib import contextmanager
from typing import (
    cast,
    Any,
    Callable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
    TypeVar,
    Generic,
    overload,
    Literal,
)
from typing_extensions import TypeAlias, ParamSpec

import sqlbind
from sqlalchemy import create_engine, exc
from sqlalchemy.pool.impl import SingletonThreadPool
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.orm.session import Session
from sqlalchemy.engine.base import Engine
from sqlalchemy.engine.interfaces import DBAPIConnection
from pysqlite3.dbapi2 import Connection, Cursor

from eportal import metrics, utils
import config


log = logging.getLogger('sql.trace')

PType: TypeAlias = Optional[Union[sqlbind.QueryParams, Sequence, dict]]

params_factory = sqlbind.Dialect.sqlite_named

P = ParamSpec('P')
R = TypeVar('R', covariant=True)
T = TypeVar('T')


def scope_func() -> int:
    return threading.get_ident()


def trace_sqlite(q: str) -> None:  # pragma: no cover
    ts = int(time.time() * 1000)
    tid = threading.get_ident()
    print('{0} th-{1}:'.format(ts, tid), q)


def sqlite_init(conn: Connection) -> None:
    if (os.environ.get('TRACE_SQL') or '0') != '0':  # pragma: no cover
        conn.set_trace_callback(trace_sqlite)

    # disable pysqlite's emitting of the BEGIN statement entirely.
    # also stops it from emitting COMMIT before any DDL.
    conn.isolation_level = None

    conn.execute('pragma journal_mode=wal')
    conn.execute('pragma cache_size=-100000')
    conn.execute('pragma busy_timeout=10000')
    conn.execute('pragma temp_store=2')  # temp tables in memory


def sqlite_creator(uri: str) -> Callable:
    """Need creator to force sqlalchemy to use pysqlite instead of stdlib sqlite"""
    SQLITE_PREFIX = 'sqlite:///'

    def creator() -> Connection:
        conn = sqlite3.connect(uri[len(SQLITE_PREFIX) :])
        sqlite_init(conn)
        return conn

    return creator


class ExecutableQueryParams(sqlbind.QueryParams):  # pragma: no cover
    @overload  # type: ignore[no-overload-impl]
    def execute(self, query: str, as_dict: Literal[True]) -> 'QueryList[dict]': ...

    @overload
    def execute(self, query: str, as_dict: Literal[False]) -> 'QueryList[tuple]': ...

    @overload
    def execute(self, query: str, as_dict: bool = False) -> 'Union[QueryList[tuple], QueryList[dict]]': ...


class EportalSession(scoped_session[Session]):  # pragma: no cover
    def query_params(self) -> ExecutableQueryParams:
        raise NotImplementedError


def make_engine(uri: str) -> Tuple[Engine, EportalSession]:
    engine = create_engine(
        uri,
        poolclass=SingletonThreadPool,
        pool_recycle=60,
        pool_size=10,
        pool_reset_on_return=False,
        isolation_level='AUTOCOMMIT',
        creator=sqlite_creator(uri),
    )

    def in_transaction(conn: Connection) -> bool:
        return cast(bool, conn.in_transaction)

    session_class = sessionmaker(bind=engine, autoflush=False, expire_on_commit=False)
    session: scoped_session[Session] = scoped_session(session_class, scopefunc=scope_func)

    session.IntegrityError = sqlite3.IntegrityError, exc.IntegrityError
    session.helper_in_transaction = in_transaction
    session.uri = uri
    session.write_lock_file = os.path.basename(uri) + '.app-lock'
    session.tlock = threading.RLock()
    session.local_data = threading.local()
    session.LEVEL_DEFAULT = 'SERIALIZABLE'
    session.LEVEL_AUTO = 'AUTOCOMMIT'
    session.BEGIN_TRANSACTION = 'BEGIN IMMEDIATE'

    session.query_params = lambda: SQLiteQueryParams(session)

    return engine, cast(EportalSession, session)


@contextmanager
def outer_db_lock(session: scoped_session[Session]) -> Iterator[None]:
    """Gets per sqlite db global writer lock

    We use a composition of advisory lock (utils.lock) and threading RLock
    to enforce single writer and allow writers to get lock in order.
    """
    if session.local_data.__dict__.get('in_transaction'):
        yield
    else:
        with utils.lock(os.path.join(config.BASE_DIR, session.write_lock_file), block=True):
            with session.tlock:
                session.local_data.in_transaction = True
                try:
                    yield
                finally:
                    session.local_data.in_transaction = False


def native_conn(session: scoped_session[Session]) -> Connection:
    return session.connection()._dbapi_connection.dbapi_connection  # type: ignore[union-attr]


# the following adjustments speed up SQLite
# https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#serializable-isolation-savepoints-transactional-ddl
@contextmanager
def transaction(session: scoped_session[Session]) -> Iterator[None]:
    """Manual transaction manager.

    * Handles session switch between autocommit and transactional modes.
    * Manages proper locks for sqlite.
    * Provides proper nesting to be noop if upper level already got a transaction.
    """
    # check for nested transaction and pass silently
    # TODO: it's possible to get a error here, maybe we need connection
    # invalidation at this point
    if session.helper_in_transaction(native_conn(session)):
        yield
        return

    with outer_db_lock(session):
        try:
            # close current autocommit transaction
            session.commit()

            # switch session into transactional mode and begin new transaction
            conn = session.connection(execution_options={'isolation_level': session.LEVEL_DEFAULT})
            dbapi_connection: DBAPIConnection = conn._dbapi_connection.dbapi_connection  # type: ignore[union-attr,assignment]
            cur = dbapi_connection.cursor()
        except Exception:
            metrics.db_errors += 1
            raise

        cur.execute(session.BEGIN_TRANSACTION)

        try:
            yield
            session.commit()
        except BaseException:
            session.rollback()
            raise
        finally:
            cur.close()
            session.connection(execution_options={'isolation_level': session.LEVEL_AUTO})


@contextmanager
def cursor(session: scoped_session[Session]) -> Iterator[Cursor]:
    cur = native_conn(session).cursor()
    try:
        yield cur
    finally:
        cur.close()


@contextmanager
def cursor_execute(session: scoped_session[Session], query: str, params: Any) -> Iterator[Cursor]:
    if params is None:  # pragma: no cover
        params = ()

    connection = native_conn(session)
    cur = connection.cursor()
    cur.execute(query, params)
    try:
        yield cur
    finally:
        cur.close()


class QueryList(list, Generic[T]):
    rowcount: int

    def first(self) -> Optional[T]:
        if self:
            return cast(T, self[0])
        return None

    def scalar(self) -> Any:
        v = self.first()
        if v is not None:
            return v[0]  # type: ignore[index]
        return None

    def column(self, n: int = 0) -> List[Any]:
        return [it[n] for it in self]


class ExecuteMixin:
    def __init__(self, session: scoped_session[Session]):
        self._session = session

    def execute(self, query: str, as_dict: bool = False) -> Union[QueryList[dict], QueryList[tuple]]:
        with cursor_execute(self._session, query, self) as cur:
            data = cur.fetchall()
            if as_dict:
                fields = [it[0] for it in cur.description]
                data = (dict(zip(fields, row)) for row in data)

            result = QueryList(data)  # type: ignore[var-annotated]
            result.rowcount = cur.rowcount
            return result


class SQLiteQueryParams(sqlbind.NamedQueryParams, ExecuteMixin):
    def __init__(self, session: scoped_session[Session]):
        sqlbind.NamedQueryParams.__init__(self, sqlbind.SQLiteDialect)
        ExecuteMixin.__init__(self, session)


def concat(*fragments: str) -> str:
    return sqlbind.join_fragments(' ', fragments)


def explain_sqlite3(session: scoped_session[Session], query: str, params: PType = None) -> None:  # pragma: no cover
    query = 'EXPLAIN QUERY PLAN ' + query
    with cursor_execute(session, query, params) as cur:
        for line in cur:
            print(line)
