import logging
import os
import re
import time
from typing import cast, Iterable, List, Tuple, Union, Any, Dict, Optional, Set, TYPE_CHECKING
from hashlib import sha256
from functools import partial, lru_cache
from shutil import copyfile

import config
from eportal import db, utils, audit, patches_constants as pc, patchsource as ps_module
from eportal.db.models import (
    DEFAULT_FEED,
    Feed,
    patch_level_statistics_kernel,
    patch_level_statistics_userspace,
)
from eportal.patches import api, serve

if TYPE_CHECKING:  # pragma: no cover
    from hashlib import _Hash

log = logging.getLogger(__name__)

FLAVOR_PATTERNS = {
    'amazon': '^amazon',
    'centos6': '^centos6',
    'centos7': '^centos7',
    'centos8': '^centos8',
    'cl6': '^cl6',
    'cl7': '^cl7',
    'custom': '^custom-0',
    'debian10': '^debian10',
    'debian11': '^debian11',
    'debian12': '^debian12',
    'debian8': '^debian8',
    'debian9': '^debian9',
    'oel6': '^oel6',
    'oel7': '^oel7',
    'oel8': '^oel8',
    'ubuntu14.04': '^ubuntu-trusty',
    'ubuntu16.04': '^ubuntu-xenial',
    'ubuntu18.04': '^ubuntu-bionic',
    'ubuntu20.04': '^ubuntu-focal',
    'ubuntu22.04': '^ubuntu-jammy',
    'pve': '^pve-',
}

DISTRO_FLAVORS = {distro: re.compile(flavor) for distro, flavor in FLAVOR_PATTERNS.items()}

PARENT_DISTRO_MAP = {
    (pc.Stream.USER, 'debian12'): 'pve',  # Proxmox VE 8 is based on Debian12
}


class SimultaneousUpdate(Exception):
    pass


class DeployError(Exception):
    pass


def get_checksum(filename: str, chunk_size: int = 4096) -> '_Hash':
    checksum = sha256()
    with open(filename, 'rb') as fd:
        for byte_block in iter(lambda: fd.read(chunk_size), b''):
            checksum.update(byte_block)
    return checksum


def get_feed(feed: Optional[str]) -> str:
    return feed or DEFAULT_FEED


def filter_patchsets_up_to(patchsets: Iterable[Tuple[str, str]], name: str) -> List[Tuple[str, str]]:
    result = []
    for it in patchsets:
        result.append(it)
        if it[0] == name:
            break
    else:
        return []

    return result


class PatchDeployManager:
    def __init__(
        self,
        store: serve.PatchStore,
        patchsource: Optional[ps_module.PatchSource] = None,
        is_local: Optional[bool] = None,
        forbidden_ok: bool = True,
    ):
        self.store = store
        self.patchsource = patchsource or ps_module.PatchSource()
        self.patch_ids: Dict[pc.PSFetchSpec, List[Tuple[str, str]]] = {}
        self.is_cache_mode = isinstance(store, serve.CachedPatchStore)
        self.forbidden_ok = forbidden_ok
        if is_local is None:
            self.is_local = not self.patchsource.is_configured()
        else:
            self.is_local = is_local

        if store:
            utils.ensure_dir(store.arch_dir)
            utils.ensure_dir(store.resource_dir)

            self.lock = os.path.join(store.resource_dir, 'deploy.lock')
            self.status = os.path.join(store.resource_dir, 'deploy.status')

    def clear_status(self) -> None:
        with utils.lock(self.lock, block=False) as success:
            if success:
                self.set_status('')

    def set_status(self, status: str) -> None:
        utils.atomic_write(self.status, status)

    def get_status(self) -> str:  # pragma: no cover TODO
        return (utils.read_file(self.status) or b'').decode()

    def another_update_is_running(self) -> bool:  # pragma: no cover TODO
        with utils.lock(self.lock, block=False) as success:
            return not success

    def get_patch_ids(self, spec: pc.PSFetchSpec) -> List[Tuple[str, str]]:
        try:
            return self.patch_ids[spec]
        except KeyError:
            pass

        releases = self.patchsource.get_updates(self.store.stream, spec.psproduct, spec.channel, self.forbidden_ok)[
            'releases'
        ]
        if config.LIMIT_PATCHSETS:
            releases = releases[-config.LIMIT_PATCHSETS :]

        result = self.patch_ids[spec] = [(it['release'], it['hash']) for it in releases]
        return result

    def download_patches(self, patchset: str, expected_checksum: Optional[str], force: bool = False) -> float:
        if self.is_cache_mode:
            file_name = patchset + '.content.json'
            expected_checksum = None
            check_is_json = True
        else:
            file_name = patchset + '.tar.bz2'
            check_is_json = False

        destination = os.path.join(self.store.arch_dir, file_name)

        if expected_checksum == 'skip':  # offline mode
            last_modified = os.stat(destination).st_mtime
        else:
            cb = partial(self.set_status, f'Downloading {patchset}')
            last_modified = self.patchsource.download(
                file_name,
                destination,
                download_cb=cb,
                check_is_json=check_is_json,
                force=force,
                expected_checksum=expected_checksum,
            )

        return last_modified

    def deploy_up_to(
        self,
        feed: Optional[str],
        name: str,
        deploy_after: Optional[int] = None,
        deadline: Optional[Union[float, int]] = None,
        enabled: bool = True,
        upto: bool = True,
        user: Optional[str] = None,
    ) -> None:
        feed = get_feed(feed)
        with utils.lock(self.lock, block=False) as success:
            if not success:  # pragma: no cover
                raise SimultaneousUpdate('Update is already running')

            self.set_status(f'Starting deployment {name}')
            try:
                available_patch_ids = self.get_available_patches_for_feed(feed)
                patchsets = filter_patchsets_up_to(available_patch_ids, name)
                if not patchsets:
                    raise DeployError(f'Patchset {name} not found')

                enabled_patchsets = self.store.get_releases(feed, enabled=True)
                disabled_patchsets = self.store.get_releases(feed, enabled=False)
                deployed = set(enabled_patchsets) | set(disabled_patchsets)
                updated = False
                for patchset_id, checksum in patchsets:
                    # skip disabled patchsets if we deploying whole feed
                    if upto and patchset_id in disabled_patchsets:
                        continue

                    # skip other patchsets if we deploying single release
                    if not upto and patchset_id != name:
                        continue

                    # exit from long update process
                    if deadline is not None and deadline < time.time():
                        break

                    if patchset_id in deployed:
                        if enabled and patchset_id not in enabled_patchsets:
                            self.store.enable_release(patchset_id, feed)
                            updated = True
                        elif not enabled and patchset_id not in disabled_patchsets:
                            self.store.disable_release(patchset_id, feed)
                            updated = True
                    else:
                        last_modified = self.download_patches(patchset_id, checksum)
                        if deploy_after is None or last_modified + deploy_after < time.time():
                            self.rollout_patches(feed, patchset_id, enabled=enabled)
                            updated = True

                if updated:
                    user = user or 'AUTO'
                    state = ['disabled', 'enabled'][bool(enabled)]
                    if upto:
                        # Here is a compromise. Technically we can't sort
                        # patchsets in general case due to legacy kernel
                        # patchsets names. But they are pretty rare and
                        # eventually would disappear completely.
                        # Another consideration: hidden local releases
                        # which customer can't see anymore but they would appear
                        # as previous patchsets in audit message. We should
                        # show locally available patchsets.
                        # To address both issues we need a hardcoded list
                        # of all previous legacy patchset names to be able to
                        # order patchsets. Alternative solution is to make a sort key
                        # by using quite reliable format of legacy releases:
                        # DDMMYYYY_N. Some patchsets could be outliers with
                        # MMDDYYYY_N (swapped day and month) format, but it's
                        # super rare situation.
                        previous = next(iter(sorted(enabled_patchsets, reverse=True)), 'NONE')
                        msg = f'updated feed {feed} to patchset {name} with {state} state [previous enabled patchset {previous}]'
                    else:
                        msg = f'{state} patchset {name} in feed {feed}'
                    audit.add_record(self.store.unit_id, user, audit.ACTION_MANAGE_PATCHSET, msg)

                self.clear_status()
            except Exception as e:
                self.set_status(f'Error: {e}')
                raise

    def rollback_patches(
        self, feed: Optional[str], patchset_id: str, user: Optional[str] = None, write_audit: bool = True
    ) -> None:
        feed = get_feed(feed)
        log.info('Disable %s in %s', patchset_id, feed)
        self.store.disable_release(patchset_id, feed)
        if write_audit:
            audit.add_record(
                self.store.unit_id,
                user or 'AUTO',
                audit.ACTION_MANAGE_PATCHSET,
                f'disabled patchset {patchset_id} in feed {feed}',
            )

    def rollout_patches(self, feed: Optional[str], patchset_id: str, enabled: bool = True) -> serve.PatchStatus:
        feed = get_feed(feed)
        log.info('Enable %s in %s', patchset_id, feed)
        if self.is_cache_mode:
            return self.store.deploy_resources(patchset_id, feed, enabled=enabled)
        else:
            file_name = patchset_id + '.tar.bz2'
            archive_path = os.path.join(self.store.arch_dir, file_name)
            unpack_cb = partial(self.set_status, f'Unpacking {patchset_id}')
            return self.store.deploy_archive(archive_path, feed, enabled=enabled, unpack_cb=unpack_cb)

    def undeploy_down_to(self, feed: Optional[str], name: str, user: Optional[str] = None) -> None:
        feed = get_feed(feed)
        available_patchsets = self.get_available_patches_for_feed(feed)
        patchsets = filter_patchsets_up_to(reversed(available_patchsets), name)
        names = [it for it, _ in patchsets]
        self.store.remove_releases_from_feed(names, feed)
        audit.add_record(
            self.store.unit_id,
            user or 'AUTO',
            audit.ACTION_MANAGE_PATCHSET,
            f'undeployed patchsets in feed {feed} down to patchset {name}',
        )

    def undeploy(self, feed: Optional[str], name: str, user: Optional[str] = None) -> None:
        feed = get_feed(feed)
        try:
            available_patchsets = self.get_available_patches_for_feed(feed)
            patchsets = filter_patchsets_up_to(reversed(available_patchsets), name)
            if not patchsets:
                raise DeployError(f'Patchset {name} not found')

            deployed = self.store.get_releases(feed, enabled=True)

            for patchset_id, _ in patchsets:
                if patchset_id in deployed:  # do not check patchset hash: rollback is always available
                    self.rollback_patches(feed, patchset_id, write_audit=False)

            audit.add_record(
                self.store.unit_id,
                user or 'AUTO',
                audit.ACTION_MANAGE_PATCHSET,
                f'disabled patchsets in feed {feed} down to patchset {name}',
            )

        except Exception as e:
            self.set_status(f'Error: {e}')
            raise

    def get_available_patches(self, spec: pc.PSFetchSpec) -> List[Tuple[str, str]]:
        if self.is_local:
            # TODO: return stubs when user is able to see list of kernels inside release
            return [(it, 'skip') for it in self.store.get_releases(order_by_deployed=True) if it != api.STUB_RELEASE]
        else:
            return self.get_patch_ids(spec)

    def get_available_patches_for_feed(self, feed: str) -> List[Tuple[str, str]]:
        return self.get_available_patches(Feed.fetch_spec(self.store.unit_id, feed))

    def get_patch_id_info(self, feed: Optional[str]) -> List[Tuple[str, bool]]:
        feed = get_feed(feed)
        result = []
        deployed = set(self.store.get_releases(feed, enabled=True))

        for patchset_id, _ in self.get_available_patches_for_feed(feed):
            status = patchset_id in deployed
            data = (patchset_id, status)
            result.append(data)
        return result

    def update_to_latest(
        self,
        feed: Optional[str],
        deploy_after: Optional[int] = None,
        deadline: Optional[float] = None,
        enabled: bool = True,
        user: Optional[str] = None,
    ) -> None:
        feed = get_feed(feed)
        patches = self.get_available_patches_for_feed(feed)
        if not patches:
            return
        latest, _ = patches[-1]
        self.deploy_up_to(feed, latest, deploy_after=deploy_after, deadline=deadline, enabled=enabled, user=user)

    def disable_blocked(self) -> None:
        for spec, feeds in Feed.get_feeds_by_fetch_spec(self.store.unit_id).items():
            blocked = set(
                self.patchsource.get_updates(self.store.stream, spec.psproduct, spec.channel).get('blocked', [])
            )
            if not blocked:
                continue

            for feed in feeds:
                available = set(self.store.get_releases(feed, enabled=True))
                for patchset in blocked & available:
                    log.info(f'Disabling patchset {patchset} due to block on the patchserver')
                    self.store.disable_release(patchset, feed)

    def update_feeds(self, deadline: Optional[float] = None, user: Optional[str] = None) -> Optional[bool]:
        success = True
        if not self.is_local:
            self.disable_blocked()

        for feed in db.session.query(Feed).order_by(Feed.name):
            # early exit not to fetch public list too often
            if deadline is not None and deadline < time.time():
                return None

            if feed.auto:
                try:
                    # deploy_after is stored in hours in db, but we'll use it in seconds
                    deploy_after = feed.deploy_after * 3600 if feed.deploy_after else None
                    self.update_to_latest(feed.name, deploy_after, deadline=deadline, user=user)
                except SimultaneousUpdate as e:  # pragma: no cover
                    log.warning(e)
                    return False
                except Exception as e:
                    log.exception('< %s > feed could not be updated: %s', feed.name, e)
                    success = False
                else:
                    log.info('< %s > feed is updated' % feed.name)
        return success

    def deploy_from_file(self, feed: Optional[str], filepath: str, enabled: bool = True, user: str = 'UNKNOWN') -> None:
        feed = get_feed(feed)
        filename = os.path.basename(filepath)
        checksum = get_checksum(filepath).hexdigest()
        dest = os.path.join(self.store.arch_dir, filename)
        if os.path.abspath(filepath) != os.path.abspath(dest):
            copyfile(filepath, dest)
        utils.atomic_write(dest + '.sha256', checksum)
        patch_status = self.store.deploy_archive(dest, feed, enabled=enabled)
        audit.add_record(self.store.unit_id, user, audit.ACTION_MANAGE_PATCHSET, str(patch_status))

    def dashboard(self, feeds: List[str], usage: bool = True) -> Dict[str, Any]:
        stat = []
        if usage:
            if self.store.stream in pc.USER_STREAMS:
                stat = patch_level_statistics_userspace(
                    self.store.unit_id, self.store.stream.product.value, active=True
                )
            else:
                stat = patch_level_statistics_kernel(self.store.unit_id, active=True)

        feed_specs = {it: Feed.fetch_spec(self.store.unit_id, it) for it in feeds}
        available_channel_releases = {
            spec.channel: [it for it, _ in self.get_available_patches(spec)] for spec in set(feed_specs.values())
        }

        channel_order = pc.Channel.UNSTABLE, pc.Channel.TEST, pc.Channel.DEFAULT

        available_releases = []
        for ch in channel_order:
            if ch not in available_channel_releases:
                continue
            available_releases = available_channel_releases[ch]
            break

        status = cast(Dict[Tuple[str, str], Union[int, str]], self.store.release_deploy_status())
        distros = get_distros(self.store, available_releases)

        result = {
            'usage': api.release_usage_stat(stat),
            'distros': distros,
            'status': status,
            'patchsets': available_releases,
            'unpacked': [it for it in available_releases if self.store.release_is_unpacked(it)],
        }

        # fix status for N/A patchsets for single dashboard
        arset = set(available_releases)
        for feed in feeds:
            ch = feed_specs[feed].channel
            ch_rel = set(available_channel_releases[ch])
            for ps in arset - ch_rel:
                status[(feed, ps)] = 'NA'

        return result

    def get_feed_status(self, feed: str) -> List[Tuple[str, str]]:
        available_releases = [it for it, _ in self.get_available_patches_for_feed(feed)]
        status = self.store.get_release_status(feed)
        unpacked = {it for it in available_releases if self.store.release_is_unpacked(it)}
        status_title = ['disabled', 'enabled']
        unpacked_title = ['not-downloaded', 'undeployed']
        return [
            (r, status_title[status[r]] if (r in status) else unpacked_title[r in unpacked]) for r in available_releases
        ]


@lru_cache(None)
def match_flavor(raw_flavor: str) -> str:
    for f_key, f_value in DISTRO_FLAVORS.items():
        if f_value.match(raw_flavor):
            return f_key
    return raw_flavor


def get_distros(store: serve.PatchStore, releases: List[str]) -> Dict[str, Set[str]]:
    distros: Dict[str, Set[str]] = {}
    for release in releases:
        r_flavors = store.release_flavors(release)
        matched_flavors = {match_flavor(f) for f in r_flavors}
        child_flavors = {
            PARENT_DISTRO_MAP[(store.stream, fl)] for fl in matched_flavors if (store.stream, fl) in PARENT_DISTRO_MAP
        }
        distros.setdefault(release, matched_flavors.union(child_flavors))
    return distros
