import distro
import os
import json
import subprocess
import tempfile
from functools import wraps
from runpy import run_path

import requests
from requests import Session

from requests.structures import CaseInsensitiveDict
from typing import Callable, Dict, Tuple, Union, Any, List, Optional

from eportal.web.views.main import get_dashboard_content
from eportal.patches_constants import Stream
from eportal.patches_impl import get_repo
from eportal.utils import get_instance_id, get_version
import config
from config import jroot


EXTERNAL_URLS = [
    'https://patches.kernelcare.com',
    'https://repo.cloudlinux.com/kernelcare/kernelcare_install.sh',
    'https://repo.eportal.kernelcare.com/',
    'https://downloads.kernelcare.com/',
]

EPORTAL_USER_CONFIG = '/etc/eportal/config'


def add_header(text: str) -> str:
    return f'------ {text} ---'


def format_structure(structure: Dict) -> str:
    return json.dumps(structure, ensure_ascii=False, sort_keys=True, indent=2)


def concat_list(items: List) -> str:
    return '\n'.join(items)


def prepare_data_block(header: str) -> Callable:
    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> str:
            try:
                result = func(*args, **kwargs)
            except Exception as error:
                result = repr(error)
            return concat_list([add_header(header), str(result)])

        return wrapper

    return decorator


def compact_patches_info(feeds: List[str], dashboard: Dict[str, Any]) -> Dict[str, Any]:
    status = dashboard['status']
    unpacked = dashboard['unpacked']
    patchsets = dashboard['patchsets']
    stitle = {None: 'Undeployed', True: 'Enabled', False: 'Disabled', 'NA': 'N/A'}

    ps_state: Dict[str, Dict[str, List[str]]] = {}
    for ps in patchsets:
        for f in feeds:
            key = f, ps
            if key in status:
                s = stitle.get(status[key], 'UNKNOWN')
            else:
                s = 'Undeployed' if ps in unpacked else 'Not downloaded'
            ps_state.setdefault(ps, {}).setdefault(s, []).append(f)

    return ps_state


@prepare_data_block('patches info')
def get_patches_info() -> str:
    patches_info = {}
    dc = get_dashboard_content(0, None)
    for stream in Stream:
        repo = get_repo(0, stream)
        stat = repo.dashboard(dc.feeds, usage=False)
        patches_info[stream.value] = compact_patches_info(dc.feeds, stat)
    return format_structure(patches_info)


def get_proxy_settings() -> Dict[str, str]:
    def get_proto_proxy(proto: str) -> Optional[str]:
        return os.environ.get(proto.upper() + '_PROXY', os.environ.get(proto + '_proxy', config.PROXY))

    return {k: v for k, v in [(proto, get_proto_proxy(proto)) for proto in ('http', 'https')] if v}


def get_resource(url: str) -> Tuple[int, CaseInsensitiveDict, bytes]:
    s = Session()
    s.proxies = get_proxy_settings()
    r = s.get(url, timeout=10)
    return r.status_code, r.headers, r.content


def check_proxy() -> bool:
    for envvar in list(os.environ):
        if envvar.lower() in ('http_proxy', 'https_proxy'):
            return True
    return False


@prepare_data_block('connectivity')
def check_connectivity() -> str:
    result: Dict[str, Union[Dict[str, Any], bool]] = {}
    for url in EXTERNAL_URLS:
        result[url] = {}
        try:
            status, headers, _ = get_resource(url)
            result[url]['status'] = status  # type: ignore
            result[url]['headers'] = headers  # type: ignore
        except Exception as error:
            result[url]['error'] = str(error)  # type: ignore
    result['proxy'] = check_proxy()
    return format_structure(result)


def run_command(command: str) -> Tuple[str, str]:
    result = subprocess.run(command, shell=True, capture_output=True)  # nosec: B602
    out = result.stdout.decode('utf8', errors='replace')
    err = result.stderr.decode('utf8', errors='replace')
    return out, err


def collect_logs() -> str:
    nginx_acces_log_content, _ = run_command('tail -n 1000 /var/log/nginx/access.log')
    nginx_eportal_log_content, _ = run_command('tail -n 1000 /var/log/nginx/kcare-eportal.log')
    nginx_error_log_content, _ = run_command('tail -n 1000 /var/log/nginx/error.log')
    eportal_log_content, _ = run_command('journalctl -u eportal -ocat -n 10000')
    logs = [
        add_header('/var/log/nginx/access.log'),
        nginx_acces_log_content,
        add_header('/var/log/nginx/kcare-eportal.log'),
        nginx_eportal_log_content,
        add_header('/var/log/nginx/error.log'),
        nginx_error_log_content,
        add_header('journalctl eportal log'),
        eportal_log_content,
    ]
    return concat_list(logs)


@prepare_data_block('disk usage')
def get_disk_usage() -> str:
    output, _ = run_command('df -h')
    return output


@prepare_data_block('OS version')
def get_os_version() -> str:
    distro_id = distro.id()
    distro_version = distro.version()
    distro_name = distro.name()
    return concat_list([f'ID: {distro_id}', f'Version: {distro_version}', f'Name: {distro_name}'])


@prepare_data_block('process list')
def get_process_list() -> str:
    result, _ = run_command('ps aux | grep -E "(nginx|eportal)"')
    return result


@prepare_data_block('audit log events')
def get_audit_log_events() -> str:
    user = os.getuid()
    if user == 0:
        result, _ = run_command('tail -10000 /var/log/audit/audit.log')
        return result
    else:
        print('SELinux audit log will not be attached to the report. You are not root.')
        return 'No access to audit log. User is not root.'


SECRETS_BLACKLIST = ['KEY', 'SECRET', 'DATABASE', 'PROXY', 'PASSWORD']


def prepare_config(path: str) -> dict:
    config = run_path(path)
    for key, value in config.copy().items():
        if not key.isupper():
            del config[key]
        if any(it in key for it in SECRETS_BLACKLIST):
            config[key] = '********'
    return config


@prepare_data_block('config options')
def get_config_options() -> str:
    paths = [jroot('config', 'local.py'), jroot('config', 'docker.py'), EPORTAL_USER_CONFIG]
    config: Dict[str, Any] = {}
    for path in paths:
        if os.path.exists(path):
            conf = prepare_config(path)
        else:
            conf = {'msg': 'No user defined config options.'}
        config[path] = conf
    return format_structure(config)


@prepare_data_block('Main IP')
def get_eportal_ip() -> str:
    _, _, content = get_resource('https://patches.kernelcare.com/myip')
    return content.decode('utf-8')


def make_doctor_report() -> str:
    version, platform = get_version()
    report = [
        get_eportal_ip(),
        get_os_version(),
        concat_list([add_header('eportal package version'), version]),
        concat_list([add_header('platform'), platform]),
        get_patches_info(),
        check_connectivity(),
        get_disk_usage(),
        get_process_list(),
        get_audit_log_events(),
        get_config_options(),
        collect_logs(),
    ]
    return concat_list(report)


def save_report(report_content: str) -> str:
    with tempfile.NamedTemporaryFile(mode='w', delete=False, prefix='eportal-doctor-') as tmp_file:
        tmp_file.write(report_content)
        filename = tmp_file.name
    return filename


def send_report(report_content: str, save_only: bool = False) -> None:
    headers = {'serverid': get_instance_id()}
    files = {'reportfile': ('cl-report', report_content)}
    if save_only:
        filename = save_report(report_content)
        print(f"Please provide this file to support:\n{filename}")
    else:
        try:
            response = requests.post(
                "https://patches.kernelcare.com/doctor/upload",
                headers=headers,
                files=files,
                timeout=30,
                proxies=get_proxy_settings(),
            )
            response.raise_for_status()
            print(f"Please provide this key to support:\n{response.text}")
        except Exception as error:
            print(f"Error while sending report:\n{error}\n\n")
            filename = save_report(report_content)
            print(f"Please provide this file to support:\n{filename}")
