import os
import json

from functools import cached_property
from typing import Dict, Any, List, Optional, cast
from typing_extensions import TypedDict

import config
from eportal import utils, oplog


class PatchRepoSettings(TypedDict):
    enabled: bool
    login: str
    password: str
    url: str
    distros: List[str]


class LDAPSettings(TypedDict):
    url: str
    bind_template: str
    filter_template: str


class Config(object):
    def __init__(self) -> None:
        self.config_file = os.path.join(config.BASE_DIR, 'config', 'config.json')

    @cached_property
    def data(self) -> Dict[str, Any]:
        return self.read()

    def read(self) -> Dict[str, Any]:
        content = utils.read_file(self.config_file)
        if content:
            return cast(Dict[str, Any], json.loads(content))
        else:
            return {'version': 1}

    def configure_patchrepo(self, login: str, password: str, url: str) -> None:
        self.update('patchrepo', enabled=True, login=login.strip(), password=password.strip(), url=url.strip())

    def reset_patchrepo(self) -> None:
        self.update('patchrepo', enabled=False, login='', password='', url='')  # nosec hardcoded_password_funcarg

    def configure_enabled_distros(self, distros: Optional[List[str]]) -> None:
        self.update('patchrepo', distros=distros or [])

    @cached_property
    def patchrepo(self) -> PatchRepoSettings:
        cfg = self.data.get('patchrepo') or {}
        if config.PATCH_SOURCE_USER:
            return {
                'url': config.PATCH_SOURCE_URL,
                'login': config.PATCH_SOURCE_USER,
                'password': config.PATCH_SOURCE_PASSWORD,
                'enabled': True,
                'distros': cfg.get('distros') or [],
            }

        return {
            'url': cfg.get('url') or '',
            'login': cfg.get('login') or '',
            'password': cfg.get('password') or '',
            'enabled': cfg.get('enabled', False),
            'distros': cfg.get('distros') or [],
        }

    @cached_property
    def ldap(self) -> LDAPSettings:
        cfg = self.data.get('ldap', {})
        return {
            'url': cfg.get('url') or '',
            'bind_template': cfg.get('connection_string') or '',
            'filter_template': cfg.get('path_part') or '',
        }

    def configure_ldap(self, url: str, bind_template: str, filter_template: Optional[str]) -> None:
        self.update('ldap', url=url, connection_string=bind_template, path_part=filter_template or '')

    def reset_ldap(self) -> None:
        self.update('ldap', url='', connection_string='', path_part='')

    def update(self, section: str, **kwargs: Any) -> None:
        with utils.lock(self.config_file, block=True):
            self.data = self.read()
            self.data.setdefault(section, {}).update(kwargs)
            try:
                delattr(self, section)
            except AttributeError:
                pass
            save_config_data(data=self.data)


@oplog.replicate('manage-sync-config')
def save_config_data(data: Dict[str, Any]) -> None:
    config_dir = os.path.join(config.BASE_DIR, 'config')
    config_file = os.path.join(config_dir, 'config.json')
    utils.ensure_dir(config_dir)
    utils.atomic_write(config_file, json.dumps(data))
