import logging
from typing import Optional, Tuple, Union
from collections import namedtuple
from urllib.parse import urlsplit, parse_qs, urlunsplit

import ldap

logger = logging.getLogger(__name__)

ldap_scopes = {'': ldap.SCOPE_BASE, 'base': ldap.SCOPE_BASE, 'one': ldap.SCOPE_ONELEVEL, 'sub': ldap.SCOPE_SUBTREE}
ldap_path = namedtuple('ldap_path', ['base', 'attributes', 'scope', 'filter', 'extensions'])
ldap_settings = namedtuple('ldap_settings', ['url', 'strict_check', 'timeout', 'tls'])


class LDAPSearchFailed(Exception):
    pass


def parse_path(ldap_path_string: str) -> ldap_path:
    max_attr_num = 5
    parts = ldap_path_string.split('?')[:max_attr_num]
    parts.extend([''] * (max_attr_num - len(parts)))  # extend with default values
    return ldap_path(*parts)


def get_scope(user_input: str) -> Tuple[int, str]:
    int_scope = ldap_scopes[user_input]
    # if scope is omitted, a scope of "base" is assumed
    # (see RFC 2255 "The LDAP URL Format", section "3. URL Definition")
    str_scope = user_input or 'base'
    return int_scope, str_scope


def parse_url(url: str, strict_check: int = 1, timeout: int = 5, tls: int = 1) -> ldap_settings:
    parsed = urlsplit(url)
    qs = parse_qs(parsed.query)

    url = urlunsplit([parsed.scheme, parsed.netloc, '', '', ''])
    tls = int(qs.get('tls', [str(tls)])[0])
    timeout = int(qs.get('timeout', [str(timeout)])[0])
    strict_check = int(qs.get('strict_check', [str(strict_check)])[0])

    return ldap_settings(url, strict_check, timeout, tls)


def verify(url: str, connection_string: str, path: str) -> None:
    parsed = urlsplit(url)

    url = urlunsplit([parsed.scheme, parsed.netloc, '', '', ''])

    if not url:
        raise ValueError('Incorrect LDAP URL provided, please fix it in accordance with the requirements.')
    if not url.startswith('ldap://') and not url.startswith('ldaps://'):
        raise ValueError('Incorrect LDAP URL scheme provided: use `ldap://` or `ldaps://`.')

    try:
        connection_string % 'check'
    except TypeError:
        message = (
            'Invalid connection string provided (no login placeholder), '
            'please fix it in accordance with the requirements.'
        )
        raise ValueError(message)

    if path:
        if parse_path(path).scope not in ldap_scopes.keys():
            message = (
                'Incorrect value for "scope" LDAP attribute. Must be one of: "base", "one", "sub". '
                'Please, see RFC 2255 "The LDAP URL Format"'
            )
            raise ValueError(message)


def login(url: str, path: Optional[str], connection_string: str, password: str) -> bool:
    settings = parse_url(url)
    logger.info('LDAP bind: %s with %s', settings, connection_string)

    conn = ldap.initialize(settings.url)
    conn.set_option(ldap.OPT_NETWORK_TIMEOUT, settings.timeout)
    conn.set_option(ldap.OPT_REFERRALS, 0)

    if not settings.strict_check:
        conn.set_option(ldap.OPT_X_TLS_REQUIRE_CERT, ldap.OPT_X_TLS_NEVER)
        # TLS options require a new TLS context, 0 denotes client context
        conn.set_option(ldap.OPT_X_TLS_NEWCTX, 0)

    if not url.startswith('ldaps://') and settings.tls:
        conn.start_tls_s()

    try:
        conn.simple_bind_s(connection_string, password)
    except ldap.INVALID_CREDENTIALS:
        return False

    if path:
        path_parts = parse_path(path)
        int_scope, str_scope = get_scope(path_parts.scope)
        logger.info(
            'LDAP search parameters: url={0}, bind={1}, dn={2}, scope={3}, filter={4}'.format(
                settings.url, connection_string, path_parts.base, str_scope, path_parts.filter
            )
        )
        search_result = conn.search_ext_s(path_parts.base, int_scope, path_parts.filter, sizelimit=10)

        # filter empty fields that could be returned by LDAP server in case of no valid results
        # bad case: [(None, ['ldap://ex.com/...']), (...)]
        # good case:  [('CN=Name Name,DC=ex,DC=com', {'some': 'data'}), (...)]
        filtered_search_result = list(filter(is_valid, search_result))

        if search_result:
            logger.info('LDAP search results: {0}'.format(search_result))
        if not filtered_search_result:
            raise LDAPSearchFailed('User with provided credentials is valid, but has insufficient rights, sorry.')

    return True


def is_valid(entry: Tuple[Optional[str], Union[list, dict]]) -> bool:
    return entry and isinstance(entry, (tuple, list)) and entry[0] is not None
