from collections import defaultdict
import threading
from typing import DefaultDict, cast

dropped_checkins = 0
db_errors = 0

REFRESH_PERIOD = 3


def get_uwsgi_worker_id() -> int:
    try:
        # uwsgi module exists only if running as uwsgi app
        import uwsgi

        return cast(int, uwsgi.worker_id())
    except ImportError:
        # According to the uWSGI docs 0 means not a worker
        return 0


class _Count2Increment:
    def __init__(self) -> None:
        self._history: DefaultDict[str, int] = defaultdict(int)

    def __call__(self, name: str, count: int) -> int:
        count_previous = self._history[name]
        self._history[name] = count
        # always >= 0
        return count - min(count, count_previous)

    def reset(self) -> None:
        self._history.clear()


COUNT_2_INCREMENT = _Count2Increment()


def statsd_pusher(stop: threading.Event) -> None:
    import logging
    import time
    from eportal import checkins
    from statsd.defaults.env import statsd
    import resource

    def statsd_incr_by_count(stat_name: str, count: int) -> None:
        """Wrapper to send counter's delta (increment)"""
        statsd.incr(stat_name, COUNT_2_INCREMENT(stat_name, count))

    wid = get_uwsgi_worker_id()

    def loop(stop: threading.Event) -> None:
        while not stop.is_set():
            try:
                s = resource.getrusage(resource.RUSAGE_SELF)
                statsd.gauge(f'eportal.maxrss,worker-{wid}', s.ru_maxrss)
                statsd_incr_by_count('eportal.block.read', s.ru_inblock)
                statsd_incr_by_count('eportal.block.write', s.ru_oublock)
                statsd_incr_by_count('eportal.dropped_checkins', dropped_checkins)
                statsd_incr_by_count('eportal.db_errors', db_errors)
                statsd.gauge(f'eportal.checkin_queue,worker-{wid}', len(checkins.process_kernel_stat.queue))
            except Exception:
                logging.getLogger(__name__).exception('failed to update statsd counters')
            finally:
                time.sleep(REFRESH_PERIOD)

    threading.Thread(target=loop, daemon=True, args=(stop,)).start()
