import time
import pysqlite3 as sqlite3
import ssl
import logging
import json
import threading
from typing import cast, Any, Callable, Optional, Union, Dict, List, TypeVar, Generic
from functools import lru_cache, wraps
from urllib.parse import urlsplit
from urllib.request import urlopen, Request, HTTPError
from http.client import HTTPResponse

from pysqlite3.dbapi2 import Connection
from sqlbind import WHERE
from typing_extensions import Protocol, TypedDict, ParamSpec
from statsd.defaults.env import statsd

import config
from eportal.dbutils import params_factory
from eportal import utils
from eportal.utils import ContinuousScheduler

log = logging.getLogger('oplog')

PAST_CHANGES = 300  # in case of empty last id get changes for last 5min
CLEAN_POLICY_NORMAL = 'normal'
CLEAN_POLICY_AGGRESSIVE = 'aggressive'

# replication proto version, should be increased on incompatible changes
VERSION = 2

P = ParamSpec('P')
R = TypeVar('R', covariant=True)


class FullChanges(TypedDict, total=False):
    units: List[dict]
    unit_users: List[dict]
    users: List[dict]
    feeds: List[dict]
    keys: List[dict]
    patches: Dict[str, list]
    config: Dict[str, Any]


class ActionCallable(Generic[P, R], Protocol):
    action_name: str
    clean_policy: str

    @property
    def write(self) -> Callable[P, R]:  # pragma: no cover
        ...

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:  # pragma: no cover
        ...


class ReplicatedCallable(Generic[P, R], Protocol):
    @property
    def action(self) -> ActionCallable[P, R]:  # pragma: no cover
        ...

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:  # pragma: no cover
        ...


_actions: Dict[str, ActionCallable] = {}


def get_real_connection() -> Connection:
    conn = sqlite3.connect(config.OPLOG_DATABASE)
    conn.isolation_level = None
    conn.execute('pragma journal_mode=wal')
    conn.execute('pragma busy_timeout=10000')
    return conn


@lru_cache(5)
def get_cached_connection(_cache_key: int) -> Connection:
    return get_real_connection()


@lru_cache(None)
def node_from_url(url: str) -> str:
    host = urlsplit(url).hostname
    if not host:
        raise ValueError(f'Invalid url: {url}')
    return host


def get_connection() -> Connection:
    return get_cached_connection(threading.get_ident())


def table_columns(conn: Connection, table: str) -> List[str]:
    cur = conn.execute(f"pragma table_info('{table}')")
    result = [it[1] for it in cur]
    cur.close()
    return result


def create_table() -> None:
    with get_real_connection() as conn:
        conn.execute(
            '''
            CREATE TABLE IF NOT EXISTS oplog (
                id INTEGER PRIMARY KEY,
                rid TEXT,
                ts REAL,
                node TEXT,
                action TEXT,
                params TEXT
            )
        '''
        )
        conn.execute('CREATE INDEX IF NOT EXISTS idx_oplog_ts ON oplog (ts)')
        conn.execute('CREATE INDEX IF NOT EXISTS idx_oplog_rid ON oplog (rid)')

        conn.execute(
            '''
            CREATE TABLE IF NOT EXISTS cursors (
                node TEXT PRIMARY KEY,
                instance_id TEXT,
                ts REAL,
                last_id INTEGER
            )
        '''
        )

        conn.execute(
            '''
            CREATE TABLE IF NOT EXISTS followers (
                node TEXT PRIMARY KEY,
                url TEXT,
                ts REAL
            )
        '''
        )

        if 'instance_id' not in table_columns(conn, 'cursors'):
            conn.execute('ALTER TABLE cursors ADD COLUMN instance_id TEXT')

        if 'change_ts' not in table_columns(conn, 'cursors'):
            conn.execute('ALTER TABLE cursors ADD COLUMN change_ts REAL')

    conn.close()


def drop_table() -> None:
    with get_real_connection() as conn:
        conn.execute('DROP TABLE IF EXISTS oplog')
        conn.execute('DROP TABLE IF EXISTS followers')
        conn.execute('DROP TABLE IF EXISTS cursors')
    conn.close()


unverified_context = ssl.create_default_context()
unverified_context.check_hostname = False
unverified_context.verify_mode = ssl.CERT_NONE


def oplog_urlopen(url: str) -> HTTPResponse:  # pragma: no cover TODO
    headers = utils.basic_auth_headers('replication', config.REPLICATION_SHARED_KEY)
    # headers['Accept-Encoding'] = 'deflate'
    if url.startswith("file:"):
        raise ValueError(f"Cannot urlopen file: {url}")

    req = Request(url, headers=headers)
    context = None
    if not config.REPLICATION_CHECK_CERT:
        context = unverified_context
    # req here cannot be a request to a file based on the check above, ignore Bandit warning.
    return cast(HTTPResponse, urlopen(req, timeout=10, context=context))  # nosec B310


def write(
    action: str,
    params: Dict[str, Any],
    node: Optional[str] = None,
    rid: Optional[str] = None,
    now: Optional[float] = None,
) -> None:
    node = node or (config.NODE_URL and node_from_url(config.NODE_URL))
    if not node:
        return
    rid = rid or utils.random_id()
    now = now or time.time()
    params_json = json.dumps(params, ensure_ascii=False)
    q = 'INSERT INTO oplog(rid, ts, node, action, params) VALUES (?, ?, ?, ?, ?)'
    conn = get_connection()
    node = '|' + node.strip('|') + '|'
    conn.execute(q, (rid, now, node, action, params_json))

    # just in case, force transaction closing
    if conn.in_transaction:  # pragma: no cover
        conn.commit()


def read(
    last_id: Optional[int], from_ts: Optional[float], node: str, action_prefix: Optional[str] = None
) -> Dict[str, Any]:
    conn = get_connection()
    from_ts = from_ts or time.time() - PAST_CHANGES

    if last_id is not None:
        q_params = params_factory()
        filters = [
            q_params.id > last_id,
            q_params.cond(action_prefix, f"action like '{q_params/(action_prefix or '')}%'"),
            q_params('node not like {}', f'%|{node}|%'),
        ]
        q = f'SELECT id, rid, ts, node, action, params FROM oplog {WHERE(*filters)} ORDER BY id LIMIT 1000'
        cur = conn.execute(q, q_params)
        changes = [
            {'id': opid, 'rid': rid, 'ts': ts, 'node': node, 'action': action, 'params': json.loads(params)}
            for opid, rid, ts, node, action, params in cur.fetchall()
        ]
        if changes:
            new_last_id = changes[-1]['id']
        else:
            # here we need to track last_id. If it is greater than max_id or less
            # than min_id then we need to reset it to min_id
            max_id = conn.execute('SELECT max(id) FROM oplog').fetchone()[0]
            min_id = conn.execute('SELECT max(id) FROM oplog WHERE ts <= ?', (from_ts,)).fetchone()[0] or 0
            if last_id < min_id:
                new_last_id = min_id
            elif max_id and last_id > max_id:
                new_last_id = min_id
            else:
                new_last_id = last_id
    else:
        changes = []
        new_last_id = conn.execute('SELECT max(id) FROM oplog WHERE ts <= ?', (from_ts,)).fetchone()[0] or 0

    return {'changes': changes, 'last_id': new_last_id}


def replicate(name: str, clean_policy: Optional[str] = None) -> Callable[[Callable[P, R]], ReplicatedCallable[P, R]]:
    def decorator(fn: Callable[P, R]) -> ReplicatedCallable[P, R]:
        @wraps(fn)
        def inner(**kwargs: P.kwargs) -> R:
            do_replicate = kwargs.pop('replicate', True)
            now = kwargs.pop('replicate_now', None)
            result = fn(**kwargs)  # type: ignore[call-arg]
            if do_replicate:
                write(name, kwargs, now=now)
            return result

        inner_aug = cast(ReplicatedCallable[P, R], inner)
        inner_aug.action = register_action(name, clean_policy)(fn)  # type: ignore[misc]
        return inner_aug

    return decorator


def register_action(name: str, clean_policy: Optional[str] = None) -> Callable:
    def decorator(fn: Callable) -> ActionCallable:
        fn = cast(ActionCallable, fn)
        _actions[name] = fn
        fn.action_name = name
        fn.clean_policy = clean_policy or CLEAN_POLICY_NORMAL
        fn.write = lambda **kwargs: write(name, kwargs)  # type: ignore[misc]
        return fn

    return decorator


def set_last_id(
    leader: str,
    last_id: Optional[int],
    instance_id: Optional[str],
    change_ts: Optional[float] = None,
    now: Optional[float] = None,
) -> None:
    now = now or time.time()
    change_ts = change_ts or now
    get_connection().execute(
        'REPLACE INTO cursors (node, ts, last_id, instance_id, change_ts) VALUES (?, ?, ?, ?, ?)',
        (leader, now, last_id, instance_id, change_ts),
    )


def get_last_id(leader: str) -> Optional[int]:
    result = get_connection().execute('SELECT last_id FROM cursors WHERE node = ?', (leader,)).fetchone()
    return result and cast(int, result[0])


def get_node_instance_id(leader: str) -> Optional[str]:
    result = get_connection().execute('SELECT instance_id FROM cursors WHERE node = ?', (leader,)).fetchone()
    return result and cast(str, result[0])


def update_follower(node: str, url: str, now: Optional[float] = None) -> None:
    ts = now or time.time()
    get_connection().execute('REPLACE INTO followers (node, url, ts) VALUES (?, ?, ?)', (node, url, ts))


def get_followers(updated_since: float) -> List[str]:
    result = get_connection().execute('SELECT url FROM followers WHERE ts > ?', (updated_since,)).fetchall()
    return [it for it, in result]


def join_node(nodes: str, node: str) -> str:
    node = f'|{node}|'
    if node not in nodes:
        return nodes.rstrip('|') + node
    return nodes


def is_op_applied(rid: str, conn: Optional[Connection] = None) -> bool:
    q = 'SELECT rid FROM oplog WHERE rid = ? LIMIT 1'
    result = (conn or get_connection()).execute(q, (rid,)).fetchone()
    return bool(result)


def apply_remote_changes(leader: str, changes: Dict[str, Any]) -> int:
    conn = get_connection()
    node = node_from_url(config.NODE_URL)
    cnt = 0
    for op in changes['changes']:
        if is_op_applied(op['rid'], conn):
            continue
        cnt += 1
        action = op['action']
        params = op['params']
        try:
            _actions[action](**params)
        except Exception as ex:
            if config.PUSH_METRICS:
                statsd.incr('eportal.replication.errors')
            log.exception(ex)
        else:
            write(action, params, join_node(op['node'], node), op['rid'], now=op['ts'])

    if cnt:
        change_ts = changes['changes'][-1]['ts']
    else:
        change_ts = None

    set_last_id(leader, changes['last_id'], changes.get('instance_id'), change_ts=change_ts)
    return cnt


def clean(upto_ts: float, clean_policy: Optional[str] = None, vacuum: bool = False) -> None:
    q_params = params_factory()
    filters = [q_params.ts < upto_ts]
    if clean_policy:
        actions = [k for k, v in _actions.items() if v.clean_policy == clean_policy]
        if not actions:
            return
        filters.append(q_params.action.IN(actions))

    q = f'DELETE FROM oplog {WHERE(*filters)}'
    conn = get_connection()
    conn.execute(q, q_params)

    if vacuum:
        conn.execute('VACUUM')


@utils.error_rate_limit(60)
def get_remote_changes(remote_url: str, node_url: str, action_prefix: Optional[str] = None) -> Any:
    from eportal.web import app

    leader = node_from_url(remote_url)
    params = {
        'node_url': node_url,
        'last_id': get_last_id(leader),
        'action_prefix': action_prefix,
        'version': VERSION,
    }

    if config.REPLICATION_STANDBY:
        params['active'] = 'false'

    url = remote_url.rstrip('/') + app.url_map.bind('fake').build('oplog_changes', params)

    response = oplog_urlopen(url)
    raw = response.read()
    result = json.loads(raw)
    return result


def leader_url_is_self(leader_url: str) -> bool:
    return leader_url.rstrip('/') == (config.NODE_URL or '').rstrip('/')


def get_full_changes(remote_url: str) -> FullChanges:  # pragma: no cover TODO
    from eportal.web import app

    url = remote_url.rstrip('/') + app.url_map.bind('fake').build('full_changes', {'version': VERSION})
    response = oplog_urlopen(url)
    raw = response.read()
    result = json.loads(raw)
    return cast(FullChanges, result)


def get_server_registrations(remote_url: str) -> Any:  # pragma: no cover TODO
    from eportal.web import app

    url = remote_url.rstrip('/') + app.url_map.bind('fake').build('server_registrations', {'version': VERSION})
    response = oplog_urlopen(url)
    raw = response.read()
    result = json.loads(raw)
    return result


def relative_time(now: Union[float, int], ts: Optional[Union[int, float]]) -> str:
    if not ts:
        return 'never'
    return '{}s ago'.format(int(now - ts))


def relative_time_num(now: float, ts: Optional[float]) -> int:
    if not ts:
        return 99999
    return int(now - ts)


def status(now: Optional[Union[float, int]] = None) -> Dict[str, Any]:
    conn = get_connection()
    now = now or time.time()
    ago = now - 86400 * 7  # week ago

    q = 'SELECT node, url, ts FROM followers WHERE ts > ?'
    followers = {
        node: {'url': url, 'last_sync': ts, 'msg': f'last sync {int(now - ts)}s ago'}
        for node, url, ts in conn.execute(q, (ago,))
    }

    q = 'SELECT node, ts, last_id, instance_id, change_ts FROM cursors WHERE ts > ?'
    leaders = {
        node: {
            'last_sync': ts,
            'change_ts': change_ts,
            'msg': f'last sync {relative_time(now, ts)}, lag {relative_time(now, change_ts)}',
            'last_id': last_id,
            'instance_id': iid,
        }
        for node, ts, last_id, iid, change_ts in conn.execute(q, (ago,))
    }

    last_id = conn.execute('SELECT max(id) FROM oplog').fetchone()[0]

    return {
        'role': 'FOLLOWER' if config.LEADER_URL else 'LEADER',
        'node': config.NODE_URL,
        'current_leader': config.LEADER_URL,
        'oplog_size': conn.execute('SELECT count(1) FROM oplog').fetchone()[0],
        'oplog_last_id': last_id,
        'followers': followers,
        'leaders': leaders,
        'instance_id': utils.get_instance_id(),
    }


def short_status(now: Optional[float] = None, ok_threshold: int = 60) -> Dict[str, Any]:
    now = now or time.time()
    data = status(now)
    if not data['leaders']:
        return {
            'ok': False,
            'last_sync': 99999,
            'lag': 99999,
        }

    upstream = {}
    for k, v in data['leaders'].items():
        last_sync = relative_time_num(now, v['last_sync'])
        replication_lag = relative_time_num(now, v['change_ts'])
        is_ok = last_sync < ok_threshold and replication_lag < ok_threshold
        upstream[k] = {'ok': is_ok, 'last_sync': last_sync, 'lag': replication_lag}

    return {
        'ok': all(it['ok'] for it in upstream.values()),
        'last_sync': max(it['last_sync'] for it in upstream.values()),
        'lag': max(it['lag'] for it in upstream.values()),
        'upstream': upstream,
    }


def forget(node: str) -> None:
    q = params_factory()
    filters = [q.node == node]
    with get_connection() as conn:
        conn.execute(f'DELETE FROM followers {WHERE(*filters)}', q)
        conn.execute(f'DELETE FROM cursors {WHERE(*filters)}', q)


def sync_with_node(url: str) -> Optional[bool]:
    node = node_from_url(url)
    try:
        changes = get_remote_changes(url, config.NODE_URL)
    except HTTPError as e:
        log.error('Unable fetch remote changes from %s, status: %s, response: %s', e.url, e.code, e.read())
        return None
    except Exception:
        log.exception('Error while fetching changes from %s:', url)
        return None

    if changes is None:  # rate limited exception occured
        return None

    count = apply_remote_changes(node, changes)

    if count:
        log.info(f'Applied {count} ops from {url}')
        return True

    return None  # pragma: no cover


def push_metrics() -> None:
    status = short_status()
    upstream_stats = list(status.get('upstream', {}).items())
    upstream_stats.append(('max', status))
    with statsd.pipeline() as sclient:
        for node, node_status in upstream_stats:
            sclient.gauge(f'eportal.replication.lag,upstream={node}', node_status['lag'])
            sclient.gauge(f'eportal.replication.last_sync,upstream={node}', node_status['last_sync'])


def make_scheduler() -> ContinuousScheduler:
    s = utils.ContinuousScheduler(log)
    scheduled_followers: Dict[str, object] = {}

    def sync_followers() -> None:
        followers = get_followers(time.time() - 60)

        # remove inactive followers from queue
        for node, tid in list(scheduled_followers.items()):
            if node not in followers:
                s.cancel(tid)
                scheduled_followers.pop(node)

        # add new followers to queue
        for node in followers:
            if node not in scheduled_followers:
                scheduled_followers[node] = s.add(3, sync_with_node, node)

    if config.NODE_URL and config.LEADER_URL:
        s.add(3, sync_with_node, config.LEADER_URL)

    if config.NODE_URL:
        s.add(5, sync_followers)

    if config.PUSH_METRICS:
        s.add(30, push_metrics)

    return s


def _check_mypy() -> None:  # pragma: no cover
    @replicate('boo')
    def boo(some: str) -> str:
        return some

    # boo triggers error intentionally, it allows to catch unused-ignore if
    # replicated decorator typehints would be broken
    boo(some=10)  # type: ignore[arg-type]
