import logging
from datetime import datetime, timedelta
from itertools import groupby
from typing import Any, Dict, List, Optional, Set, Tuple, Union, Iterable
from typing_extensions import TypedDict

from sqlbind import WHERE, not_none

from eportal import patches_constants as pc
from eportal.db import models as me
from eportal.patches import models as mp, session
from eportal.dbutils import transaction

import config

log = logging.getLogger('wetches')

STATUS_ENABLED = 'enabled'
STATUS_PARTLY = 'partly'
STATUS_DISABLED = 'disabled'
STUB_RELEASE = 'stubs'

EMPTY = '__empty__'


@transaction(session)
def insert_release(
    unit_id: int, stream: pc.Stream, release: str, feed: str, enabled: bool, deploy_date: datetime
) -> bool:
    already_deployed = (
        session.query(mp.Release).filter_by(unit_id=unit_id, stream=stream.value, release=release, feed=feed).first()
    )
    if already_deployed:
        switch(unit_id, stream, feed, enabled, release=release)
        return False
    else:
        r = mp.Release(
            unit_id=unit_id, stream=stream.value, feed=feed, release=release, deployed=deploy_date, enabled=enabled
        )
        session.add(r)
        return True


@transaction(session)
def insert_release_info(release: str, info: Dict[str, Dict[str, Any]]) -> None:
    """
    Update `patches` and `fixup` table with info about a new release (resources description).
    :param release: release name
    :param info: dict with release content: {'kernel_id': {'patch_level_v1': 8, 'patch_level_v2': 8, 'fixups': <...>}}
    :return: None
    """
    already_deployed = session.query(mp.Patch).filter_by(release=release).first()
    if already_deployed:
        return

    for kernel, v in info.items():
        patch_level_v1 = v['patch_level_v1']
        patch_level_v2 = v['patch_level_v2']

        # old latest versions are available to freshest clients, e.g.
        # latest.v1 is available to client with latest.v2 support;
        # if patch contains only latest.v1, it should be available to all KC clients (with/without latest.v2 support);
        # if release contains both latest.v1 and latest.v2 -- the same;
        # this patch could be served in response to latest requests with version 1+
        if patch_level_v1 is not None:
            version = 1
            level = patch_level_v1

        # if patch contains only latest.v2, it's available only to KC clients
        # with latest.v2 support (and fresher versions), and could not be served in response to latest.v1 requests;
        # this patch could be served in response to latest requests with version 2+
        elif patch_level_v2 is not None:
            version = 2
            level = patch_level_v2

        info[kernel]['version'] = version
        info[kernel]['level'] = level

    for kernel, v in info.items():
        patch = mp.Patch(patch_level=v['level'], version=v['version'], release=release, kernel_id=kernel)
        session.add(patch)

        for level_to_fix in v['fixups']:
            f = mp.Fixup(kernel_id=kernel, fixup_for=level_to_fix, fixup_from=v['level'], release=release)
            session.add(f)


class ReleaseName(str):
    pass


StickyT = Union[datetime, ReleaseName]


class Latest(TypedDict):
    name: str
    latest: int


def parse_sticky(value: str, now: Optional[datetime] = None) -> StickyT:
    """
    Parses sticky latest

    :param value: <num>h | <num>d | release-<release>

    <num>h and <num>d are relative sticky dates to specify patchsets released num hours or days ago
    respectively.

    release-<release> allows to filter out patchsets released after specified <release>
    """
    value = value.strip()
    rprefix = 'release-'
    if value.startswith(rprefix):
        return ReleaseName(value[len(rprefix) :])
    elif value.endswith('d'):
        return (now or datetime.utcnow()) - timedelta(days=int(value[:-1]))
    elif value.endswith('h'):
        return (now or datetime.utcnow()) - timedelta(hours=int(value[:-1]))
    raise ValueError('Invalid sticky format')


def get_klar(
    unit_id: int,
    stream: pc.Stream,
    feed: str,
    kernel: str,
    version: Optional[int] = None,
    sticky: Optional[StickyT] = None,
) -> Optional[Latest]:
    """
    Get last available release and it's data for specified kernel (KernelLastAvailableRelease).
    :param version: which latest version is supported by this patch level (latest.v1 = 1, latest.v2 = 2, etc.)
    :param feed: feed name
    :param kernel: kernel hash
    :return: {'name': '11122019_1', 'latest': 11}
    """

    sticky_release = None
    sticky_deployed = None
    if isinstance(sticky, ReleaseName):
        sticky_release = sticky
    else:
        sticky_deployed = sticky

    q = session.query_params()
    filters = [
        q.p.kernel_id == kernel,
        q.r.unit_id == unit_id,
        q.r.stream == stream.value,
        q.r.feed == feed,
        q.r.enabled,
        q.version <= not_none / version,
        q.release <= not_none / sticky_release,
        q.deployed <= not_none / sticky_deployed,
    ]
    sql = f'''\
        SELECT p.release, p.patch_level
        FROM patches p INDEXED BY idx_patches_kernel_id_patch_level
        INNER JOIN releases AS r USING (release)
        {WHERE(*filters)}
        ORDER BY p.patch_level DESC
        LIMIT 1
    '''

    klar = q.execute(sql).first()
    if not klar:
        return None

    release, patch_level = klar
    return {'name': release, 'latest': patch_level}


def get_flar(unit_id: int, stream: pc.Stream, feed: str, include_disabled: bool = False) -> Optional[str]:
    q = session.query_params()
    filters = [
        q.unit_id == unit_id,
        q.stream == stream.value,
        q.feed == feed,
        q.cond(not include_disabled, q.enabled),
    ]
    sql = f'''\
        SELECT release
        FROM releases
        {WHERE(*filters)}
        ORDER BY deployed DESC
        LIMIT 1
    '''

    return q.execute(sql).scalar()


@transaction(session)
def switch(unit_id: int, stream: pc.Stream, feed: Optional[str], enable: bool, release: str) -> None:
    upd = {'enabled': enable}

    flt = {'unit_id': unit_id, 'stream': stream.value, 'release': release}
    if feed:
        flt['feed'] = feed

    session.query(mp.Release).filter_by(**flt).update(upd, synchronize_session=False)  # type: ignore[arg-type]


def get_release_by_level(
    unit_id: int,
    stream: pc.Stream,
    feed: str,
    kernel: str,
    level: int,
) -> Optional[str]:
    """
    Find release that contains files from specified patch level
    :param feed: feed name
    :param kernel: kernel hash
    :param level: patch level
    :return:
    """
    q = session.query_params()
    sql = f'''\
        SELECT release
        FROM patches INNER JOIN releases AS r USING (release)
        WHERE kernel_id = {q/kernel}
          AND unit_id = {q/unit_id}
          AND stream = {q/stream.value}
          AND feed = {q/feed}
          AND patch_level = {q/level}
          AND enabled = 1
        LIMIT 1
    '''
    return q.execute(sql).scalar()


def get_release_with_fixup(unit_id: int, stream: pc.Stream, feed: str, kernel: str, level_to_fix: int) -> Optional[str]:
    """
    Find release that contains fixups for specified patch level:
    * omit "enabled" field, because fixups are never disabled.
    :param feed: feed name
    :param kernel: kernel hash
    :param level_to_fix: patch level that should be fixed
    :return: release name in DDMMYYYY_NN format
    """
    q = session.query_params()
    sql = f'''\
        SELECT release
        FROM fixups INNER JOIN releases USING (release)
        WHERE unit_id = {q/unit_id}
          AND stream = {q/stream.value}
          AND feed = {q/feed}
          AND kernel_id = {q/kernel}
          AND fixup_for = {q/level_to_fix}
        ORDER BY fixup_from DESC
        LIMIT 1
    '''
    return q.execute(sql).scalar()


def get_kernel_level_pairs(unit_id: int, stream: pc.Stream, feed: str) -> List[Tuple[str, int]]:
    q = session.query_params()
    sql = f'''\
        SELECT p.kernel_id, max(p.patch_level)
        FROM patches p
        INNER JOIN releases r USING (release)
        WHERE unit_id = {q/unit_id} AND stream = {q/stream.value} AND r.feed = {q/feed} AND r.enabled
        GROUP BY p.kernel_id
    '''
    return q.execute(sql)


def get_kernel_level_pairs_for_all_feeds(unit_id: int, stream: pc.Stream) -> List[Dict[str, Union[str, int]]]:
    q = session.query_params()
    sql = f'''\
        SELECT feed, p.kernel_id kernel_id, max(p.patch_level) patch_level
        FROM patches p
        INNER JOIN releases r USING (release)
        WHERE unit_id = {q/unit_id} AND stream = {q/stream.value} AND r.enabled
        GROUP BY feed, p.kernel_id
    '''
    return q.execute(sql, as_dict=True)


def get_release_status(unit_id: int, stream: pc.Stream, feed: str) -> Dict[str, int]:
    q = session.query_params()
    sql = f'SELECT release, enabled FROM releases WHERE unit_id = {q/unit_id} AND stream = {q/stream.value} AND feed = {q/feed}'
    return dict(q.execute(sql))


def release_deploy_status(unit_id: int, stream: pc.Stream) -> Dict[Tuple[str, str], int]:
    q = session.query_params()
    sql = f'SELECT feed, release, enabled FROM releases WHERE unit_id = {q/unit_id} AND stream = {q/stream.value}'
    return {(feed, release): status for feed, release, status in q.execute(sql)}


def instance_release_deploy_status(stream: pc.Stream) -> List[Tuple[int, str, str, int]]:
    q = session.query_params()
    sql = f'SELECT unit_id, feed, release, enabled FROM releases WHERE stream = {q/stream.value}'
    return q.execute(sql)


def get_releases(
    unit_id: int, stream: pc.Stream, feed: str = EMPTY, enabled: Optional[bool] = None, order_by_deployed: bool = False
) -> List[str]:
    q = session.query_params()
    filters = [
        q.unit_id == unit_id,
        q.stream == stream.value,
        q.cond(feed != EMPTY, q.feed == feed),
        q.cond(enabled is not None, q.enabled if enabled else ~q.enabled),  # could not work in PG
    ]

    group_by = q.cond(feed == EMPTY, 'GROUP BY release')
    order_by = q.cond(order_by_deployed, 'ORDER BY min(deployed)')

    sql = f'SELECT release FROM releases {WHERE(*filters)} {group_by} {order_by}'
    return q.execute(sql).column()


def get_minimum_releases(unit_id: int, stream: pc.Stream, feed: str, min_patches_per_kernel: int) -> List[str]:
    """
    Returns the minimum number of releases that contain minimum
    defined number of the last patches for all supported kernels.
    :param feed: feed name
    :param min_patches_per_kernel: minimum number of freshest patches for one kernel
    :return: list of release names
    """
    q = session.query_params()
    sql = f'''\
        SELECT ks.kernel_id, (
            SELECT group_concat(release, ',') AS releases
            FROM (
                SELECT release
                FROM patches INNER JOIN releases USING (release)
                WHERE kernel_id = ks.kernel_id
                  AND unit_id = {q/unit_id}
                  AND stream = {q/stream.value}
                  AND feed = {q/feed}
                  AND enabled
                ORDER BY patch_level DESC
                LIMIT {q/min_patches_per_kernel}
                )
            ) AS releases
        FROM (SELECT DISTINCT kernel_id FROM patches) AS ks
    '''

    data = q.execute(sql)
    if not data:
        return []

    releases: Set[str] = set()
    for _, r in data:
        if r:
            releases.update(r.split(','))

    return sorted(releases)


@transaction(session)
def remove_releases_from_feed(unit_id: int, stream: pc.Stream, names: Iterable[str], feed: str) -> None:
    if names:
        q = session.query_params()
        sql = f'''\
            DELETE FROM releases
            WHERE {q.release.IN(names)}
              AND unit_id = {q/unit_id}
              AND stream = {q/stream.value}
              AND feed = {q/feed}
        '''
        q.execute(sql)


@transaction(session)
def remove_feed(unit_id: int, feed: str) -> None:
    q = session.query_params()
    sql = f'''\
        DELETE FROM releases
        WHERE unit_id = {q/unit_id}
          AND feed = {q/feed}
    '''
    q.execute(sql)


@transaction(session)
def remove_all_releases_data(unit_id: int, stream: pc.Stream, names: Iterable[str]) -> None:
    if names:
        q = session.query_params()
        q.execute(
            f'DELETE FROM releases WHERE {q.release.IN(names)} AND unit_id = {q/unit_id} AND stream = {q/stream.value}'
        )
        # patches and fixups are shared between units and couldn't be deleted
        if not config.BUNITS:
            q.execute(f'DELETE FROM patches WHERE {q.release.IN(names)}')
            q.execute(f'DELETE FROM fixups WHERE {q.release.IN(names)}')


def is_stub_deployed(unit_id: int, stream: pc.Stream, kernel: str) -> bool:
    feeds = me.Feed.get_all_names(unit_id)
    for f in feeds:
        if not get_klar(unit_id, stream, f, kernel):
            return False
    return True


@transaction(session)
def add_stub(unit_id: int, stream: pc.Stream, kernel: str, now: datetime) -> None:
    feeds = set(me.Feed.get_all_names(unit_id))
    release_filters = dict(release=STUB_RELEASE, unit_id=unit_id, stream=stream.value)
    existing_stub_feeds = set(it for it, in session.query(mp.Release.feed).filter_by(**release_filters))
    stub_feeds_to_add = feeds - existing_stub_feeds

    for f in stub_feeds_to_add:
        r = mp.Release(**release_filters, feed=f, deployed=now, enabled=True)
        session.add(r)

    (session.query(mp.Release).filter(mp.Release.feed.in_(existing_stub_feeds)).filter_by(**release_filters)).update(
        {'deployed': now}, synchronize_session=False
    )

    already_deployed = session.query(mp.Patch).filter_by(release=STUB_RELEASE, kernel_id=kernel).first()
    if not already_deployed:
        p = mp.Patch(patch_level=0, version=1, release=STUB_RELEASE, kernel_id=kernel)
        session.add(p)


PatchLevelStat = List[Tuple[str, str, int]]


def release_usage_stat(stat: PatchLevelStat) -> Dict[str, Dict[Any, Any]]:
    """
    :param stat: list of tuples (feed, release, server_count)
    """
    result = {}
    cnt_sum = lambda g: sum(it[2] for it in g)

    key = lambda it: (it[0], it[1])
    result['feed_release'] = {k: cnt_sum(g) for k, g in groupby(sorted(stat, key=key), key)}

    key = lambda it: it[0]
    result['feed'] = {k: cnt_sum(g) for k, g in groupby(sorted(stat, key=key), key)}

    key = lambda it: it[1]
    result['release'] = {k: cnt_sum(g) for k, g in groupby(sorted(stat, key=key), key)}
    return result


def get_release_by_kernel_level(stream: pc.Stream, kernel_id: str, level: int) -> Optional[str]:
    q = session.query_params()
    sql = f'''\
        SELECT release
        FROM patches INDEXED BY idx_patches_kernel_id_patch_level
        INNER JOIN releases INDEXED BY idx_releases_stream_release USING (release)
        WHERE kernel_id = {q/kernel_id}
          AND patch_level = {q/level}
          AND stream = {q/stream.value}
        LIMIT 1
    '''
    return q.execute(sql).scalar()


def get_release_kernels(release: str) -> List[str]:
    q = session.query_params()
    sql = f'SELECT kernel_id FROM patches WHERE release = {q/release}'
    return q.execute(sql).column()
