import contextlib
import glob
import hashlib
import json
import logging
import os
import shutil
from datetime import datetime
from typing import List, Set, Any, Dict, Optional, Union, Callable, Iterable, NamedTuple, Tuple
from typing_extensions import Self, TypedDict

from werkzeug.exceptions import NotFound
import werkzeug.security

from eportal import utils, patches_constants as pc
from eportal.dbutils import transaction
from eportal.patches import api, session

log = logging.getLogger('eportal.patches')


def get_latest_version(fname: str) -> int:
    _, _, version = fname.partition('.v')
    return int(version)


def safe_join_or_404(directory: str, *pathnames: str) -> str:
    joined = werkzeug.security.safe_join(directory, *pathnames)
    if joined is None:
        # Unable to join into a safe path
        raise NotFound()
    return joined


class PatchStatus(NamedTuple):
    feed: str
    is_enabled: bool
    is_new: bool
    release: str

    def __str__(self) -> str:
        if self.is_new:
            return f'{self.release} deployed to {self.feed} with {"enabled" if self.is_enabled else "disabled"} status'
        else:
            return f'{self.release} was {"enabled" if self.is_enabled else "disabled"} in {self.feed}'


class ReleaseAwareLocation(str):
    release: str

    def __new__(cls, release: str, loc: str) -> Self:
        obj = str.__new__(cls, loc)
        obj.release = release
        return obj


class LatestJson(TypedDict):
    release: str
    level: int
    baseurl: str
    blocklist: List[str]


class PatchStore:
    _flavor_cache: Dict[str, Set[str]] = {}

    def __init__(self, unit_id: int, stream: pc.Stream, resource_dir: str, arch_dir: str):
        self.unit_id = unit_id
        self.stream = stream
        self.resource_dir = resource_dir
        self.arch_dir = arch_dir

    def release_flavors(self, release: str) -> Set[str]:
        cached = self._flavor_cache.get(release)
        if cached:
            return cached

        flavors = set(it.get('distro') or it['flavor'] for it in self.release_meta(release).values())
        result = self._flavor_cache[release] = flavors
        return result

    def get_release_content(self, release: str) -> Dict[str, Dict[str, Any]]:
        """
        Analyze the content of defined release and returns a dictonary with corresponding info.
        :param release: name of release in DDMMYYY_NN format (11122019_1)
        :return: {'kernel-1': {'patch_level': 12}, 'kernel-2': {'patch_level': 8}}
        """
        content: Dict[str, Dict[str, Any]] = {}
        release_dir = os.path.join(self.resource_dir, release)
        kernels = [d for d in os.listdir(release_dir) if os.path.isdir(os.path.join(release_dir, d))]
        for k in kernels:
            content[k] = {}
            latest_data = self.get_latest_data(release, k)
            content[k]['patch_level_v1'] = latest_data.get('latest.v1')
            content[k]['patch_level_v2'] = latest_data.get('latest.v2')
            content[k]['fixups'] = self.get_fixup_data(release, k)
        return content

    @transaction(session)
    def deploy_resources(
        self,
        release: str,
        feed: str,
        enabled: bool,
        deploy_date: Optional[datetime] = None,
    ) -> PatchStatus:
        deploy_date = deploy_date or datetime.utcnow()
        is_release_created = api.insert_release(self.unit_id, self.stream, release, feed, enabled, deploy_date)
        if is_release_created:
            api.insert_release_info(release, self.get_release_content(release))
        result = PatchStatus(
            feed=feed,
            is_enabled=bool(enabled),
            is_new=is_release_created,
            release=release,
        )
        log.info(str(result))
        return result

    def deploy_archive(
        self,
        archive_path: str,
        feed: str,
        enabled: bool,
        deploy_date: Optional[datetime] = None,
        unpack_cb: Optional[Callable[[], None]] = None,
    ) -> PatchStatus:
        """
        :param archive_path: absolute path to release archive
        :param feed: where release should be deployed
        :param enabled: should release be deployed as enabled or disabled one
        :param deploy_date: deploy with this release date
        :param unpack_cb: function to be called right before archive unpacking
        :return: deploy status info
        """
        release = utils.get_release_name(archive_path)
        self.unpack_archive(archive_path, unpack_cb=unpack_cb)
        return self.deploy_resources(release, feed, enabled, deploy_date)

    def unpack_archive(self, archive_path: str, unpack_cb: Optional[Callable[[], None]] = None) -> None:
        utils.unpack(archive_path, self.resource_dir, unpack_cb=unpack_cb)

    def get_release(self, feed: str, kernel: str, level: str = '', fname: str = '') -> Optional[str]:
        """
        Return release from /resources directory that should be used to serve requested file.
        There is different logic used for:
            1) fixup files (fixup_ext4.ko, fixup_ext4.ko.sig, kpatch.fixups, kpatch.fixups.sig)
               patch level in route means level to be fixed not the level that contains requested files,
               this information (release and fixed patch level pairs) is stored in separate DB table
            2) files with defined patch level
               exact patch level that contains requested files, just find a release that contains it,
               also, no restriction logic here, if the file exists -- serve it
            3) all other kernel files not restricted by patch level
               find KLAR (kernel last available release)
        :param feed: feed name
        :param kernel: kernel hash
        :param level: patch level (which should be fixed in case of fixup files or which contains requested files)
        :param fname: requested file name, only fixup files are counted
        :return: release name in DDMMYYYY_NN format
        """
        if level:
            try:
                int_level = int(level)
            except (ValueError, TypeError):
                return None

        if 'fixup' in fname:
            if not level:
                return None
            return api.get_release_with_fixup(self.unit_id, self.stream, feed, kernel, level_to_fix=int_level)
        elif level:
            return api.get_release_by_level(self.unit_id, self.stream, feed, kernel, int_level)
        else:
            latest = api.get_klar(self.unit_id, self.stream, feed, kernel)
            return latest and latest['name']

    def get_file_path(self, feed: str, kernel: str, level: str, fname: str) -> Optional[ReleaseAwareLocation]:
        release = self.get_release(feed, kernel, level, fname)
        if not release:
            return None

        if fname == 'kcare.ko' or fname == 'kcare.ko.sig' or fname == 'kcare.ko.sig2' or fname == 'kcare.ko.json-sig':
            loc = self.get_resource_location(release, kernel, '', fname)
        else:
            loc = self.get_resource_location(release, kernel, level, fname)
        if loc:
            return ReleaseAwareLocation(release, loc)
        return None

    def get_root_file_path(self, feed: str, fname: str) -> Optional[str]:
        release = api.get_flar(self.unit_id, self.stream, feed)
        if not release:
            return None
        return self.get_resource_location(release, fname)

    def get_resource_location(self, *parts: str) -> Optional[str]:
        fpath = safe_join_or_404(self.resource_dir, *parts)
        if os.path.exists(fpath) and os.path.isfile(fpath):
            return fpath
        return None

    def get_latest_content(
        self, feed: str, kernel: str, latest: str, sticky: Optional[api.StickyT] = None
    ) -> Optional[int]:
        release_info = api.get_klar(
            self.unit_id, self.stream, feed, kernel, version=get_latest_version(latest), sticky=sticky
        )
        return release_info and release_info['latest']

    def get_latest_json(
        self, feed: str, kernel: str, version: int, sticky: Optional[api.StickyT] = None
    ) -> Optional[LatestJson]:
        data = api.get_klar(self.unit_id, self.stream, feed, kernel, version=version, sticky=sticky)
        if not data:
            return None

        # Should be changed later. There is no correct implementation for getting 'blocklist' field right now.
        return {
            'release': data['name'],
            'level': data['latest'],
            'baseurl': f'/patches/{data["name"]}',
            'blocklist': [],
        }

    def disable_release(self, release: str, feed: Optional[str]) -> None:
        return api.switch(self.unit_id, self.stream, feed, enable=False, release=release)

    def enable_release(self, release: str, feed: Optional[str]) -> None:
        return api.switch(self.unit_id, self.stream, feed, enable=True, release=release)

    def get_patches_json(self, feed: str) -> Dict[str, Dict[str, Union[str, int]]]:
        # we cannot use info just from DB,
        # because patches.json contains additional information
        pjson_path = self.get_root_file_path(feed, 'patches.json')
        if not pjson_path:
            return {}

        with open(pjson_path, 'r') as f:
            ps_from_file = json.load(f)

        ps_from_db = api.get_kernel_level_pairs(self.unit_id, self.stream, feed)
        result = {}
        for kernel, level in ps_from_db:
            if kernel in ps_from_file:
                result[kernel] = ps_from_file[kernel]
                result[kernel]['latest'] = level
        return result

    def get_userspace_patches_json(self, feed: str) -> Dict[str, Dict[str, Union[str, int]]]:
        releases = self.get_releases(feed, enabled=True)

        res = {}
        for release in releases:
            ps_file = os.path.join(self.resource_dir, release, 'userspace-patches.json')
            if os.path.exists(ps_file):
                with open(ps_file) as fp:
                    ps_from_file = json.load(fp)
                res.update(ps_from_file)
        return res

    def get_latest_data(self, release: str, kernel: str) -> Dict[str, int]:
        """
        For specified release return values from it's latest.vX files.
        :param kernel: kernel hash
        :param release: release name
        :return: {'latest.v1': 1, 'latest.v2': 2}
        """
        kernel_dir = os.path.join(self.resource_dir, release, kernel)
        result = {}
        for latest_type in ['latest.v1', 'latest.v2']:
            lpath = os.path.join(kernel_dir, latest_type)
            if os.path.exists(lpath):
                with open(lpath, 'r') as lfile:
                    result[latest_type] = int(lfile.read().strip())
        return result

    def get_latest_libcare(
        self, feed: str, build_id: str, latest: str, prefix: str
    ) -> Optional[Dict[str, Union[str, int]]]:
        data = api.get_klar(self.unit_id, self.stream, feed, build_id, version=get_latest_version(latest))
        if not data:
            return None
        return {
            'release': data['name'],
            'level': data['latest'],
            'patch_url': f'{prefix}/patches/{data["name"]}/{build_id}/patch.tar.gz',
        }

    def get_libcare_info(self, build_id: str, level: Union[str, int]) -> Optional[Dict[str, str]]:
        if self.stream in (pc.Stream.QEMU, pc.Stream.DB):
            release = utils.libcare_level_to_release(level, self.stream.value)
        else:
            release = utils.libcare_level_to_release(level)

        f = self.get_resource_location(release, 'userspace-patches.json')
        if not f:
            return None
        with open(f) as fp:
            patches_json = json.load(fp)

        f = self.get_resource_location(release, build_id, 'info.json')
        if not f:  # pragma: no cover
            return None
        with open(f) as fp:
            info_json = json.load(fp)

        info_json['time'] = patches_json[build_id]['time']
        return info_json

    def get_fixup_data(self, release: str, kernel: str) -> List[str]:
        kernel_dir = os.path.join(self.resource_dir, release, kernel)
        patch_level = '*'
        kpatch_fixups_mask = os.path.join(kernel_dir, patch_level, 'kpatch.*fixups')
        return [os.path.basename(os.path.dirname(fixup)) for fixup in glob.glob(kpatch_fixups_mask)]

    def remove_releases(self, releases: Iterable[str]) -> None:
        api.remove_all_releases_data(self.unit_id, self.stream, releases)
        for r in releases:
            shutil.rmtree(os.path.join(self.resource_dir, r), True)
            for ext in ('.tar.bz2', '.tar.bz2.lock', '.tar.bz2.sha256'):
                filename = os.path.join(self.arch_dir, r + ext)
                with contextlib.suppress(FileNotFoundError):
                    os.remove(filename)

    def get_available_resources(self) -> List[str]:
        if not os.path.exists(self.resource_dir):
            return []
        names = os.listdir(self.resource_dir)
        result = []
        for name in names:
            release_dir = os.path.join(self.resource_dir, name)
            patches_json = os.path.join(release_dir, 'patches.json')
            if os.path.isdir(release_dir) and os.path.isfile(patches_json):
                result.append(name)
        return result

    def clean_releases(
        self, public_releases: List[str], feeds: List[str], min_patches_per_kernel: int = 3, dry_run: bool = False
    ) -> Set[str]:
        public = set(public_releases)
        overall_min = set()
        for feed in feeds:
            feed_min = set(api.get_minimum_releases(self.unit_id, self.stream, feed, min_patches_per_kernel))
            overall_min.update(feed_min)
            feed_keep = public | feed_min
            feed_deployed = set(self.get_releases(feed, enabled=True))
            if not dry_run:
                self.remove_releases_from_feed(feed_deployed - feed_keep, feed)

        resources = set(self.get_available_resources())
        obsolete_releases = resources - (overall_min | public)
        if not dry_run:
            self.remove_releases(obsolete_releases)
        return obsolete_releases

    @transaction(session)
    def prepare_stub(self, version: str, now: Optional[datetime] = None) -> str:
        # kcarectl generate kernel_id by reading /proc/version and using SHA1,
        # this file contains \n in the end, so, add it here
        version = version.strip() + '\n'

        now = now or datetime.utcnow()
        # sha1 is not used for security here, using a broad nosec as Bandit does
        # not recognize a specific nosec.
        kernel_id = hashlib.sha1(version.encode()).hexdigest()  # nosec
        stub_dir = os.path.join(self.resource_dir, 'stubs', kernel_id)
        if os.path.exists(os.path.join(stub_dir, 'version')) and api.is_stub_deployed(
            self.unit_id, self.stream, kernel_id
        ):
            return kernel_id
        api.add_stub(self.unit_id, self.stream, kernel_id, now)
        os.makedirs(stub_dir, exist_ok=True)
        utils.atomic_write(os.path.join(stub_dir, 'version'), version)
        return kernel_id

    def release_is_unpacked(self, name: str) -> bool:
        return os.path.exists(os.path.join(self.resource_dir, name, 'unpack.done'))

    def release_meta(self, release: str) -> Dict[str, Any]:
        if self.stream != pc.Stream.KERNEL:
            # userspace patches.json already contains only relevant buildid data
            # and can be used as-is
            fname = os.path.join(self.resource_dir, release, 'userspace-patches.json')
            return json.loads(utils.read_file(fname) or b'{}')

        # kernel patches.json contains cumulative info for all previous releases
        # and we need extract only kernels included in a corresponding release
        fname = os.path.join(self.resource_dir, release, 'patches.json')
        data = json.loads(utils.read_file(fname) or b'{}')
        release_kernels = set(api.get_release_kernels(release))
        return {it: data[it] for it in release_kernels if it in data}

    def get_releases(
        self, feed: str = api.EMPTY, enabled: Optional[bool] = None, order_by_deployed: bool = False
    ) -> List[str]:
        return api.get_releases(
            self.unit_id, self.stream, feed=feed, enabled=enabled, order_by_deployed=order_by_deployed
        )

    def remove_releases_from_feed(self, names: Iterable[str], feed: str) -> None:
        return api.remove_releases_from_feed(self.unit_id, self.stream, names, feed)

    def release_deploy_status(self) -> Dict[Tuple[str, str], int]:
        return api.release_deploy_status(self.unit_id, self.stream)

    def get_release_status(self, feed: str) -> Dict[str, int]:
        return api.get_release_status(self.unit_id, self.stream, feed)


class CachedPatchStore(PatchStore):
    _flavor_cache = {}

    def get_release_content(self, release: str) -> Dict[str, Dict[str, Any]]:
        fname = os.path.join(self.arch_dir, f'{release}.content.json')
        data = json.load(open(fname))
        info = {}
        for k, meta in data['result'].items():
            info[k] = {
                'patch_level_v1': meta['patch_level'] if int(meta['version']) == 1 else None,
                'patch_level_v2': meta['patch_level'] if int(meta['version']) == 2 else None,
                'fixups': meta.get('fixups') or [],
            }
        return info

    def release_is_unpacked(self, name: str) -> bool:
        return os.path.exists(os.path.join(self.arch_dir, f'{name}.content.json'))

    def release_meta(self, release: str) -> Dict[str, Any]:
        fname = os.path.join(self.arch_dir, f'{release}.content.json')
        data = utils.read_file(fname)
        return data and json.loads(data)['result'] or {}

    def get_resource_location(self, *parts: str) -> str:
        return safe_join_or_404(self.resource_dir, *parts)

    def get_patches_json(self, feed: str) -> Dict[str, Union[Dict[str, Union[str, int]]]]:
        result: Dict[str, Any] = {}
        for release in self.get_releases(feed, enabled=True):
            data = utils.read_file(os.path.join(self.arch_dir, f'{release}.content.json'))
            if not data:
                continue

            for k, v in json.loads(data)['result'].items():
                try:
                    rv = result[k]
                except KeyError:
                    result[k] = v
                else:
                    if v['patch_level'] > rv['patch_level']:
                        result[k] = v
        return result
