import os
import json
import hashlib
import logging

from urllib.parse import urlencode
from shutil import copyfileobj, disk_usage
from typing import cast, Callable, List, Tuple, Any, Dict, Optional

import config
from eportal import dynconfig, utils, patches_constants as pc
from eportal.db.models import server_statistics

log = logging.getLogger(__name__)


class ChecksumMismatch(Exception):
    pass


class RepoConfigRequired(Exception):
    pass


class PatchSource:
    def __init__(self, config: Optional[dynconfig.Config] = None):
        self.config = config or dynconfig.Config()

    def is_configured(self) -> bool:
        return self.config.patchrepo['enabled']

    def check_configuration(self) -> None:
        if not self.is_configured():
            raise RepoConfigRequired('ePortal patch source is not configured.')

    @property
    def auth(self) -> Tuple[str, str]:
        return self.config.patchrepo['login'], self.config.patchrepo['password']

    def get_url(self, name: str) -> str:
        url = self.config.patchrepo['url'].rstrip('/')
        name = name.lstrip('/')
        return f'{url}/{name}'

    def get_available_distros(self) -> List[str]:
        if not self.is_configured():
            return FALLBACK_AVAILABLE_DISTROS

        try:
            with utils.get_session() as s:
                r = s.get(self.get_url('available-distros'), **utils.get_req_args(auth=self.auth, timeout=15))
            if r.status_code != 200:
                return FALLBACK_AVAILABLE_DISTROS
        except Exception:  # pragma: no cover, ignore any other possible exceptions
            return FALLBACK_AVAILABLE_DISTROS

        return r.json()['result'] or FALLBACK_AVAILABLE_DISTROS

    def probe(self) -> None:
        with utils.get_session() as s:
            r = s.get(self.get_url('signatures.json'), **utils.get_req_args(auth=self.auth, timeout=15))
            r.raise_for_status()

    def get_updates(
        self, stream: pc.Stream, psproduct: pc.PSProduct, channel: pc.Channel, forbidden_ok: bool = True
    ) -> Dict[str, Any]:
        self.check_configuration()
        stat = {
            'stat': server_statistics(active=True),
            'overall': server_statistics(active=False),
        }

        # params are used to cache updates.json response
        params = {
            'product': psproduct.psvalue,
            'stream': stream.psvalue,
            'channel': channel.psvalue,
        }

        if stream == pc.Stream.KERNEL:
            enabled_distros = self.config.patchrepo['distros']
            if enabled_distros:
                params['distros'] = ','.join(enabled_distros)

        fname = os.path.join(config.PATCH_RESOURCES_DIR, 'updates.json')
        fname_cached, data = utils.try_file(fname, expire=600, hash_args=params.items())

        if data is None:
            url = f'updates.json?{urlencode({**params, **stat})}'
            full_url = self.get_url(url)
            with utils.get_session() as s:
                r = s.get(full_url, **utils.get_req_args(auth=self.auth, timeout=15))

            if r.status_code != 200:
                if r.status_code == 403:
                    error = f'ePortal account has no access to {psproduct}/{stream}.'
                else:
                    error = (
                        f'{full_url} is not available ({r.status_code}). '
                        'Please check you patch source url, credentials, proxy settings.'
                    )

                if r.status_code == 403:
                    log.info('Forbidden update request %r', params)
                    data = json.dumps({'blocked': [], 'releases': [], 'error': error}).encode()
                else:
                    raise Exception(error)
            else:
                data = r.content

            utils.atomic_write(fname_cached, data, safe=True)

        result = json.loads(data)
        if not forbidden_ok and (result_error := result.get('error')):
            raise Exception(result_error)

        return cast(Dict[str, Any], result)

    def get_changelog_text(self, name: str) -> str:
        self.check_configuration()
        fname = os.path.join(config.ARCH_DIR, name + '.changelog')
        fname_cached, data = utils.try_file(fname, 86400)
        if data is None:
            url = self.get_url(name + '.changelog')
            with utils.get_session() as s:
                response = s.get(url, **utils.get_req_args(auth=self.auth, timeout=15))
            response.raise_for_status()
            utils.atomic_write(fname_cached, response.content)
            data = response.content
        return data.decode()

    def download(
        self,
        remote_file: str,
        destination: str,
        chunk_size: int = 1048576,
        download_cb: Optional[Callable] = None,
        check_is_json: bool = False,
        force: bool = False,
        expected_checksum: Optional[str] = None,
    ) -> float:
        self.check_configuration()
        tmp = destination + '.tmp'
        hash_path = destination + '.sha256'

        if (
            not force
            and os.path.exists(destination)
            and (not expected_checksum or expected_checksum == (utils.read_file(hash_path) or b'').decode())
        ):
            return os.stat(destination).st_mtime

        url = self.get_url(remote_file)
        with utils.lock(destination + '.lock', block=True):
            download_cb and download_cb()
            with utils.get_session() as s:
                response = s.get(url, **utils.get_req_args(auth=self.auth, stream=True, timeout=15))
                response.raise_for_status()

            # check for enough disk space
            content_length = response.headers.get('Content-Length')
            hashed_response = utils.HashedStream(hashlib.sha256, response.raw)
            disk_space = disk_usage(os.path.dirname(destination)).free

            if content_length and disk_space < int(content_length):
                file_size = int(content_length) // 1024 // 1024
                raise Exception(f'Not enough disk space. Patch size: {file_size}Mb')

            # save file temporary
            with open(tmp, 'wb') as fd:
                copyfileobj(hashed_response, fd, chunk_size)

            # check if the body's control sum is the expected one
            if content_length and hashed_response.length != int(content_length):
                raise Exception('Incomplete read {0} != {1}: {2}'.format(hashed_response.length, content_length, url))

            # check if the checksum is the expected one
            checksum = hashed_response.hexdigest()
            if expected_checksum and expected_checksum != checksum:
                raise ChecksumMismatch(
                    'Invalid checksum for {0}: expected {1}, received {2}'.format(tmp, expected_checksum, checksum)
                )

            if check_is_json:
                data = (utils.read_file(tmp) or b'').decode()
                json.loads(data)

            # keep the file for further usage
            os.rename(tmp, destination)

            if expected_checksum:
                utils.atomic_write(hash_path, checksum)

            return os.stat(destination).st_mtime


def check_repo(login: str, password: str, url: str) -> None:
    # try to read repo with the given parameters
    # PatchRepo will raise exception on error
    config = dynconfig.Config()
    setattr(config, 'data', {'patchrepo': {'url': url, 'login': login, 'password': password}})

    ps = PatchSource(config)
    ps.probe()


FALLBACK_AVAILABLE_DISTROS = [
    "almalinux8",
    "almalinux8-arm64",
    "almalinux8-rpi",
    "almalinux9",
    "almalinux9-arm64",
    "amazon",
    "centos6",
    "centos7",
    "centos8",
    "cl6",
    "cl7",
    "cl8",
    "custom",
    "debian10",
    "debian11",
    "debian11-cloud",
    "debian12",
    "debian8",
    "debian9",
    "oel6",
    "oel7",
    "oel8",
    "oel9",
    "oel9-uek7",
    "openvz",
    "pve",
    "raspberrypi",
    "rhel6",
    "rhel7",
    "rhel8",
    "rhel8-arm64",
    "rhel9",
    "rockylinux8",
    "rockylinux9",
    "rpios-arm64",
    "sl6",
    "ubuntu14.04",
    "ubuntu16.04",
    "ubuntu18.04",
    "ubuntu20.04",
    "ubuntu22.04",
    "ubuntu24.04",
]
