import logging
import re
from typing import Callable, Iterable, Tuple, Any, cast, Dict, Optional, IO

from requests import Response

from eportal import dynconfig
from eportal.utils import get_session, get_req_args
from eportal.db.models import get_kc_key_fields_by_server, is_product_allowed
from eportal.patches_constants import Product

log = logging.getLogger('eportal-proxy')
session = get_session()

# # Uncomment to debug http.client send/recv data
# import http.client
# http.client.HTTPConnection.debuglevel = 1


def extract_headers(env: Dict[str, str]) -> Dict[str, str]:
    headers = {}
    for k, v in env.items():
        if k.startswith('HTTP_'):
            h = k[5:].replace('_', '-').title()
            if h.lower() not in ('host', 'connection', 'x-forwarded-proto', 'x-forwarded-for'):
                headers[h] = v

    if 'CONTENT_LENGTH' in env:
        headers['Content-Length'] = env['CONTENT_LENGTH']

    if 'CONTENT_TYPE' in env:
        headers['Content-Type'] = env['CONTENT_TYPE']

    return headers


# requests is quite dumb and forces chunked encoding if body is a stream.
# This wrapper prevents requests to use chunked encoding.
# Moreover for chunked encoding requests prefers iter. And for uwsgi.input
# it iterates by lines. Yes, one socket send per one input line.
class LengthAwareStream:
    def __init__(self, fp: IO[bytes], length: int) -> None:
        self.fp = fp
        self.length = length
        self.__iter__ = fp.__iter__  # makes requests believe it's a good body
        self.read = fp.read  # makes http.client believe it's a good body

    def __len__(self) -> int:  # requests checks availability of __len__ to prevent chunked encoding
        return self.length


def reader(resp: Response) -> Iterable[bytes]:
    try:
        while True:
            data = resp.raw.read(65536)
            if not data:
                break
            yield data
    finally:
        resp.close()


def get_patch_source_url_and_auth(args: dict) -> Tuple[str, dict]:
    conf = dynconfig.Config()
    auth = conf.patchrepo['login'], conf.patchrepo['password']
    url = conf.patchrepo['url'].rstrip('/')
    return url, get_req_args(auth=auth, **args)


def _parse_tuxcare_url(url: str) -> Tuple[Optional[str], Optional[str], str]:
    product, *parts = url.lstrip('/').split('/')
    if not parts:
        # []
        server_id = None
        proxied_url = ''
    elif product == 'tuxcare' and len(parts) == 1:
        # ['tuxcare-release-latest-9.noarch.rpm'] or ['']
        server_id = None
        proxied_url = parts[0]
    elif product == 'tuxcare' and parts[1] == 'base':
        # ['9', 'base' ...]
        server_id = None
        release, reponame, *url_end = parts
        proxied_url = '/'.join([release, reponame, *url_end])
    else:
        # ['9', '<server_id>', 'updates' ...]
        release, server_id, reponame, *url_end = parts
        proxied_url = '/'.join([release, reponame, *url_end])

    return product, server_id, proxied_url


def application(env: dict, start_response: Callable) -> Iterable:
    patchserver_prefix = '/patchserver/'
    try:
        url_to_download = env['PATH_INFO']

        # By default it will redirect to the repo
        url = 'https://repo.cloudlinux.com' + url_to_download
        args: Dict[str, Any] = dict(headers=None)

        if url_to_download.startswith('/kernel-patches'):
            base_url, args = get_patch_source_url_and_auth(args)
            url = base_url + url_to_download
        elif url_to_download.startswith('/doctor'):
            url = 'https://patches.kernelcare.com' + url_to_download
            args['headers'] = extract_headers(env)
        elif url_to_download.startswith(patchserver_prefix):
            ps_path = url_to_download[len(patchserver_prefix) - 1 :]
            if re.match('/(almalinux|tuxcare)/.*', ps_path):
                try:
                    product, server_id, proxied_url = _parse_tuxcare_url(ps_path)
                except Exception:
                    log.warning(f'failed to parse tuxcare url: {ps_path}')
                    start_response('400 Bad Request', [])
                    return []

                if server_id:
                    sinfo = get_kc_key_fields_by_server(server_id)
                    if not sinfo:
                        start_response('401 Unauthorized', [])
                        return []

                    if not is_product_allowed(Product.TUXCARE_ESU, sinfo.products):
                        start_response('403 Forbidden', [])
                        return []

                    _, args = get_patch_source_url_and_auth(args)

                ps_path = f'/proxy/{product}/' + proxied_url

            url = 'https://patches.kernelcare.com' + ps_path

        query = env['QUERY_STRING']
        if query:
            url = f'{url}?{query}'

        if 'CONTENT_LENGTH' in env:
            body = LengthAwareStream(env['wsgi.input'], int(env['CONTENT_LENGTH']))
            args.setdefault('headers', {})['Content-Length'] = env['CONTENT_LENGTH']
        else:
            body = None

        args.update(dict(timeout=10, stream=True))
        resp = session.request(env['REQUEST_METHOD'], url, data=body, **cast(Dict[Any, Any], args))

        if 'Content-Length' in resp.headers:
            content = reader(resp)
            length = str(resp.headers['Content-Length'])
        else:
            data = resp.content
            length = str(len(data))
            content = [data]

        headers = {
            'Upstream-Url': url,
            'Content-Length': length,
            'Content-Type': resp.headers.get('Content-Type', 'application/octet-stream'),
        }
        start_response(f'{resp.status_code} {resp.reason}', list(headers.items()))
        return content
    except Exception:
        log.exception(f'Error fetching {url_to_download}')
        start_response('500 PROXY ERROR', [])
        return []
