import json
import time
from datetime import datetime
from typing import Any, Dict, List, Optional, Set, Tuple, Union, TypeVar

from sqlalchemy.dialects import sqlite
from sqlalchemy.sql.elements import quoted_name

from eportal import checkin_saver, oplog, audit, patches_constants as pc, dynconfig
from eportal.utils import TaskOpts, adapt_args, iter_chunks
from eportal.db import models as m, session as main_session
from eportal.dbutils import transaction, EportalSession
from eportal.patches import api

ReplicatedModel = TypeVar(
    'ReplicatedModel',
    bound=Union[m.User, m.Feed, m.KcKey, m.KcServer, m.Unit, m.UnitUser],
)

TABLE_MODELS = {M.__tablename__: M for M in (m.User, m.Feed, m.KcKey, m.KcServer)}


# deprecated, need for compatibility
@oplog.replicate('server-checkin', oplog.CLEAN_POLICY_AGGRESSIVE)  # pragma: no cover
def checkin_server(data: Dict[str, Any]) -> None:
    checkin_saver.save_kernel_stat([data])


@oplog.replicate('server-checkin-many', oplog.CLEAN_POLICY_AGGRESSIVE)
def checkin_server_many(data: List[Dict[str, Any]]) -> None:
    checkin_saver.save_kernel_stat(data)


@oplog.replicate('userspace-checkin', oplog.CLEAN_POLICY_AGGRESSIVE)
def checkin_userspace(data: List[Dict[str, Any]]) -> None:
    checkin_saver.save_userspace_stat(data)


@oplog.register_action('server-register')
def register_server_by_id(server_id: str, ip: str, hostname: str, key: str) -> None:
    m.add_server(ip, hostname, key, server_id=server_id)


def register_server(ip: str, hostname: str, key: str, replicate: bool = True) -> m.KcServer:
    server = m.add_server(ip, hostname, key)
    params = {
        'server_id': server.server_id,
        'ip': ip,
        'hostname': hostname,
        'key': key,
    }
    replicate and register_server_by_id.write(**params)
    return server


@oplog.replicate('server-unregister')
def unregister_server(server_id: str) -> None:
    return m.delete_server(server_id)


@oplog.replicate('server-unregister-by-key-and-ip')
def unregister_server_by_key_and_ip(key: str, ip: str) -> int:
    return m.delete_server_by_key_and_ip(key, ip)


@oplog.replicate('server-tag')
def tag_server(server_id: str, tags: str) -> bool:
    return m.tag_server(server_id, tags)


@oplog.replicate('manage-sync-table-record')
@transaction(main_session)
def sync_table_record(table: Union[str, quoted_name], fields: Dict[str, Any]) -> None:
    q = main_session.query_params()
    q.execute(f'REPLACE INTO {table} {q.VALUES(**fields)}')


@oplog.replicate('manage-delete-table-record-v2')
@transaction(main_session)
def delete_table_record(model_name: str, pk: Any) -> None:
    obj = main_session.get(getattr(m, model_name), pk)
    if obj:
        if hasattr(obj, 'on_delete'):
            obj.on_delete()
        main_session.delete(obj)


@oplog.replicate('manage-delete-table-record')
@transaction(main_session)
def delete_table_record_legacy(table: str, pk: str, value: Any, fix_keys_for_feed: Optional[str] = None) -> None:
    _ = pk
    pkvalue: Any
    if table == 'feed':
        pkvalue = {'unit_id': 0, 'name': value}
    else:
        pkvalue = value
    delete_table_record.action(TABLE_MODELS[table].__name__, pkvalue)


def sync_model(model: ReplicatedModel, user: str = 'UNKNOWN') -> None:
    data = model.to_json()
    sync_table_record.action.write(table=model.__tablename__, fields=data)

    if hasattr(model, 'audit_change_msg'):
        audit_log(getattr(model, 'unit_id', 0), user, *model.audit_change_msg())


def delete_model(model: ReplicatedModel, user: str = 'UNKNOWN') -> None:
    pk = {col.name: getattr(model, col.name) for col in model.__table__.primary_key}
    delete_table_record.action.write(model_name=model.__class__.__name__, pk=pk)

    if hasattr(model, 'audit_delete_msg'):
        audit_log(getattr(model, 'unit_id', 0), user, *model.audit_delete_msg())


@oplog.replicate('manage-deploy-patchset')
@adapt_args(defaults={'unit_id': 0}, rename={'store': 'stream'})
def deploy_patchset(
    stream: str,
    unit_id: int,
    patchset: str,
    deploy_orders: Dict[str, str],
    user: Optional[str] = None,
) -> None:
    from eportal.patches_impl import get_repo
    from eportal import tasks

    repo = get_repo(unit_id, pc.Stream.items[stream])
    for feed, status in deploy_orders.items():
        if status == 'disable':
            repo.rollback_patches(feed, patchset, user=user)
        elif status == 'enable':
            tasks.deployment_worker_single.push(
                TaskOpts(urgent=True), unit_id=unit_id, feed=feed, patchset=patchset, store_type=stream, user=user
            )
        elif status == 'enable-upto':
            tasks.deployment_worker.push(
                TaskOpts(urgent=True), unit_id=unit_id, feed=feed, patchset=patchset, store_type=stream, user=user
            )
        elif status == 'undeploy-downto':
            repo.undeploy_down_to(feed, patchset, user=user)


def get_registrations() -> Dict[str, list]:
    sql = '''SELECT s.kc_key, s.server_id, s.remote_ip, s.node, m.mltag
        FROM kc_servers s LEFT JOIN kc_servers_mltag_view m USING(server_id)'''
    fields = ['kc_key', 'server_id', 'remote_ip', 'node', 'mltag']
    return {'fields': fields, 'result': main_session.query_params().execute(sql)}


def _add_tags(session: EportalSession, tags: Set[Tuple[int, str]]) -> None:
    values = [{'unit_id': unit_id, 'name': name} for unit_id, name in tags]
    query = sqlite.insert(m.Tag).values(values).on_conflict_do_nothing()
    session.execute(query)
    session.flush()


def _add_tags_assignment(session: EportalSession, tagged_servers: List[Tuple[str, int, Dict[str, Any]]]) -> None:
    tags_map = {(t.unit_id, t.name): t.tag_id for t in session.query(m.Tag).all()}
    values: List[Dict[str, Any]] = []
    for server_id, unit_id, tags in tagged_servers:
        values.extend(
            {'server_id': server_id, 'tag_id': tags_map[(unit_id, name)], 'tag_value': None if value is True else value}
            for name, value in tags.items()
        )
    for chunk in iter_chunks(values, 1000):
        query = sqlite.insert(m.KcServerTag).values(chunk).on_conflict_do_nothing()
        session.execute(query)


def apply_server_registrations(registrations: Dict[str, list]) -> None:
    # m.add_server is too slow to use directly, here we use sqlite json
    # functions to upsert registrations, it provides 50x speedup

    q = main_session.query_params()
    fields = [f for f in registrations['fields'] if f != 'mltag']
    jfields = ', '.join(f"json_extract(jd.value, '$[{i}]')" for i in range(len(fields)))
    sql = f'''\
        REPLACE INTO kc_servers ({', '.join(fields)})
        SELECT {jfields}
        FROM json_each(:data) jd
    '''

    unit_by_key = {k.key: k.unit_id for k in main_session.query(m.KcKey).all()}
    for chunk in iter_chunks(registrations['result'], 1000):
        # chunk is a tuple with fields from :func:`get_registration`:
        # ['kc_key', 'server_id', 'remote_ip', 'node', 'mltag']

        with transaction(main_session):
            q['data'] = json.dumps(chunk)  # type: ignore[call-arg]
            q.execute(sql)

            tagged_servers = [(r[0], unit_by_key.get(r[1], 0), m.Tag.tag_to_map(r[-1])) for r in chunk if r[-1]]
            if tagged_servers:
                tags_set = set((unit_id, name) for _, unit_id, tags in tagged_servers for name in tags)
                _add_tags(main_session, tags_set)
                _add_tags_assignment(main_session, tagged_servers)


def full_changes() -> oplog.FullChanges:
    return {
        'users': [it.to_json() for it in main_session.query(m.User)],
        'feeds': [it.to_json() for it in main_session.query(m.Feed)],
        'keys': [it.to_json() for it in main_session.query(m.KcKey)],
        'units': [it.to_json() for it in main_session.query(m.Unit)],
        'unit_users': [it.to_json() for it in main_session.query(m.UnitUser)],
        'patches': {it.value: api.instance_release_deploy_status(it) for it in pc.REPLICATED_STREAMS},
        'config': dynconfig.Config().data,
    }


def apply_full_changes(changes: oplog.FullChanges, audit_user: str = 'AUTO') -> None:
    from eportal import tasks

    for it in changes.get('units', []):
        sync_table_record.action(m.Unit.__tablename__, it)

    for it in changes.get('unit_users', []):
        sync_table_record.action(m.UnitUser.__tablename__, it)

    for it in changes.get('users', []):
        sync_table_record.action(m.User.__tablename__, it)

    for it in changes.get('feeds', []):
        sync_table_record.action(m.Feed.__tablename__, it)

    for it in changes.get('keys', []):
        sync_table_record.action(m.KcKey.__tablename__, it)

    config_data = changes.get('config')
    if config_data:
        dynconfig.save_config_data.action(config_data)

    patches = changes.get('patches', {})
    for stream_type, remote_status in patches.items():
        stream = pc.Stream.items[stream_type]
        local_status = {
            (unit_id, feed, release): status
            for unit_id, feed, release, status in api.instance_release_deploy_status(stream)
        }
        for unit_id, feed, patchset, enabled in remote_status:
            local_enabled = local_status.get((unit_id, feed, patchset))
            if local_enabled == enabled:
                continue

            if local_enabled is None:
                audit_msg = (
                    f'downloading patchset {patchset} with enabled={enabled} to feed {feed} during reconciliation'
                )
            elif enabled:
                audit_msg = f'enabling patchset {patchset} in feed {feed} during reconciliation'
            else:
                audit_msg = f'disabling patchset {patchset} in feed {feed} during reconciliation'

            audit.add_record(unit_id, audit_user, audit.ACTION_MANAGE_PATCHSET, audit_msg)
            tasks.deployment_worker_single.push(
                TaskOpts(urgent=True),
                unit_id=unit_id,
                feed=feed,
                patchset=patchset,
                store_type=stream_type,
                enabled=enabled,
            )

        remote_patchsets = set((unit_id, feed, patchset) for unit_id, feed, patchset, _ in remote_status)
        for unit_id, feed, patchset in set(local_status) - remote_patchsets:
            audit.add_record(
                unit_id,
                audit_user,
                audit.ACTION_MANAGE_PATCHSET,
                f'removed {patchset} patchset from feed {feed} during reconciliation',
            )
            api.remove_releases_from_feed(unit_id, stream, [patchset], feed)


@oplog.replicate('server-cleanup')
def delete_servers(
    server_ids: Optional[List[str]] = None,
    ips: Optional[List[str]] = None,
    hosts: Optional[List[str]] = None,
    last_check_in_age: Optional[int] = None,
) -> int:
    return m.delete_servers(server_ids=server_ids, ips=ips, hosts=hosts, last_check_in_age=last_check_in_age)


@oplog.register_action('audit-log')
def _audit_log(unit_id: int, user: str, action: str, details: str, ts: int) -> None:
    audit.add_record(unit_id, user, action, details, datetime.utcfromtimestamp(ts))


def audit_log(unit_id: int, user: str, action: str, details: str) -> None:
    ops = dict(unit_id=unit_id, user=user, action=action, details=details, ts=int(time.time()))
    _audit_log(**ops)
    _audit_log.write(**ops)


@oplog.replicate('create-user')
@adapt_args(defaults={'unit_id': 0})
def get_or_create_user(
    unit_id: int, username: str, description: str, role: str = 'readonly'
) -> Tuple[Union[m.User, m.UnitUser], bool]:
    with transaction(main_session):
        user = m.get_user(username, unit_id)
        if user:
            return user, False

        if unit_id:
            user = m.UnitUser(  # nosec hardcoded_password_funcarg
                unit_id=unit_id, username=username, password_hash='', role=None, description=description
            )
        else:
            # empty password to prevent access after user deletion from identity provider
            user = m.User(  # nosec hardcoded_password_funcarg
                username=username, description=description, password_hash='', active=True
            )
            user.set_role(role)

        main_session.add(user)
        main_session.flush()

    return user, True
