import hashlib
import json
import logging
import os

import distro
import socket
from datetime import datetime
from urllib.parse import urlencode
from typing import Any, Dict

from requests.exceptions import RequestException

import config
from eportal import oplog, patchsource
from eportal.db import models
from eportal.utils import get_version, get_session, atomic_write, bstr, get_req_args, get_instance_id, Table
from eportal.patches_constants import Product

log = logging.getLogger(__name__)


def send_report(source: str = 'schedule') -> None:
    """
    :param source: can be `schedule` or `doctor`
            `schedule` request is made by ePortal itself using uWSGI Spooler --> eportal.kcare.eportal.scheduler
            `doctor` request is made by --doctor option from kc.eportal
    :return:
    """
    ps = patchsource.PatchSource()
    if not ps.is_configured():
        log.warning('Report could not be sent: ePortal patch source is not configured.')
        return

    hostname = socket.gethostname()
    data = get_data_report(ps.auth[0], hostname)

    log.info('Reporting: data collected...')

    try:
        save_as_file(data)
    except Exception as err:  # pragma: no cover TODO
        log.warning(f'Report saving failed: {err}')

    version, platform = get_version()
    query = urlencode(dict(source=source, version=version, platform=platform))
    url = ps.get_url('server_stats?{0}'.format(query))

    try:
        with get_session() as s:
            response = s.post(url, json=data, timeout=30, **get_req_args(auth=ps.auth))
            if response.status_code >= 300:
                log.error(f'Reporting: failed {response.status_code}!')
            else:
                log.info('Reporting: data is sent!')
    except RequestException as err:
        log.warning(f'Report sending failed: {err}. Trying a fallback method.')
        data.update(dict(fallback_reason=str(err), source=source, version=version, platform=platform))
        try:
            send_via_email(data)
        except Exception as err:  # pragma: no cover TODO
            log.warning(f'Report sending via mail failed: {err}')


def save_as_file(data: Dict[str, Any]) -> None:
    report_name = os.path.join(config.BASE_DIR, 'reports', 'report.txt')
    timestamp = datetime.now().strftime('%Y-%m-%d-%H:%M:%S')
    data['timestamp'] = timestamp
    atomic_write(report_name, json.dumps(data, ensure_ascii=False), ensure=True)
    log.info(f'Report saved as `{report_name}`')


def send_via_email(data: Dict[str, Any]) -> None:
    from email.mime.text import MIMEText
    from subprocess import Popen, PIPE

    message = MIMEText(str(data))
    message['Subject'] = 'Eportal server stats for {0}:{1}'.format(
        data.pop('username'),
        data.pop('hostname'),
    )

    message['From'] = 'eportal@kernelcare.com'
    message['To'] = 'eportal.stats@cloudlinux.com'
    p = Popen(['/usr/sbin/sendmail', '-t', '-oi'], stdin=PIPE)
    p.communicate(message.as_bytes())


def _value_or_null(title: str, value: Any) -> Any:
    # previous behaviour kept
    if title in ('servers', 'active_servers'):
        return value
    return value or 'null'


def _get_tbl_tuxcare_usage() -> Table:
    # TODO: required, excluded is too fragile and not backward compatible, should be explicit list of
    # products
    return models.get_statistics(
        models.DIMENSION_TOTAL,
        products_required=(Product.TUXCARE_ESU,),
    ).rename(
        count='servers',
    )


def _get_tbl_kcare_versions(active: bool) -> Table:
    return models.get_statistics(models.DIMENSION_VERSIONS, active=active)


def _get_tbl_kernel_usage(active: bool) -> Table:
    # "Kernel" products only, exclude any of "AlmaCare", which includes "Kernel" as well
    return models.get_statistics(
        models.DIMENSION_USAGE,
        active=active,
        products_excluded=(Product.TUXCARE_ESU,),
    ).rename(
        **{
            models.KcServer.patchset.name: 'releases',
            models.KcServer.kernel_id.name: 'kernel_ids',
        },
    )


def _get_tbl_userspace_total(active: bool) -> Table:
    return models.get_userspace_statistics(models.DIMENSION_TOTAL, active=active)


def _get_tbl_userspace_usage(active: bool) -> Table:
    return models.get_userspace_statistics(
        models.DIMENSION_USAGE,
        active=active,
    ).rename(
        **{
            models.KcUserspaceCheckins.product.name: 'products',
            models.KcUserspaceCheckins.patchset.name: 'releases',
            models.KcUserspaceCheckins.build_id.name: 'build_ids',
        },
    )


def _get_kcare_distros_report_data() -> dict:
    return (
        # TODO KPT-4085 maybe we should move `get_statistics` and `utils.Table` to this module
        models.get_statistics(models.DIMENSION_DISTROS)
        .rename(count='servers')
        .join(
            models.get_statistics(models.DIMENSION_DISTROS, active=True).rename(count='active_servers'),
            default_map={'servers': 0, 'active_servers': 0},
        )
        .map(_value_or_null)
        .rotate(
            pivot='distro',
            ordering=('servers', 'active_servers'),
        )
        .as_dict()
    )


def get_data_report(username: str, hostname: str) -> Dict[str, Any]:
    models.cleanup_userspace_checkins()
    all_servers = models.server_statistics(active=False)
    active_servers = models.server_statistics(active=True)
    eportal_os_name = distro.linux_distribution(full_distribution_name=False)[0:2]

    tuxcare_esu_usage: int

    if config.REDUCED_REPORT:
        # sha1 hash is not used for security here.
        sha1_hash = hashlib.sha1(bstr(hostname)).hexdigest()  # nosec
        hostname = 'disabled-{0}'.format(sha1_hash)
        tuxcare_esu_usage = 0
        usage = {}
        kcare_versions = {}
        userspace_usage = {}
        userspace_total = {}
    else:
        # yes Table doesn't help at all ))
        tuxcare_esu_usage = _get_tbl_tuxcare_usage().map(_value_or_null).as_dict().get('servers', [0])[0]

        # TODO KPT-4085 extract such xforms to separate functions so we can test each xform
        kcare_versions = (
            _get_tbl_kcare_versions(active=False)
            .rename(count='servers')
            .join(
                _get_tbl_kcare_versions(active=True).rename(count='active_servers'),
                default_map={'servers': 0, 'active_servers': 0},
            )
            .map(_value_or_null)
            .rotate(
                pivot=models.KcServer.kcare_version.name,
                ordering=('servers', 'active_servers'),
            )
            .as_dict()
        )

        usage = (
            _get_tbl_kernel_usage(active=False)
            .rename(count='servers')
            .join(
                _get_tbl_kernel_usage(active=True).rename(count='active_servers'),
                default_map={'servers': 0, 'active_servers': 0},
            )
            .map(_value_or_null)
            .as_dict()
        )

        userspace_total = (
            _get_tbl_userspace_total(active=False)
            .rename(count='servers')
            .join(
                _get_tbl_userspace_total(
                    active=True,
                ).rename(count='active_servers'),
                default_map={'servers': 0, 'active_servers': 0},
            )
            .map(_value_or_null)
            .rotate(
                pivot=models.KcUserspaceCheckins.product.name,
                ordering=('servers', 'active_servers'),
            )
            .as_dict()
        )

        userspace_usage = (
            _get_tbl_userspace_usage(active=False)
            .rename(count='servers')
            .join(
                _get_tbl_userspace_usage(active=True).rename(count='active_servers'),
                default_map={'servers': 0, 'active_servers': 0},
            )
            .map(_value_or_null)
            .as_dict()
        )

    try:
        instance_id = get_instance_id()
    except Exception as e:
        instance_id = f'error: {e}'

    feed_count = len(models.Feed.get_all_names(0))
    key_count = models.get_number_of_keys()

    data = {
        'all_servers': all_servers,
        'active_servers': active_servers,
        'eportal_os_name': f"{eportal_os_name[0]}-{eportal_os_name[1]}",
        'usage': usage,
        'kcare_versions': kcare_versions,
        'kcare_distros': _get_kcare_distros_report_data(),
        'username': username,
        'hostname': hostname,
        'leader_instance_id': config.LEADER_URL and oplog.get_node_instance_id(oplog.node_from_url(config.LEADER_URL)),
        'instance_id': instance_id,
        'feed_count': feed_count,
        'autoupdated_feed_count': models.Feed.autoupdated_count(),
        'delayed_feed_count': models.Feed.delayed_count(),
        'key_count': key_count,
        'userspace_total': userspace_total,
        'userspace_usage': userspace_usage,
        'tuxcare_esu_usage': tuxcare_esu_usage,
    }

    return data
