import os
import logging
import hashlib
import tempfile
from subprocess import check_call
from typing import Optional
import config
from eportal import utils, audit, patchsource

log = logging.getLogger('eportal.signature')


def _get_signature_archive_marker_path(release_name: str, sha256: str) -> str:
    return os.path.join(config.RESOURCES_OVERRIDE_DIR, f'{release_name}-{sha256}')


def is_signature_archive_unpacked(release_name: str, sha256: str) -> bool:
    return os.path.exists(_get_signature_archive_marker_path(release_name, sha256))


def unpack_signature_archive(archive: str, user: str = 'UNKNOWN', sha256: Optional[str] = None) -> bool:
    release_name = utils.get_release_name(archive)

    if sha256 is None:
        with open(archive, 'rb') as f:
            sha256 = hashlib.sha256(f.read()).hexdigest()

    if is_signature_archive_unpacked(release_name, sha256):
        return False

    os.makedirs(config.RESOURCES_OVERRIDE_DIR, exist_ok=True)
    check_call(['/bin/tar', '--no-same-owner', '-xf', archive, '-C', config.RESOURCES_OVERRIDE_DIR])

    msg = f'update signature {release_name}'
    audit.add_record(0, user, audit.ACTION_UPDATE_SIGNATURE, msg)
    utils.atomic_write(_get_signature_archive_marker_path(release_name, sha256), b'')
    return True


def update_signatures(user: str = "UNKNOWN") -> None:
    ps = patchsource.PatchSource()
    if not ps.is_configured():
        log.warning('Skip signatures update, patchsource is not configured')
        return

    log.info('Updating signatures: started...')
    url = ps.get_url('signatures.json')
    with utils.get_session() as s:
        response = s.get(url, **utils.get_req_args(auth=ps.auth, timeout=15))

    response.raise_for_status()

    for item in response.json()['signatures']:
        if is_signature_archive_unpacked(item['name'], item['sha256']):
            log.debug('signature for %r already downloaded, skipping', item['name'])
            continue

        with tempfile.TemporaryDirectory() as tmp_dirname:
            file_name = item['name'] + '.tar.bz2'

            downloaded_file_path = os.path.join(tmp_dirname, file_name)

            log.debug('downloading %r to %r', file_name, downloaded_file_path)
            ps.download(
                file_name,
                destination=downloaded_file_path,
                expected_checksum=item['sha256'],
            )

            log.info('extracting %r', downloaded_file_path)
            try:
                unpack_signature_archive(downloaded_file_path, user=user, sha256=item['sha256'])
            except Exception:  # pragma: no cover TODO
                log.exception('failed to unpack signature file %r', item['name'])
