import json
import time
import logging
from typing import cast, Any, Callable, Dict, Optional, Union, TypeVar, overload, Literal

from sqlalchemy import String, Integer, Float
from sqlalchemy.orm.decl_api import Mapped, mapped_column
from sqlalchemy.orm.decl_api import DeclarativeBase
from typing_extensions import ParamSpec, Concatenate

from eportal.dbutils import make_engine, transaction
from eportal.utils import TaskCallable, ContextTaskCallable, TaskOpts
import config

engine, session = make_engine(config.SPOOLER_DATABASE_URI)

T = TypeVar('T')
R = TypeVar('R', covariant=True)
P = ParamSpec('P')


class Base(DeclarativeBase):
    pass


IDLE = 'idle'
IN_PROGRESS = 'in-progress'

log = logging.getLogger('eportal.spooler')


class Task(Base):
    __tablename__ = 'tasks'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    # key: can be used to replace existing idle task with a new one with the same key
    key: Mapped[Optional[str]] = mapped_column(String, nullable=True, index=True)
    # at: unix timestamp when start a task
    at: Mapped[float] = mapped_column(Float, nullable=False, index=True)
    # status: task status, idle (can be picked for execution) or in-progress
    status: Mapped[str] = mapped_column(String, nullable=False)
    # task: json string with task data
    task: Mapped[str] = mapped_column(String, nullable=False)


@transaction(session)
def task_push(
    name: str,
    args: Optional[tuple] = None,
    kwargs: Optional[Dict[str, Any]] = None,
    key: Optional[str] = None,
    urgent: bool = False,
    eta: Union[int, float, None] = None,
    now: Union[int, float, None] = None,
) -> None:
    now = now or time.time()
    if urgent:
        now -= 86400

    if eta:
        eta = now + eta

    if key:
        q = session.query_params()
        flt = (q.key == key) & (q.status == IDLE)
        min_at = q.execute(f'SELECT min(at) FROM tasks WHERE {flt} LIMIT 1').scalar()
        if min_at:
            q.execute(f'DELETE FROM tasks WHERE {flt}')
            now = min(min_at, now)

    task = json.dumps({'name': name, 'args': args, 'kwargs': kwargs, 'eta': eta}, ensure_ascii=False)
    q = session.query_params()
    q.execute(f'INSERT INTO tasks {q.VALUES(key=key, at=now, status=IDLE, task=task)}')


@transaction(session)
def task_pop(now: Optional[float] = None) -> Any:
    now = now or time.time()

    q = session.query_params()
    sql = f"SELECT id, task FROM tasks WHERE status = {q/IDLE} and at <= {q/now} ORDER BY at LIMIT 1"
    result = q.execute(sql).first()

    if result:
        task = json.loads(result[1])
        task['id'] = result[0]
        q = session.query_params()
        q.execute(f"UPDATE tasks SET status = {q/IN_PROGRESS} WHERE id = {q/task['id']}")
        return task


@transaction(session)
def task_done(task: Dict[str, Any]) -> None:
    q = session.query_params()
    q.execute(f"DELETE FROM tasks WHERE id = {q/task['id']}")


def task_count() -> int:
    q = session.query_params()
    return cast(int, q.execute('SELECT count(1) FROM tasks').scalar())


@transaction(session)
def task_clean() -> None:
    q = session.query_params()
    q.execute(f"DELETE FROM tasks WHERE status = {q/IN_PROGRESS}")


_tasks: Dict[str, TaskCallable] = {}


@overload
def task(
    name: str, context: Literal[True], keyfn: Optional[Callable] = None
) -> Callable[[Callable[Concatenate[Dict[str, Any], P], R]], ContextTaskCallable[P, R]]: ...  # pragma: no cover


@overload
def task(
    name: str, keyfn: Optional[Callable] = None
) -> Callable[[Callable[P, R]], TaskCallable[P, R]]: ...  # pragma: no cover


def task(  # type: ignore[misc]
    name: str, context: bool = False, keyfn: Optional[Callable] = None
) -> Callable[[Callable[P, R]], TaskCallable[P, R]]:
    def decorator(fn: Callable[P, R]) -> TaskCallable[P, R]:
        def push(
            opts: Optional[TaskOpts],
            *args: P.args,
            **kwargs: P.kwargs,
        ) -> None:
            if opts is None:
                opts = TaskOpts()
            if not opts.key and keyfn:
                opts.key = keyfn(*args, **kwargs)
            task_push(name, args, kwargs, opts.key, urgent=opts.urgent, eta=opts.eta, now=opts.at)

        fn = cast(TaskCallable[P, R], fn)
        _tasks[name] = fn
        fn.push = push  # type: ignore[method-assign]
        fn.provide_context = context
        return fn

    return decorator


def task_runner(exit_if_empty: bool = False, max_tasks: int = 1000) -> None:
    count = 0
    while count < max_tasks:
        task = task_pop()
        if task:
            fn: Optional[TaskCallable] = _tasks.get(task['name'])
            if fn and (not task['eta'] or task['eta'] > time.time()):
                count += 1
                args = task['args'] or ()
                kwargs = task['kwargs'] or {}
                try:
                    if fn.provide_context:
                        fn(task, *args, **kwargs)
                    else:
                        fn(*args, **kwargs)
                except Exception:
                    log.exception(f"Error executing {task['name']}")
            task_done(task)
        else:
            if exit_if_empty:
                break
            time.sleep(0.3)  # pragma: no cover


def _check_mypy() -> None:  # pragma: no cover
    from eportal.utils import adapt_args

    @task('boo')
    @adapt_args(defaults={'some': ''})
    def boo(some: str) -> str:
        return some

    @task('foo', context=True)
    @adapt_args(defaults={'some': ''})
    def foo(task: Dict[str, Any], some: str) -> str:
        return some

    # following code triggers mypy errors intentionally, it allows to catch
    # unused-ignore mypy warnings if task decorator typehints would be broken

    boo(some=10)  # type: ignore[arg-type]
    boo.push(None, some=10)  # type: ignore[arg-type]
    boo()  # type: ignore[call-arg]
    boo.push(None)  # type: ignore[call-arg]

    foo({}, some=10)  # type: ignore[arg-type]
    foo(some='')  # type: ignore[call-arg]
    foo.push(None, some=10)  # type: ignore[arg-type]
    foo({})  # type: ignore[call-arg]
    foo.push(None)  # type: ignore[call-arg]
