import copy
import dataclasses
import enum
import fcntl
import hashlib
import inspect
import itertools
import logging
import os.path
import sched
import shutil
import threading
import time
from subprocess import call
from collections import OrderedDict, deque
from io import BufferedReader
from typing import cast, Any, Type, Callable, Dict, Iterable, Iterator, List
from typing import Optional, Set, Tuple, Union, Generator, TypeVar, Sequence
from typing import Generic, Mapping, Deque
from base64 import urlsafe_b64encode, b64encode
from contextlib import contextmanager
from functools import wraps, partial, lru_cache
from random import getrandbits, uniform
from uuid import uuid4

import requests.adapters
from typing_extensions import Self, Protocol, ParamSpec
from requests.sessions import Session
from werkzeug.wrappers import Request

import config
import eportal

log = logging.getLogger('eportal.utils')

T = TypeVar('T')
CT = TypeVar('CT')
R = TypeVar('R', covariant=True)
P = ParamSpec('P')


class CacheInFileCallable(Protocol):
    fn: Callable

    def __call__(self, *args: Any, **kwargs: Any) -> Any:  # pragma: no cover
        ...


utype = type(u'')

FEATURES = 'filter-old'
# filter-old: ban old clients with bug leading to inability to reload new kmod from patch.
ESTIMATION_UNPACKED_SIZE_COEFFICIENT = 5


def bstr(data: Union[bytes, str], encoding: str = 'latin1') -> bytes:
    if isinstance(data, utype):
        return data.encode(encoding)
    return cast(bytes, data)


def get_ca_bundle() -> str:
    from requests.utils import DEFAULT_CA_BUNDLE_PATH

    result = DEFAULT_CA_BUNDLE_PATH
    if config.CA_BUNDLE:
        result = config.CA_BUNDLE
    elif not config.USE_CERTIFI:  # try to set system certificate bundle
        try:
            import ssl

            paths = ssl.get_default_verify_paths()
        except (ImportError, AttributeError):  # pragma: no cover
            pass
        else:
            if paths.cafile and os.path.isfile(paths.cafile):
                result = paths.cafile
            elif paths.capath and os.path.isdir(paths.capath):
                result = paths.capath
    return cast(str, result)


requests.adapters.DEFAULT_CA_BUNDLE_PATH = get_ca_bundle()


def get_version(data: Optional[str] = None) -> Tuple[str, Optional[str]]:
    data = data or eportal.EPORTAL_VERSION

    if data == '__VERSION__':
        return 'DEV', 'DEV'

    version, _, tail = data.partition('-')
    parts = tail.split('.')

    if len(parts) >= 2:
        release = parts[0]
        platform = parts[1]
        version = '{0}-{1}'.format(version, release)
        return version, platform
    if len(parts) == 1:
        release = parts[0]
        version = '{0}-{1}'.format(version, release)

    return version, None


def get_session() -> Session:
    s = requests.Session()
    ver = get_version()[0]
    s.headers = {
        'User-Agent': 'ePortal/{0}'.format(ver),
        'Eportal-Features': FEATURES,
        'Eportal-Version': ver,
    }
    if config.PROXY:
        s.proxies = {'http': config.PROXY, 'https': config.PROXY}
    return s


def get_req_args(**kwargs: Any) -> Dict[str, Any]:
    if config.CLIENT_CERTIFICATE:  # pragma: no cover
        kwargs.pop('auth', None)
        kwargs['cert'] = config.CLIENT_CERTIFICATE

    return kwargs


def md5(fname: str) -> str:
    with open(fname, 'rb') as f:
        return HashedStream(hashlib.md5, f).consume().hexdigest()


class HashedStream(object):
    def __init__(self, algo: Callable, fp: BufferedReader) -> None:
        try:
            # KPT-1041 FISP kernels support
            self.hsh = algo(usedforsecurity=False)
        except TypeError:  # pragma: no cover
            self.hsh = algo()

        self.fp = fp
        self.length = 0

    def read(self, size: Optional[int] = None) -> bytes:
        data = self.fp.read(size)
        self.length += len(data)
        self.hsh.update(data)
        return data

    def consume(self, bufsize: int = 1048576) -> Self:
        while self.read(bufsize):
            pass
        return self

    def hexdigest(self) -> str:
        return cast(str, self.hsh.hexdigest())


def get_release_name(fname: str) -> str:
    name, _, _ = os.path.basename(fname).partition('.')
    return name


def libcare_level_to_release(patch_level: Union[int, str], postfix: str = '') -> str:
    """
    Convert Libcare patch_level 'YYYYMMDDNN' to release 'UYYYYMMDD_NN'
    """
    if postfix:
        postfix = '-' + postfix
    patch_level = str(patch_level)
    return 'U' + patch_level[:-2] + '_' + patch_level[-2:] + postfix


def unpack(archive: str, destination: str, unpack_cb: Optional[Callable[[], None]] = None) -> None:
    release = get_release_name(archive)
    unpack_done = os.path.join(destination, release, 'unpack.done')
    if os.path.isfile(unpack_done):
        return
    ensure_dir(destination)

    with open(archive):
        pass

    disk_usage = shutil.disk_usage(destination).free
    estimated_unpack_size = os.path.getsize(archive) * ESTIMATION_UNPACKED_SIZE_COEFFICIENT

    if disk_usage < estimated_unpack_size:
        raise Exception(f'Not enough disk space to unpack {release}')

    unpack_cb and unpack_cb()
    if call(['/bin/tar', '--no-same-owner', '-xf', archive, '-C', destination]) != 0:
        raise Exception('Uncompress failed: %s' % archive)
    atomic_write(unpack_done, 'unpacked')


def extract_ip(request: Request) -> str:
    ip_list = request.headers.get('X-Forwarded-For')
    return parse_proxy_header(ip_list) or request.remote_addr or ''


def parse_proxy_header(header_value: Optional[str]) -> Optional[str]:
    return header_value.split(',')[0].strip() if header_value else None


def ensure_dir(path: str) -> None:
    if not os.path.isdir(path):
        os.makedirs(path)


def try_lock_file(file_name: str) -> Optional[int]:
    # `os.open()` is used intentionally. Built-in `open()` returns an object which
    # could be easily GCed (and therefore closed) and hard to track. Integer fd
    # is guaranteed to stay opened until process exit or explicit `os.close()`
    fd = os.open(file_name, os.O_CREAT | os.O_WRONLY)
    try:
        fcntl.lockf(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
        return fd
    except IOError:
        pass

    return None


@contextmanager
def lock(file_name: str, block: bool = False) -> Iterator[bool]:
    # `a` is needed to prevent content of locked file from zeroing
    with open(file_name, 'a') as fp:
        opts = fcntl.LOCK_EX
        if not block:
            opts |= fcntl.LOCK_NB

        error = None
        while True:
            try:
                fcntl.lockf(fp, opts)
            except IOError as e:
                if block and e.errno == 11:  # Resource busy
                    time.sleep(0.1)
                    continue
                error = e
            else:
                error = None
            break

        if error:
            if block:
                raise error
            yield False  # pragma: no cover
        else:
            try:
                yield True
            finally:
                fcntl.lockf(fp, fcntl.LOCK_UN)


def swallow_and_log(fn: Callable) -> Callable:
    @wraps(fn)
    def inner(*args: Any, **kwargs: Any) -> Any:
        try:
            return fn(*args, **kwargs)
        except Exception:
            log.exception('Unhandled exception')

    return inner


def atomic_write(fname: str, content: Union[bytes, str], safe: bool = False, ensure: bool = False) -> None:
    if ensure:
        ensure_dir(os.path.dirname(fname))

    if safe:
        tmp = fname + '.tmp'
    else:
        tmp = fname + '.' + uuid4().hex
    mode = 'wb' if type(content) is bytes else 'w'
    with open(tmp, mode) as f:
        f.write(content)
    os.rename(tmp, fname)


def read_file(fname: str) -> Optional[bytes]:
    if os.path.exists(fname):
        with open(fname, 'rb') as f:
            return f.read()
    return None


@dataclasses.dataclass
class TaskOpts:
    urgent: bool = False
    at: Optional[float] = None
    eta: Optional[float] = None
    key: Optional[str] = None


class TaskCallable(Generic[P, R], Protocol):
    provide_context: bool

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R: ...  # pragma: no cover

    def push(self, opts: Optional[TaskOpts], *args: P.args, **kwargs: P.kwargs) -> None: ...  # pragma: no cover


class ContextTaskCallable(Generic[P, R], Protocol):
    provide_context: bool

    def __call__(self, task: Dict[str, Any], *args: P.args, **kwargs: P.kwargs) -> R: ...  # pragma: no cover

    def push(self, opts: Optional[TaskOpts], *args: P.args, **kwargs: P.kwargs) -> None: ...  # pragma: no cover


class Scheduler:
    def __init__(self, state_dir: str, log: logging.Logger) -> None:
        self.state_dir = state_dir
        self.log = log
        self.scheduler = sched.scheduler(time.time, time.sleep)

    def periodic(self, state_file: str, delay: int, action: Callable) -> None:
        # reschedule current task again to make it work infinitely
        self.scheduler.enter(delay, 1, self.periodic, (state_file, delay, action))
        try:
            try:
                # refresh mtime for the action state file
                # (time when the action launched the last time),
                # see Scheduler.add() below
                atomic_write(state_file, '')
            except OSError:  # pragma: no cover
                pass
            action()
        except Exception as error:
            self.log.exception(error)

    def add_task(self, name: str, delay: int, task: TaskCallable, *args: Any, **kwargs: Any) -> None:
        # eta is used to skip expired tasks
        # (the task may take so long to complete
        # that new same tasks will be scheduled, in this
        # case expired tasks should be skipped
        # to maintain queue clogging, see skip_expired() above)
        self.add(name, delay, partial(task.push, TaskOpts(eta=delay), *args, **kwargs))

    def add(self, action_name: str, delay: int, action: Callable) -> None:
        now = time.time()
        state_file = os.path.join(self.state_dir, action_name + '.sched')
        try:
            ts = os.path.getmtime(state_file)
        except OSError:  # pragma: no cover
            try:
                atomic_write(state_file, '')
            except OSError:  # pragma: no cover
                pass
            ts = now

        # adjust tasks launch in case of ePortal restart/downtime, examples:
        # 1) task is launched each 10 minutes, ePortal was off for 20 minutes,
        # launch this task immediately after restart;
        # 2) task with daily periodicity, the last launch was 20 hours ago,
        # ePortal is restarted, => the next launch will be in 4 hours
        initial_delay = max(0, ts + delay - now)
        self.scheduler.enter(initial_delay, 1, self.periodic, (state_file, delay, action))

    def run(self) -> None:  # pragma: no cover
        self.scheduler.run()


class ContinuousScheduler(sched.scheduler):
    """
    This scheduler can reschedule action immediately if action returns True.
    It can be useful for data polling.
    """

    def __init__(self, logger: logging.Logger) -> None:
        self._queue: List[sched.Event]
        super().__init__()
        self._logger = logger
        self._active_tasks: Set[Any] = set()

    def _periodic(self, tid: Any, delay: int, fn: Callable, args: tuple, kwargs: Dict[str, str]) -> None:
        if tid not in self._active_tasks:
            return
        new_delay = delay
        try:
            if fn(*args, **kwargs):
                new_delay = 0  # there is some data left, we need to reschedule the function as soon as possible
        except Exception:
            self._logger.exception('Error running %s', fn)
        self.enter(new_delay, 1, self._periodic, (tid, delay, fn, args, kwargs))

    def add(self, delay: int, fn: Callable, *args: Any, **kwargs: Any) -> object:
        tid = object()
        self._active_tasks.add(tid)
        self.enter(0, 1, self._periodic, (tid, delay, fn, args, kwargs))
        return tid

    def cancel(self, tid: Any) -> None:
        self._active_tasks.discard(tid)


def become_owner() -> None:  # pragma: no cover
    # Ensure that effective user is the same as owner of config.BASE_DIR.
    # It prevents file permissions changes if various scripts are run under root user.
    if os.geteuid() == 0:
        uid, gid = get_base_owner()
        logging.debug(f'Running under root. Dropping privileges to {uid}:{gid}')
        os.setgid(gid)
        os.setuid(uid)


def get_base_owner() -> Tuple[int, int]:  # pragma: no cover
    st = os.stat(config.BASE_DIR)
    return st.st_uid, st.st_gid


def try_file(fname: str, expire: Optional[int] = None, hash_args: Iterable[Any] = []) -> Tuple[str, Optional[bytes]]:
    if hash_args:
        hsh = hashlib.sha256(''.join(map(str, hash_args)).encode()).hexdigest()[:10]
        fname += '.' + hsh

    if os.path.isfile(fname):
        valid = True
        if expire:
            now = time.time()
            ts = os.path.getctime(fname)
            valid = ts + expire > now and ts < now + 3600
        if valid:
            return fname, read_file(fname)

    return fname, None


def iter_chunks(it: Iterable, size: int) -> Iterator[List[str]]:
    it = iter(it)
    while True:
        chunk = list(itertools.islice(it, size))
        if chunk:
            yield chunk
        else:  # pragma: no cover  python optimizer can hide this block from coverage
            break


def error_rate_limit(seconds: Union[float, int], error_value: Optional[Any] = None) -> Callable:
    def decorator(fn: Callable) -> Callable:
        last_errors: Dict[Type, float] = {}

        @wraps(fn)
        def inner(*args: Any, **kwargs: Any) -> Any:
            try:
                return fn(*args, **kwargs)
            except Exception as e:
                now = time.time()
                key = type(e)
                if now - last_errors.get(key, 0) > seconds:
                    last_errors[key] = now
                    raise e
                return error_value

        return inner

    return decorator


def random_id() -> str:
    """
    Gets random 16 byte string ordered by time
    """
    value = (int(time.time()) << 56) + getrandbits(56)
    return urlsafe_b64encode(value.to_bytes(12, 'big')).decode()


def basic_auth_headers(user: str, pwd: str) -> Dict[str, str]:
    based = b64encode(f'{user}:{pwd}'.encode()).decode()
    return {'Authorization': f'Basic {based}'}


def age_t(value: str) -> int:
    if value.endswith('h'):
        return 3600 * int(value[:-1])
    return 86400 * int(value)


@lru_cache(None)
def get_instance_id() -> str:
    fname = os.path.join(config.BASE_DIR, 'instance-id')
    instance_id = (read_file(fname) or b'').decode()
    if instance_id:
        return instance_id.strip()

    instance_id = random_id()
    atomic_write(fname, instance_id, safe=True)
    return instance_id


def lru_cache_with_expire(count: int, expiration: int) -> Callable:
    def decorator(fn: Callable) -> Callable:
        @lru_cache(count)
        def cached(_cache_bin: Any, *args: Any, **kwargs: Any) -> Any:
            return fn(*args, **kwargs)

        @wraps(fn)
        def inner(*args: Any, **kwargs: Any) -> Any:
            return cached(int(time.time()) // expiration, *args, **kwargs)

        inner.__wrapped__ = fn
        inner.cache_clear = cached.cache_clear
        return inner

    return decorator


def wrap_if(predicate: Callable, wrapper: Callable) -> Callable:
    def decorator(fn: Callable) -> Callable:
        @wraps(fn)
        def inner(*args: Any, **kwargs: Any) -> Any:
            if predicate():
                return wrapper(fn)(*args, **kwargs)
            return fn(*args, **kwargs)

        return inner

    return decorator


_T_Table = TypeVar('_T_Table', bound='Table')


# TODO: replace with more simple solution I don't think anybody understands
# what's going on here
class Table:
    """
    Designed to keep query results with their field names
    and to be a suitable metaphor of merge several query results for reports
    """

    def __init__(self, titles: Sequence[str], rows: Sequence[Sequence[Any]] = ()):
        if not all(len(titles) == len(row) for row in rows):
            raise ValueError('Every row must contain elements for every title')

        self._values_by_title_map: Dict[str, List[Any]] = {
            title: [row[index] for row in rows] for index, title in enumerate(titles)
        }

    def as_dict(self) -> Dict[str, List[Any]]:
        if not self:
            # previous behaviour kept: skip if titles present but rows empty
            return {}
        return copy.deepcopy(self._values_by_title_map)

    def join(self: _T_Table, other: 'Table', default_map: Dict[str, Any]) -> _T_Table:
        """Implements Outer Full Join by common titles"""
        titles_common = tuple(self._titles.intersection(other._titles))

        row_by_vals_common_map_other: Dict[Tuple[Any, ...], Dict[str, Any]] = OrderedDict(
            (self._pick_values(row, columns=titles_common), row) for row in other._iter_rows()
        )

        joined = type(self)(titles=tuple(self._titles.union(other._titles)))
        for row_self in self._iter_rows():
            vals_common = self._pick_values(row_self, columns=titles_common)
            row_defaults = row_by_vals_common_map_other.pop(vals_common, default_map)
            joined._add_row(row_self, default_map=row_defaults)
        for row_other in row_by_vals_common_map_other.values():
            joined._add_row(row_other, default_map=default_map)
        return joined

    def map(self: _T_Table, func: Callable[[str, Any], Tuple[str, Any]]) -> _T_Table:
        for title, values in self._values_by_title_map.items():
            for idx, val in enumerate(values):
                val_new = func(title, val)
                values[idx] = val_new
        return self

    def rename(self: _T_Table, **new_titles_map: str) -> _T_Table:
        if not self._titles.issuperset(new_titles_map.keys()):
            # don't skip silently to fail fast
            unexpected_original_titles = set(new_titles_map.keys()) - self._titles
            raise ValueError(
                f'Invalid rename map, got unexpected original names: '
                f'{", ".join(str(n) for n in unexpected_original_titles)}'
            )
        self._values_by_title_map = {
            new_titles_map.get(title, title): values for title, values in self._values_by_title_map.items()
        }
        return self

    def rotate(self: _T_Table, pivot: str, ordering: Tuple[str, ...]) -> _T_Table:
        """
        Every value of column specified as a pivot becomes a new title,
        other columns in a row becomes a values for that title in a specified order.

        It may be used to show the "distribution" of values for specified "pivot" column.
        """
        titles_missed = self._titles - {pivot, *ordering}
        if titles_missed:
            raise ValueError(
                'Every title must be set as a "pivot" or listed in an "ordering", '
                f'but {", ".join(titles_missed)} missed'
            )

        titles_new: List[Any] = [row[pivot] for row in self._iter_rows()]
        rows_new: List[List[Any]] = [[] for _ in ordering]
        for row in self._iter_rows():
            for row_new, title in zip(rows_new, ordering):
                row_new.append(row[title])
        return type(self)(titles=titles_new, rows=rows_new)

    def _add_row(self, row: Dict[str, Any], default_map: Dict[str, Any]) -> None:
        for title in self._titles:
            if title in row:
                value = row[title]
            else:
                value = default_map[title]
            self._values_by_title_map[title].append(value)

    @staticmethod
    def _pick_values(row: Dict[str, Any], columns: Tuple[str, ...]) -> Tuple[Any, ...]:
        return tuple(row[column] for column in columns)

    def _iter_rows(self) -> Generator[Dict[str, Any], None, None]:
        for row_idx in range(self._rows_count):
            yield {title: values[row_idx] for title, values in self._values_by_title_map.items()}

    @property
    def _rows_count(self) -> int:
        for values in self._values_by_title_map.values():
            return len(values)
        return 0

    @property
    def _titles(self) -> Set[str]:
        return set(self._values_by_title_map.keys())

    def __bool__(self) -> bool:
        return self._rows_count > 0

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Table):
            raise TypeError(f'"{Table.__name__}" instance expected, got "{type(other).__name__}"')
        return self._values_by_title_map == other._values_by_title_map

    def __repr__(self) -> str:
        # for better output in test assertions
        titles: List[str] = []
        rows: List[List[Any]] = [[] for _ in self._titles]
        for title, values in sorted(self._values_by_title_map.items()):
            titles.append(title)
            for row, value in zip(rows, values):
                row.append(value)
        return f'{type(self).__name__}(titles={titles}, rows={rows})'

    def __str__(self) -> str:
        return '|'.join(sorted(self._titles))


class cached_cls_property(Generic[T]):
    def __init__(self, fn: Callable[..., T]):
        self.fn = fn

    def __get__(self, obj: Any, cls: Type[CT]) -> T:
        value = self.fn(cls)
        setattr(cls, self.fn.__name__, value)
        return value


class EnumEx(enum.Enum):
    """enum with additional custom attributes"""

    value: str
    title: str
    _value_: str

    def __new__(cls, value: str, title: str = '', *other: str) -> Self:
        obj = object.__new__(cls)
        obj._value_ = value
        obj.title = title
        assert len(other) == len(
            vars(cls).get('__annotations__', {})
        ), 'Insufficient enum constructor args'  # nosec assert_used
        for attr, value in zip(cls.__annotations__.keys(), other):
            setattr(obj, attr, value)
        return obj

    def __str__(self) -> str:
        return self._value_

    @cached_cls_property
    def items(cls: Type[CT]) -> Mapping[str, CT]:  # type: ignore[misc]
        return {it._value_: it for it in cls}  # type: ignore[attr-defined]

    @cached_cls_property
    def titles(cls: Type[CT]) -> Mapping[str, str]:  # type: ignore[misc]
        return {it._value_: it.title for it in cls}  # type: ignore[attr-defined]


def run_in_loop(stop: threading.Event, fn: Callable, error_message: str) -> None:
    while not stop.is_set():
        try:
            fn(stop)
        except Exception:
            log.exception(error_message)
            stop.wait(1)
    _variable_to_check_coverage_of_loop_exit = 42  # noqa: F841


class BatchProcessor(Generic[T]):
    def __init__(self, timeout: float = 5.0, batch_size: int = 200, error_callback: Optional[Callable] = None):
        self.queue: Deque[T] = deque()
        self.timeout = timeout
        self.batch_size = batch_size
        self.error_callback = error_callback

        self.append = self.queue.append
        self.clear = self.queue.clear

    def __call__(self, fn: Callable[[List[T]], None]) -> Self:
        self.fn = fn
        return self

    def tick(self, stop: Optional[threading.Event]) -> None:
        batch = popleft_n(self.queue, self.batch_size)
        full_batch = len(batch) >= self.batch_size
        if batch:
            try:
                self.fn(batch)
            except Exception:
                if self.error_callback:
                    self.error_callback(batch)
                raise

        if not full_batch:
            # Suppress Bandit warning as random.uniform is not used for security here.
            stop and stop.wait(uniform(self.timeout, self.timeout + 1))  # nosec B311
        else:
            stop and stop.wait(0.01)  # give a chance to other processes to get a db lock

    def start(self, stop: threading.Event, error_msg: str = 'error') -> threading.Thread:
        t = threading.Thread(target=swallow_and_log(run_in_loop), args=(stop, self.tick, error_msg), daemon=True)
        t.start()
        return t


def popleft_n(queue: deque, n: int) -> list:
    if not queue:
        return []

    result = []
    for _ in range(n):
        try:
            result.append(queue.popleft())
        except IndexError:
            break

    return result


def process_int(res: Dict[str, Any], label: str, default: Optional[int] = -1) -> None:
    try:
        res[label] = int(res[label])
    except (TypeError, ValueError, KeyError):
        res[label] = default


Adapter = Callable[[Dict[str, Any]], None]


def adapt_args(
    *adapters: Adapter, rename: Optional[Dict[str, str]] = None, defaults: Optional[Dict[str, Any]] = None
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """A common way to add a new argument into task/replicated action is to add it with a default value.
    It effectively prevents type hints warnings and there is a history of cases when 100% test coverage
    could not point to the issue.

    adapt_args is a decorator allowing to add arguments without a default value in signature or rename arguments.

    processing order: adapters, renames, defaults.

    :param adapters: a list of generic adapters
    :param rename: a dict or renames oldname -> newname
    :param defaults: a dict of default values
    """
    ladapters = list(adapters)
    if rename:
        ladapters.append(rename_adapter(**rename))
    if defaults:
        ladapters.append(default_adapter(**defaults))

    def decorator(fn: Callable[P, R]) -> Callable[P, R]:
        arg_names = inspect.getfullargspec(fn).args

        @wraps(fn)
        def inner(*args: P.args, **kwargs: P.kwargs) -> R:
            for i, v in enumerate(args):
                kwargs[arg_names[i]] = v
            for a in ladapters:
                a(kwargs)
            return fn(**kwargs)  # type: ignore[call-arg]

        return inner

    return decorator


def default_adapter(**kwargs: Any) -> Adapter:
    def inner(data: Dict[str, Any]) -> None:
        for k, v in kwargs.items():
            if k not in data:
                data[k] = v

    return inner


def rename_adapter(**kwargs: str) -> Adapter:
    def inner(data: Dict[str, Any]) -> None:
        for fromname, toname in kwargs.items():
            if fromname in data and toname not in data:
                data[toname] = data.pop(fromname)

    return inner


def instance_lock() -> Optional[int]:
    return try_lock_file(os.path.join(config.BASE_DIR, 'instance.lock'))
