import os
import logging
import threading
from functools import lru_cache
from typing import Any, Tuple, Optional, List

from flask import Flask, render_template, request
from flask_login import LoginManager
from flask_moment import Moment
from werkzeug.routing import Rule

import config

from eportal import db, checkins
from eportal.web import auth, units
from eportal.utils import instance_lock

# check database before starting app
with db.engine.connect():
    pass

app = Flask(__name__)
app.logger = logging.getLogger(app.name)
app.config.from_object(config)
app.url_map.strict_slashes = False
app.url_map.redirect_defaults = False
app.teardown_request(db.remove_session)
app.wsgi_app = units.unit_middleware(app.wsgi_app)  # type: ignore[method-assign,no-untyped-call]
app.session_interface = units.UnitSessionInteface()

login_manager = LoginManager(app)
login_manager.session_protection = 'strong'
login_manager.login_view = 'login'
login_manager.user_loader(auth.load_user)
login_manager.request_loader(auth.load_user_from_request)
login_manager.anonymous_user = auth.EportalAnonumousUser


@lru_cache(None)
def static_render(name: str, **kwargs: Any) -> str:
    return render_template(name, **kwargs)


@app.errorhandler(404)
def page_not_found(e: Exception) -> Tuple[str, int]:
    return static_render('404.html'), 404


@app.errorhandler(500)
def internal_server_error(e: Exception) -> Tuple[str, int]:
    return static_render('500.html'), 500  # pragma: no cover


@app.before_request
def check_for_allowed_host() -> Optional[Tuple[str, int]]:
    host = request.host
    if config.ALLOWED_HOSTS and host not in config.ALLOWED_HOSTS:
        msg = "Host not in a list of allowed hosts."
        app.logger.error(msg)
        return (
            static_render(
                '400.html', error_msg=msg + ' See the documentation https://docs.tuxcare.com/eportal/#allowed-hosts'
            ),
            400,
        )
    return None


Moment(app)

import eportal.web.views


def get_all_rules() -> List[Rule]:  # pragma no cover
    all_rules = app.url_map.iter_rules()
    rule_list = [route for route in all_rules]
    return rule_list


if config.PUSH_METRICS:  # pragma: no cover
    from eportal import metrics

    stop_statd_pusher = threading.Event()
    metrics.statsd_pusher(stop_statd_pusher)


try:  # pragma: no cover
    # Take an instance lock in the context of uwsgi worker, it doesn't matter
    # which particular worker takes the lock.
    import uwsgi

    instance_lock()
except ImportError:
    pass
