import gzip
import base64
import logging
from functools import wraps
from typing import cast, Any, Callable, Optional, TYPE_CHECKING, Union
from functools import lru_cache
from defusedxml import ElementTree

from flask import session, abort, request, jsonify, current_app, Request
from flask_login import login_user, current_user, config as flask_login_config
from flask_login.mixins import AnonymousUserMixin  # type: ignore[import-untyped]

import config
from eportal import db, ldaplib, replicated, audit, dynconfig, perms
from eportal.web import units
from eportal.db.models import User, get_verified_user, get_user, UnitUser

if TYPE_CHECKING:  # pragma: no cover
    # We use defusedxml instead of xml for security, import Element just for typing.
    from xml.etree.ElementTree import Element  # nosec B405


log = logging.getLogger('auth')


class EportalAnonumousUser(AnonymousUserMixin):  # type: ignore[misc] # pragma: no cover
    permissions: perms.PermSet = set()


def ensure_permission(permission: str, json: bool = False) -> None:
    if not perms.user_has_permission(permission):
        if json:
            resp = jsonify({'error': 'forbidden'})
            resp.status_code = 403
            abort(resp)
        else:
            abort(403)


def set_auth_session(user: User, kind: str) -> None:
    login_user(user)
    replicated.audit_log(units.current_unit_id(), user.username, audit.ACTION_LOGIN, f'logged in via {kind}')
    if config.AUTH_REFRESH_SESSION:
        session.permanent = True  # pragma: no cover


def db_login(username: str, password: str) -> bool:
    user = get_user(username, unit_id=units.current_unit_id())
    if user is not None and user.verify_password(password):
        set_auth_session(user, 'ePortal')
        return True
    return False


def ldap_login(username: str, password: str) -> bool:
    conf = dynconfig.Config().ldap
    if not conf['url'] or not conf['bind_template']:
        return False

    # unlike connection_string `%s` in filter_template is not obligatory
    flt = conf['filter_template']
    flt = flt if '%s' not in flt else flt % username

    if not ldaplib.login(conf['url'], flt, conf['bind_template'] % username, password):
        return False

    user = get_or_create_user(username, 'LDAP')
    set_auth_session(user, 'LDAP')
    return True


def get_or_create_user(username: str, description: str, role: str = 'readonly') -> User:
    unit_id = units.current_unit_id()
    user, created = replicated.get_or_create_user(
        unit_id=unit_id, username=username, description=description, role=role
    )

    if created:
        replicated.audit_log(unit_id, username, audit.ACTION_MANAGE_USER, f'created user via {description}')

    return user


def load_user(user_id: Optional[str]) -> Union[User, UnitUser, None]:
    if user_id:
        if unit_id := units.current_unit_id():
            return db.session.query(UnitUser).filter_by(id=int(user_id), unit_id=unit_id).first()
        else:
            return db.session.get(User, int(user_id))
    return None  # pragma: no cover


def load_user_from_basic_auth(request: Request) -> Optional[User]:
    auth = request.authorization
    if not auth:
        return None

    username = auth.username
    password = auth.password

    if username and password:
        user = get_verified_user(username, password, units.current_unit_id())
        return user

    return None  # pragma: no cover


def load_user_from_request(request: Request) -> Optional[User]:
    return load_user_from_basic_auth(request) or load_user_from_webseal(request)


def xml_find(elem: Optional["Element"], *args: Any) -> Optional["Element"]:
    if elem is not None:
        return elem.find(*args)
    return None  # pragma: no cover


def xml_text(elem: Optional["Element"]) -> Optional[str]:
    if elem is not None:
        return cast(str, elem.text).strip()
    return None  # pragma: no cover


@lru_cache(100)
def parse_webseal_token(token: str) -> Optional[str]:
    try:
        data = gzip.decompress(base64.b64decode(token.encode()))
        root = ElementTree.fromstring(data)
        subj = xml_find(root, '{urn:oasis:names:tc:SAML:2.0:assertion}Subject')
        nameid = xml_find(subj, '{urn:oasis:names:tc:SAML:2.0:assertion}NameID')
        return xml_text(nameid)
    except Exception as e:
        log.error('Invalid WebSeal token: %s', e)
        return None


def load_user_from_webseal(request: Request) -> Optional[User]:
    if not config.WEBSEAL_AUTH:
        return None

    token = request.headers.get(config.WEBSEAL_AUTH_HEADER)
    if not token:
        return None

    userid = parse_webseal_token(token)
    if userid:
        return get_or_create_user(userid, 'WebSeal', config.WEBSEAL_AUTH_DEFAULT_ROLE)
    return None  # pragma: no cover


def login_required_with_json_error(func: Callable) -> Callable:
    """
    this decorator is pretty much the same as login_required
    but returns JSON response if auth failed
    """

    @wraps(func)
    def decorated_view(*args: Any, **kwargs: Any) -> Any:
        if request.method in flask_login_config.EXEMPT_METHODS:
            return func(*args, **kwargs)
        elif current_app.config.get('LOGIN_DISABLED'):  # pragma: no cover
            return func(*args, **kwargs)
        elif not current_user.is_authenticated:
            return jsonify({'error': 'unauthorized'}), 401
        return func(*args, **kwargs)

    return decorated_view
