from __future__ import annotations

from typing import Any, Optional, cast

from flask import request, sessions, abort
from flask.globals import request_ctx

import config
from eportal.web import utils as web_utils
from eportal.db.models import Unit


def _update_environ(environ: dict[str, Any]) -> None:
    static_resources = ('static', 'js', 'css', 'img')
    dynamic_resources = ('admin', 'index.html')

    parts = environ['PATH_INFO'].split('/')

    if len(parts) > 2 and parts[1] and parts[1] != 'admin' and parts[2] in (dynamic_resources + static_resources):
        unit = parts[1]
        environ['PATH_INFO'] = environ['PATH_INFO'][1 + len(unit) :]
        environ['REQUEST_URI'] = environ['REQUEST_URI'][1 + len(unit) :]

        if unit == '_':
            # a special slug to distinguish super-admin selection in drop-down menu
            # for frontend only, not used in the backend
            return

        if parts[2] in dynamic_resources:
            environ['SCRIPT_NAME'] = f'/{unit}'
            environ['unit_slug'] = unit


def unit_middleware(app):
    def wsgi(environ, start_response):
        if config.BUNITS:
            _update_environ(environ)

        return app(environ, start_response)

    return wsgi


class ForcedSecureCookieSessionInterface(sessions.SecureCookieSessionInterface):
    def get_cookie_secure(self, app):
        return super().get_cookie_secure(app) or request.headers.get('X-Forwarded-Proto') == 'https'


class UnitSession(sessions.SecureCookieSession):
    def __init__(self, unit, session):
        self.unit = unit
        self.session = session
        self.is_unit_session = True

    def __getitem__(self, name):  # pragma: no cover
        name = f'{self.unit}-{name}'
        return self.session[name]

    def get(self, name, default=None):
        name = f'{self.unit}-{name}'
        return self.session.get(name, default)

    def pop(self, name, default=None):
        name = f'{self.unit}-{name}'
        return self.session.pop(name, default)

    def __setitem__(self, name, value):
        name = f'{self.unit}-{name}'
        self.session[name] = value

    def __contains__(self, name):
        name = f'{self.unit}-{name}'
        return name in self.session

    @property
    def permanent(self):  # pragma: no cover
        return self.session.permanent

    @permanent.setter
    def permanent(self, value: bool) -> None:  # pragma: no cover
        self.session.permanent = value


class UnitSessionInteface(ForcedSecureCookieSessionInterface):
    def open_session(self, app, request):
        session = super().open_session(app, request)
        if unit_slug := request.environ.get('unit_slug'):
            session = UnitSession(unit_slug, session)
        return session

    def save_session(self, app, session, response):
        if isinstance(session, UnitSession):
            session = session.session
        return super().save_session(app, session, response)


def current_unit() -> Optional[Unit]:
    if hasattr(request, 'unit'):
        return cast(Unit, request.unit)

    if unit_slug := request.environ.get('unit_slug'):
        unit = Unit.by_slug(unit_slug)
        if not unit:  # pragma: no cover
            abort(404)
    else:
        unit = None

    request.unit = unit  # type: ignore[attr-defined]
    return unit


def current_unit_id() -> int:
    unit = current_unit()
    if unit:
        return unit.id
    return 0


def global_session() -> sessions.SessionMixin:
    if getattr(request_ctx.session, 'is_unit_session', False):
        return cast(sessions.SessionMixin, request_ctx.session.session)  # type: ignore[union-attr]
    return cast(sessions.SessionMixin, request_ctx.session)


def set_current_slug(slug: str) -> None:
    """Pretend current request has a different unit"""
    request.environ['unit_slug'] = slug

    try:
        del request.unit  # type: ignore[attr-defined]
    except AttributeError:
        pass

    request_ctx.session = UnitSession(slug, global_session())


titles = web_utils.g_cache(Unit.titles)
slugs = web_utils.g_cache(Unit.slugs)
