import logging
import json
from datetime import datetime, timedelta, timezone
from urllib.parse import quote
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    List,
    Tuple,
    cast,
    Dict,
    Optional,
    Generic,
    TypeVar,
    Union,
    TYPE_CHECKING,
)

from jinja2.runtime import Context
from flask import Response
from flask import url_for, redirect, flash, request
from flask_admin import Admin, AdminIndexView, expose
from flask_admin.form import FormOpts
from flask_admin.form.rules import RuleSet
from flask_admin.contrib.sqla import ModelView
from flask_admin.form.fields import Select2Field
from flask_admin.contrib.sqla.filters import (
    BaseSQLAFilter,
    FilterEqual,
    FilterLike,
    IntGreaterFilter,
    DateEqualFilter,
    DateGreaterFilter,
    DateSmallerFilter,
)
from flask_login import current_user
from jinja2 import pass_context
from markupsafe import Markup
from sqlalchemy import and_, or_, text, func, case
from sqlalchemy.orm import Query, joinedload
from wtforms.fields import SelectField, StringField, PasswordField
from wtforms.validators import NumberRange, DataRequired, NoneOf, Length, Optional as OptionalV
from wtforms.form import Form
from markupsafe import escape

import config
from eportal import db, perms, replicated, audit, patches_constants as pc
from eportal.db.models import (
    DEFAULT_FEED,
    Base,
    User,
    KcKey,
    Feed,
    KcServer,
    Tag,
    get_unique_id,
    Unit,
    UnitUser,
)
from eportal.web import app, login_manager, units
from eportal.audit import AuditLog

logger = logging.getLogger(__name__)
T = TypeVar('T')
UM = TypeVar('UM', bound=Union[KcKey, Feed, AuditLog, UnitUser])
RM = TypeVar('RM', bound=Union[User, KcKey, Feed, Unit, UnitUser])

if not TYPE_CHECKING:

    class ModelView(Generic[T], ModelView):
        pass


def formatter(fn: Callable) -> Callable:
    def fmt(_view: ModelView, _context: Context, model: Base, name: str) -> str:
        value = getattr(model, name)
        return fn(value)  # type: ignore[no-any-return]

    return fmt


def mformatter(fn: Callable) -> Callable:
    def fmt(_view: ModelView, _context: Context, model: Base, name: str) -> str:
        return fn(model)  # type: ignore[no-any-return]

    return fmt


@formatter
def key_name_formatter(value: str) -> Markup:
    # key name is a link to server list connected to this key
    return Markup("<a href='{0}'>{1}</a>".format(url_for('kcserver.index_view', flt_key_equals=value), escape(value)))


@mformatter
def server_limit_formatter(model: 'KcKey') -> str:
    # column shows the relation of registered servers to its key limit
    limit = model.server_limit
    return '{0} / {1}'.format(model.server_count(), limit if limit else 'unlimited')


@formatter
def feed_formatter(value: str) -> str:
    return value or DEFAULT_FEED


@mformatter
def products_formatter(model: 'KcKey') -> str:
    products_list = model.products_list or []
    products_labels = [pc.Product.titles.get(prod, prod) for prod in products_list]
    return ', '.join(products_labels)


@formatter
def channel_formatter(value: str) -> str:
    return pc.Channel.titles.get(value, value)


def filter_formatter(_view: "AuditView", _context: Context, model: AuditLog, name: str) -> Markup:
    value = getattr(model, name)
    url = f'.?flt_{name}_equals=' + quote(value)
    return Markup(f'<a href="{url}">{escape(value)}</a>')


def convert_utc_date_to_local(date: datetime) -> datetime:
    return date.replace(tzinfo=timezone.utc).astimezone(tz=None)


@formatter
def local_date_formatter(value: Optional[datetime]) -> str:
    if value:
        local_date = convert_utc_date_to_local(value)
        return local_date.strftime(config.DATETIME_FORMAT)
    return ''


class Choices:
    def __init__(self, get_names_func: Callable) -> None:
        self.get_names_func = get_names_func

    def __iter__(self) -> Iterable:
        values = list(self.get_names_func())
        if values and isinstance(values[0], tuple):
            return iter(values)

        choices = [(name, name) for name in values]
        return iter(choices)


PermArg = Union[perms.Perm, perms.PermSet]


def make_permissions(
    view: PermArg = perms.NONE, modify: PermArg = perms.NONE, **kwargs: PermArg
) -> Dict[str, perms.PermSet]:
    result = {
        'index_view': perms.to_permset(view),
        'create_view': perms.to_permset(modify),
        'edit_view': perms.to_permset(modify),
        'delete_view': perms.to_permset(modify),
        'details_view': perms.to_permset(view),
        'export': perms.to_permset(view),
    }
    result.update((k, perms.to_permset(v)) for k, v in kwargs.items())
    return result


class BaseView(ModelView[T]):
    permissions = make_permissions()
    _export_columns: List[Tuple[str, str]]

    def is_accessible(self) -> bool:
        return cast(bool, current_user.is_authenticated)

    def inaccessible_callback(self, name: str, **kwargs: Any) -> Response:
        if not current_user.is_authenticated:
            return cast(Response, login_manager.unauthorized())
        return super().inaccessible_callback(name, **kwargs)  # pragma: no cover TODO

    def _handle_view(self, name: str, **kwargs: Any) -> Optional[Response]:
        if not self.has_access(name):
            return self.inaccessible_callback(name, **kwargs)
        return super()._handle_view(name, **kwargs)

    def has_access(self, name: str) -> bool:
        view_permissions = self.permissions.get(name, {perms.NONE})
        return perms.has_permissions(current_user.permissions, view_permissions)

    @property
    def can_create(self) -> bool:
        return self.has_access('create_view')

    @property
    def can_edit(self) -> bool:
        return self.has_access('edit_view')

    @property
    def can_delete(self) -> bool:
        return self.has_access('delete_view')

    def mutate_form(self, model: Optional[T], form: Form, is_created: bool) -> None:
        """Allows to modify form based on request data"""

    def mutate_form_opts(self, form_opts: FormOpts, is_created: bool) -> None:
        """Allows to modify form opts based on request data"""

    def edit_form(self, obj: Optional[T] = None) -> Form:
        form = super().edit_form(obj=obj)
        self.mutate_form(obj, form, False)
        return form

    def create_form(self) -> Form:
        form = super().create_form()
        self.mutate_form(None, form, True)
        return form

    def list_columns_ex(self, columns: List[Tuple[str, str]]) -> List[Tuple[str, str]]:
        """Allows to define list columns based on request data

        columns MUST not be changed
        """
        return columns

    def details_columns_ex(self, model: T, columns: List[Tuple[str, str]]) -> List[Tuple[str, str]]:  # pragma: no cover
        """Allows to define details columns based on request data

        columns MUST not be changed
        """
        return columns

    def export_columns_ex(self, columns: List[Tuple[str, str]]) -> List[Tuple[str, str]]:  # pragma: no cover
        """Allows to define details columns based on request data

        columns MUST not be changed
        """
        return columns

    def render(self, template: str, **kwargs: Any) -> Response:
        ep = cast(str, request.endpoint)
        if ep.endswith('.index_view'):
            kwargs['list_columns'] = self.list_columns_ex(kwargs['list_columns'])
        elif ep.endswith('.details_view'):
            kwargs['details_columns'] = self.details_columns_ex(kwargs['model'], kwargs['details_columns'])
        elif ep.endswith('.edit_view'):
            self.mutate_form_opts(kwargs['form_opts'], is_created=False)
        elif ep.endswith('.create_view'):
            self.mutate_form_opts(kwargs['form_opts'], is_created=True)
        return super().render(template, **kwargs)

    def _export_data(self) -> Tuple[int, List[T]]:
        # flask-admin has no way to override export columns besides
        # rewriting self._export_columns. Here is a nasty trick
        # to save original _export_columns and reuse the value later.
        if not hasattr(self, '_eportal_export_columns'):
            self._eportal_export_columns = self._export_columns
        self._export_columns = self.export_columns_ex(self._eportal_export_columns)
        return super()._export_data()


class Select2FieldMultiple(Select2Field):
    def iter_choices(self) -> Iterator[Tuple[str, str, bool]]:
        """fix check if list item is selected"""
        selected_items = self.data or []
        for value, label in self.choices:
            yield (value, label, value in selected_items)


class UnitBaseView(BaseView[UM]):
    unit_exclude_default = False
    form_overrides: Optional[Dict[str, Any]]
    form_args: Optional[Dict[str, Any]]
    form_edit_rules: Optional[Iterable[str]]
    form_columns: Optional[Iterable[str]]

    def __init__(self, *args: Any, **kwargs: Any):
        form_overrides = self.form_overrides or {}
        form_overrides['unit_id'] = SelectField
        self.form_overrides = form_overrides

        form_args = self.form_args or {}
        form_args['unit_id'] = dict(choices=Choices(self.make_titles), label='Business Unit')
        self.form_args = form_args

        if self.form_edit_rules:
            self.form_edit_rules = ('unit_id',) + tuple(self.form_edit_rules)

        self.form_columns = ('unit_id',) + tuple(self.form_columns or ())
        super().__init__(*args, **kwargs)

    def make_titles(self) -> Iterable[Tuple[int, str]]:
        titles = units.titles()
        if self.unit_exclude_default:
            titles = titles.copy()
            titles.pop(0)
        return titles.items()

    def list_columns_ex(self, columns: List[Tuple[str, str]]) -> List[Tuple[str, str]]:
        if config.BUNITS and not units.current_unit():
            columns = [('unit_id', 'Business Unit')] + columns
        return columns

    @pass_context
    def get_list_value(self, context: Context, model: UM, name: str) -> Any:
        if name == 'unit_id':
            return units.titles().get(model.unit_id, 'UNKNOWN')
        return super().get_list_value(context, model, name)

    def unit_filter_query(self, q: Query) -> Query:
        if (unit_id := units.current_unit_id()) or not config.BUNITS:
            q = q.filter_by(unit_id=unit_id)
        return q

    def get_one(self, id: str) -> Optional[UM]:
        model: Optional[UM] = super().get_one(id)
        if unit_id := units.current_unit_id():
            if model and model.unit_id != unit_id:
                return None
        return model

    def get_query(self) -> Query:
        return self.unit_filter_query(super().get_query())

    def get_count_query(self) -> Query:
        return self.unit_filter_query(super().get_count_query())

    def mutate_form(self, model: Optional[UM], form: Form, is_created: bool) -> None:
        if not config.BUNITS or units.current_unit():
            try:
                del form['unit_id']
            except KeyError:  # pragma: no cover
                pass

    def mutate_form_opts(self, form_opts: FormOpts, is_created: bool) -> None:
        if not config.BUNITS or units.current_unit():
            if form_opts.form_rules:
                rules = [it for it in form_opts.form_rules.rules if it.field_name != 'unit_id']
                form_opts.form_rules = RuleSet(form_opts.form_rules.view, rules)

    def on_model_change(self, form: Form, model: UM, is_created: bool) -> None:
        unit_id = units.current_unit_id()
        if is_created and unit_id:
            model.unit_id = unit_id
        super().on_model_change(form, model, is_created)


password_length_v = Length(min=config.AUTH_PASSWORD_MIN_LENGTH, max=100)


class ReplicatedMixin(Generic[RM]):
    def on_model_delete(self, model: RM) -> None:
        if hasattr(model, 'on_delete'):
            model.on_delete()
        replicated.delete_model(model, user=current_user.username)

    def after_model_change(self, form: Form, model: RM, is_created: bool) -> None:
        replicated.sync_model(model, current_user.username)


class UserView(ReplicatedMixin[User], BaseView[User]):
    permissions = make_permissions(perms.VIEW, perms.MANAGE_USERS)

    def on_model_change(self, form: Form, model: User, is_created: bool) -> None:
        if model.password:
            model.set_password(model.password)
        super().on_model_change(form, model, is_created)

    def mutate_form(self, model: Optional[User], form: Form, is_created: bool) -> None:
        form.password.validators = [DataRequired(), password_length_v]  # type: ignore[attr-defined]

    column_list = ('username', 'description', 'readonly')
    column_labels = dict(username='Username', description='Description', readonly='Readonly')

    column_exclude_list = ['password_hash']

    column_sortable_list = ('username', 'description', 'readonly')
    column_searchable_list = ['username']

    form_columns = ('username', 'password', 'readonly', 'description')
    form_extra_fields = dict(password=PasswordField(label='Password', validators=[OptionalV(), password_length_v]))


def feed_names() -> List[str]:
    return Feed.get_all_names(units.current_unit_id())


class KeyView(ReplicatedMixin[KcKey], UnitBaseView[KcKey]):
    permissions = make_permissions(perms.VIEW, perms.MANAGE_KEYS)

    def on_model_change(self, form: Form, model: KcKey, is_created: bool) -> None:
        # allow user to leave key field empty and fill it with autogenerated text
        if not model.key:
            model.key = get_unique_id()

        model.set_products(form.products.raw_data)  # type: ignore[attr-defined]
        model.validate()
        super().on_model_change(form, model, is_created)

    def on_form_prefill(self, form: Form, id: Any) -> None:
        form.products.data = json.loads(form.products.data or '[]')  # type: ignore[attr-defined] # stored as str in DB

    def handle_view_exception(self, exc: Exception) -> bool:
        if isinstance(exc, db.session.IntegrityError):  # type: ignore[attr-defined]
            error_msg = str(exc)

            if 'unique' in error_msg.lower():
                error_msg = "Key already exists"

            flash(error_msg, category='error')
            return True

        return super().handle_view_exception(exc)  # pragma: no cover TODO

    column_list = ('key', 'note', 'server_limit', 'feed', 'products')
    column_labels = dict(
        key='Key',
        feed='Feed',
        note='Description',
        server_limit='Server Count / Limit',
        products='Products',
    )

    column_sortable_list = ('key', 'note', 'feed')
    column_searchable_list = ['key', 'feed', 'note']

    column_formatters = dict(
        key=key_name_formatter,
        server_limit=server_limit_formatter,
        feed=feed_formatter,
        products=products_formatter,
    )

    form_columns = ('key', 'note', 'server_limit', 'feed', 'products')

    # remove all default primary key validators
    form_extra_fields = dict(key=StringField(label='Key', description='Leave empty to get autogenerated key name'))
    form_edit_rules = ('note', 'server_limit', 'feed', 'products')  # do not allow user to change `key` primary key

    form_overrides = dict(feed=SelectField, products=Select2FieldMultiple)

    form_args = dict(
        server_limit=dict(label='Server Limit', validators=[NumberRange(min=0)]),
        feed=dict(choices=Choices(feed_names), coerce=lambda value: value if value != DEFAULT_FEED else None),
        products=dict(render_kw={'multiple': 'multiple'}, choices=list(pc.Product.titles.items())),
    )

    def _show_missing_fields_warning(self, text: str) -> None:
        pass


class FeedView(ReplicatedMixin[Feed], UnitBaseView[Feed]):
    permissions = make_permissions(perms.VIEW, perms.MANAGE_FEEDS)

    list_template = 'delete_feed.html'

    column_list = ('name', 'auto', 'deploy_after', 'channel')
    column_labels = dict(name='Name', auto='Autoupdate', deploy_after='Deploy after X hours')
    column_formatters = dict(
        channel=channel_formatter,
    )

    column_sortable_list = ('name',)
    column_searchable_list = ['name']

    form_columns = ('name', 'auto', 'deploy_after', 'channel')
    form_edit_rules = ('auto', 'deploy_after', 'channel')  # do not allow user to change `name` primary key
    form_overrides = dict(channel=SelectField)
    form_args = dict(
        name=dict(validators=[NoneOf(DEFAULT_FEED, message=f"Feed '{DEFAULT_FEED}' is reserved")]),
        channel=dict(choices=list(pc.Channel.titles.items())),
    )

    def _show_missing_fields_warning(self, text: str) -> None:
        pass


class ServerView(BaseView):
    permissions = make_permissions(perms.VIEW, perms.MANAGE_SERVERS)

    can_edit = False
    can_create = False
    can_view_details = True
    can_export = True

    class FilterInactive(IntGreaterFilter):
        def apply(self, query: Query, value: int, alias: Optional[Any] = None) -> Query:
            since_date = datetime.now() - timedelta(days=value)
            return query.filter(
                or_(
                    KcServer.last < since_date,
                    and_(KcServer.last.is_(None), KcServer.created < since_date),
                )
            )

    class FilterFeed(FilterEqual):
        def apply(self, query: Query, value: Optional[str], alias: Optional[Any] = None) -> Query:
            if value == DEFAULT_FEED:
                value = None
            return query.filter(KcKey.key == KcServer.kc_key, KcKey.feed == value)

    # it's workaround which provides ability to get servers with userspace patches
    # todo: rework it to use combined model kc_servers + kc_ucheckins during refactoring
    class FilterPatchsetEqual(FilterEqual):
        def apply(self, query: Query, value: str, alias: Optional[Any] = None) -> Query:
            if value.startswith('U'):
                last = datetime.now() - timedelta(days=db.models.DAY_DELTA)
                f = '''
                    kc_servers.server_id IN (
                        SELECT server_id from kc_ucheckins WHERE patchset = :patchset
                        AND last >= :last
                    )
                    '''
                result = query.filter(text(f).bindparams(patchset=value, last=last))
            else:
                result = query.filter(KcServer.patchset == value)
            return result

    tag_cache: Dict[Optional[str], dict] = {}

    def get_tag(self, multi_tag: Optional[str], tag: str) -> str:
        try:
            parsed_tag = self.tag_cache[multi_tag]
        except KeyError:
            parsed_tag = self.tag_cache[multi_tag] = Tag.tag_to_map(multi_tag or '')
        return cast(str, parsed_tag.get(tag, ''))

    def on_model_delete(self, model: KcServer) -> None:
        replicated.delete_model(model, user=current_user.username)

    @property
    def _template_args(self) -> Dict[str, int]:
        checking_in_hours = 30 * 24
        in_past_time = datetime.now() - timedelta(hours=checking_in_hours)

        unit_id = units.current_unit_id()

        query = db.session.query(
            func.count(KcServer.server_id),
            func.count(case((KcServer.last >= in_past_time, KcServer.server_id), else_=None)),
        )

        if unit_id:
            query = query.join(KcKey, KcServer.kc_key == KcKey.key).filter(KcKey.unit_id == unit_id)

        total, checked_in = query.one()
        silent = total - checked_in

        return {'total': total, 'checked_in': checked_in, 'silent': silent}

    list_template = 'server_statistics.html'

    column_list = ('remote_ip', 'node', 'mltag', 'euname', 'created', 'last', 'server_id', 'kc_key', 'patchset')

    column_details_list = column_export_list = [
        'remote_ip',
        'node',
        'euname',
        'created',
        'last',
        'server_id',
        'patch_level',
        'kernel_id',
        'distro_version',
        'distro',
        'machine',
        'processor',
        'release',
        'version',
        'kcare_version',
        'virt',
        'uptime',
        'ltimestamp',
        'patch_type',
        'patchset',
        'kc_key',
    ]

    def list_columns_ex(self, columns: List[Tuple[str, str]]) -> List[Tuple[str, str]]:
        if config.BUNITS and not units.current_unit():
            columns = [('unit_id', 'Business Unit')] + columns
        return columns

    def export_columns_ex(self, columns: List[Tuple[str, str]]) -> List[Tuple[str, str]]:
        columns = columns + [('tag_' + tag, f'Tag {tag}') for tag in Tag.get_tags(units.current_unit_id())]
        if config.BUNITS and not units.current_unit():
            columns = [('unit_id', 'Business Unit')] + columns
        return columns

    def details_columns_ex(self, model: KcServer, columns: List[Tuple[str, str]]) -> List[Tuple[str, str]]:
        server_tags = Tag.tag_to_map(model.mltag or '').keys()
        columns = columns + [('tag_' + tag, f'Tag {tag}') for tag in server_tags]
        if config.BUNITS and not units.current_unit():
            columns = [('unit_id', 'Business Unit')] + columns
        return columns

    def get_export_value(self, model: KcServer, name: str) -> Any:
        if name.startswith('tag_'):
            return self.get_tag(model.mltag, name[4:])  # pragma: no cover TODO
        elif name == 'unit_id':
            return units.titles().get(model.kc_key_rel.unit_id, 'UNKNOWN')
        else:
            return super().get_export_value(model, name)

    @pass_context
    def get_detail_value(self, context: Context, model: KcServer, name: str) -> Any:
        if name.startswith('tag_'):
            return self.get_tag(model.mltag, name[4:])
        elif name == 'unit_id':
            return units.titles().get(model.kc_key_rel.unit_id, 'UNKNOWN')
        else:
            return super().get_detail_value(context, model, name)

    @pass_context
    def get_list_value(self, context: Context, model: KcServer, name: str) -> Any:
        if name.startswith('tag_'):
            return self.get_tag(model.mltag, name[4:])  # pragma: no cover TODO
        elif name == 'unit_id':
            return units.titles().get(model.kc_key_rel.unit_id, 'UNKNOWN')
        elif name == 'mltag':
            data = model.mltag or u''
            mltag = (data[:30] + u'..') if len(data) > 30 else data
            return Markup('<div title="{0}"><i>{1}</i></div>'.format(escape(data.replace(';', '\n')), escape(mltag)))
        return super().get_list_value(context, model, name)

    named_filter_urls = True

    column_labels = dict(
        last='Check In',
        node='Hostname',
        mltag='Tags',
        euname='Effective Kernel',
        created='Registered',
        ltimestamp='Last Updated',
        remote_ip='IP',
        server_id='Server ID',
        kc_key='Key',
    )

    column_sortable_list = (
        'remote_ip',
        'node',
        'euname',
        'created',
        'last',
        'server_id',
        'kc_key',
        'mltag',
        'patchset',
    )

    column_filters = (
        FilterEqual(column=KcServer.kc_key, name='Key'),
        FilterLike(column=KcServer.remote_ip, name='IP'),
        FilterEqual(column=KcServer.remote_ip, name='IP'),
        FilterLike(column=KcServer.node, name="Hostname"),
        FilterEqual(column=KcServer.node, name="Hostname"),
        FilterLike(column=KcServer.mltag, name="Tags"),  # type: ignore[arg-type] # hybrid properties are not supported
        FilterEqual(column=KcServer.server_id, name='Server ID'),
        FilterLike(column=KcServer.euname, name='Effective Kernel'),
        FilterEqual(column=KcServer.euname, name='Effective Kernel'),
        DateEqualFilter(column=KcServer.last, name='Check In'),
        DateGreaterFilter(column=KcServer.last, name='Check In'),
        DateSmallerFilter(column=KcServer.last, name='Check In'),
        DateEqualFilter(column=KcServer.created, name='Registered'),
        DateGreaterFilter(column=KcServer.created, name='Registered'),
        DateSmallerFilter(column=KcServer.created, name='Registered'),
        FilterInactive(column=KcServer.last, name='Days Inactive'),
        DateEqualFilter(column=KcServer.ltimestamp, name='Last Updated'),
        DateGreaterFilter(column=KcServer.ltimestamp, name='Last Updated'),
        DateSmallerFilter(column=KcServer.ltimestamp, name='Last Updated'),
        FilterLike(column=KcServer.patch_type, name="Patch Type"),
        FilterEqual(column=KcServer.patch_type, name="Patch Type"),
        FilterFeed(column='feed', name='Feed'),
        FilterPatchsetEqual(column='patchset', name='Patchset'),
    )

    column_searchable_list = ['kc_key', 'server_id', 'mltag', 'node', 'remote_ip']

    column_formatters = dict(last=local_date_formatter, created=local_date_formatter, ltimestamp=local_date_formatter)

    def handle_filter(self, filter: BaseSQLAFilter) -> BaseSQLAFilter:
        filter = super().handle_filter(filter)
        if filter.name == 'Tags':
            self._filter_joins[filter.column] = [KcServer._mltag_rel]
        return filter

    def unit_filter_query(self, q: Query, is_count: bool) -> Query:
        unit_id = units.current_unit_id()
        if config.BUNITS:
            q = q.join(KcKey)

        if not is_count and config.BUNITS and not unit_id:
            q = q.options(joinedload(KcServer.kc_key_rel))

        if unit_id:
            q = q.filter(KcKey.unit_id == unit_id)

        return q

    def get_query(self) -> Query:
        return self.unit_filter_query(super().get_query(), False)

    def get_count_query(self) -> Query:
        return self.unit_filter_query(super().get_count_query(), True)


class IndexView(AdminIndexView):
    @expose('/')
    @expose('/<path>')
    def index(self, path: Optional[Any] = None) -> Response:
        return cast(Response, redirect(url_for('kcserver.index_view')))


class AuditView(UnitBaseView[AuditLog]):
    permissions = make_permissions(perms.AUDIT)

    can_create = False
    can_delete = False
    can_edit = False
    can_view_details = False

    column_default_sort = ('date', True)
    column_searchable_list = ('details',)
    column_display_actions = False

    named_filter_urls = True

    column_list = ('date', 'user', 'action', 'details')

    column_filters = (
        FilterEqual(column=audit.AuditLog.user, name='User'),
        FilterEqual(column=audit.AuditLog.action, name='Action'),
    )

    column_formatters = dict(
        date=local_date_formatter,
        user=filter_formatter,
        action=filter_formatter,
    )


def unit_url_value(slug: str, url: str) -> str:
    return f'/{escape(slug)}{url}'


class UnitView(ReplicatedMixin[Unit], BaseView[Unit]):
    permissions = make_permissions(
        {perms.MANAGE_UNIT, perms.VIEW_UNIT}, perms.MANAGE_UNIT, edit_view={perms.MANAGE_UNIT, perms.EDIT_UNIT_TITLE}
    )

    column_list = ('slug', 'login_url', 'api_url', 'title')
    column_labels = {'slug': 'Name'}
    column_formatters = {
        'login_url': mformatter(lambda m: Markup(f'<a href="{unit_url_value(m.slug, "/admin/login")}">login url</a>')),
        'api_url': mformatter(lambda m: Markup(f'<a href="{unit_url_value(m.slug, "/admin/api")}">API url</a>')),
    }

    form_args = {
        'slug': {
            'description': 'A unique ID used to identify this Business Unit',
            'validators': [NoneOf(['_'], message="Name '_' is reserved")],
        },
    }

    def mutate_form(self, model: Optional[Unit], form: Form, is_created: bool) -> None:
        if not perms.user_has_permission(perms.MANAGE_UNIT):
            del form['slug']

    def get_one(self, id: str) -> Optional[Unit]:
        model: Optional[Unit] = super().get_one(id)
        if not perms.user_has_permission(perms.MANAGE_UNIT):
            if model and model.id != units.current_unit_id():
                return None
        return model

    def unit_filter_query(self, q: Query) -> Query:
        if unit_id := units.current_unit_id():
            q = q.filter_by(id=unit_id)
        return q

    def get_query(self) -> Query:
        return self.unit_filter_query(super().get_query())

    def get_count_query(self) -> Query:
        return self.unit_filter_query(super().get_count_query())


class UnitUserView(ReplicatedMixin[UnitUser], UnitBaseView[UnitUser]):
    permissions = make_permissions(perms.MANAGE_UNIT_USERS, perms.MANAGE_UNIT_USERS)

    unit_exclude_default = True

    column_exclude_list = ['password_hash']
    column_searchable_list = ['username']
    column_list = ['username', 'role', 'description']
    column_formatters = dict(role=formatter(lambda value: perms.UnitRoles.get(value).title))

    form_columns = ('username', 'password', 'role', 'description')
    form_extra_fields = dict(password=PasswordField(label='Password', validators=[OptionalV(), password_length_v]))

    form_overrides = dict(role=SelectField)

    form_args = dict(role=dict(choices=perms.UnitRoles.titles.items()))

    def on_model_change(self, form: Form, model: UnitUser, is_created: bool) -> None:
        if model.password:
            model.set_password(model.password)
        super().on_model_change(form, model, is_created)


admin = Admin(app, index_view=IndexView(), template_mode='bootstrap3', base_template='base.html')

admin.add_view(UserView(User, db.session, 'Users'))
admin.add_view(KeyView(KcKey, db.session, 'Keys'))
admin.add_view(FeedView(Feed, db.session, 'Feeds'))
admin.add_view(ServerView(KcServer, db.session, 'Servers'))
admin.add_view(AuditView(audit.AuditLog, db.session, 'User Log'))
admin.add_view(UnitView(Unit, db.session, 'Units'))
admin.add_view(UnitUserView(UnitUser, db.session, 'Unit users'))
