import logging
from typing import Callable
from functools import wraps
from urllib.parse import quote

import requests
from oauthlib.oauth2 import WebApplicationClient
from oauthlib.oauth2.rfc6749.errors import OAuth2Error
from flask import redirect, request, session, url_for, flash, g, make_response
from flask_login import logout_user, login_required, current_user
from werkzeug.wrappers import Response
from typing_extensions import ParamSpec
from covador import opt
from covador.flask import form

import config
from eportal import perms, ldaplib, replicated, audit
from eportal.utils import get_session

from eportal.web import app, auth, units
from eportal.web.views.admin import admin

logger = logging.getLogger(__name__)


P = ParamSpec('P')


class FlashException(Exception):
    pass


def sso_catch_error(func: Callable[P, Response]) -> Callable[P, Response]:
    @wraps(func)
    def inner(*args: P.args, **kwargs: P.kwargs) -> Response:
        try:
            return func(*args, **kwargs)
        except FlashException as e:
            flash(str(e), category='error')
        except requests.exceptions.RequestException as e:
            logger.exception(e)
            flash(f'SSO network error: {str(e)}', category='error')
        except OAuth2Error as e:
            msg = f'SSO error: {e.error}'
            if e.description:
                msg += f' - {e.description}'
            flash(msg, category='error')
        except Exception as e:
            logger.exception(e)
            flash(f'SSO error: {str(e)}', category='error')
        return redirect(url_for('.login'))

    return inner


def with_requests_session(func: Callable[P, Response]) -> Callable[P, Response]:
    @wraps(func)
    def inner(*args: P.args, **kwargs: P.kwargs) -> Response:
        g.session = get_session()
        with g.session:
            return func(*args, **kwargs)

    return inner


@app.route('/admin/sso/login')
@sso_catch_error
@with_requests_session
def sso_login() -> Response:
    if not config.OIDC_CLIENT_ID:
        raise FlashException('OpenID Connect is not configured')

    oidc_client = WebApplicationClient(config.OIDC_CLIENT_ID)

    callback_url = f'{request_host_url()}/admin/sso/login/callback'
    auth_url, _, _ = oidc_client.prepare_authorization_request(
        config.OIDC_AUTH_URL, redirect_url=callback_url, scope=['openid', 'email']
    )

    units.global_session()['oidc_state'] = oidc_client.state
    if unit := units.current_unit():
        units.global_session()['sso_unit_slug'] = unit.slug
    else:
        units.global_session()['sso_unit_slug'] = ''
    return redirect(auth_url)


@app.route('/admin/sso/login/callback')
@sso_catch_error
@with_requests_session
def sso_login_callback() -> Response:
    if not config.OIDC_CLIENT_ID:
        raise FlashException('OpenID Connect is not configured')

    error = request.args.get('error')
    if error:
        flash(f"SSO login failed ({error})", category='error')
        return redirect(url_for('.login'))

    oidc_state = session.get('oidc_state')
    if not oidc_state:
        return redirect(url_for('.login'))

    if request.args['state'] != oidc_state:
        return make_response("Invalid OAuth 2.0 'state' parameter", 400)

    oidc_client = WebApplicationClient(config.OIDC_CLIENT_ID)

    code = request.args['code']

    token_url, headers, body = oidc_client.prepare_token_request(
        config.OIDC_TOKEN_URL,
        redirect_url=url_for('sso_login_callback', _external=True),
        code=code,
        include_client_id=False,
    )

    token_response = g.session.post(
        token_url, headers=headers, data=body, auth=(config.OIDC_CLIENT_ID, config.OIDC_CLIENT_SECRET)
    )

    oidc_client.parse_request_body_response(token_response.content)

    uri, headers, body = oidc_client.add_token(config.OIDC_USERINFO_URL)
    userinfo_response = g.session.get(uri, headers=headers, data=body)
    userinfo = userinfo_response.json()

    if not userinfo.get('email'):
        raise ValueError("OAuth2 userinfo should contain email")

    url = url_for('kckey.index_view')
    if slug := units.global_session()['sso_unit_slug']:
        url = f'/{slug}{url}'
        units.set_current_slug(slug)

    username = userinfo['email']
    user = auth.get_or_create_user(username, 'OpenID Connect')
    auth.set_auth_session(user, 'OAuth2')
    session['id_token_hint'] = oidc_client.token.get('id_token')

    return redirect(url)


@app.route('/admin/login', methods=['GET', 'POST'])
@form(username=opt(str, ''), password=opt(str, ''))
def login(username: str, password: str) -> Response:
    if request.method == 'POST':
        if not username or not password:
            flash('Email or password field cannot be empty.', category='error')
            return make_response(admin.index_view.render('login.html', username=username))

        redirect_url = request.args.get('next')
        if not redirect_url or redirect_url == '/':
            redirect_url = url_for('kcserver.index_view')

        if auth.db_login(username, password):
            return redirect(redirect_url)

        try:
            if auth.ldap_login(username, password):
                return redirect(redirect_url)
            flash('Invalid username or password. Please try again.', category='error')
        except ldaplib.LDAPSearchFailed as e:
            flash('{0}'.format(str(e)), category='error')
        except Exception as e:
            logger.exception(e)
            flash('LDAP Error: {0}: {1}'.format(type(e), str(e)), category='error')

    return make_response(admin.index_view.render('login.html', username=username))


@app.route('/admin/logout')
@login_required
def logout() -> Response:
    replicated.audit_log(units.current_unit_id(), current_user.username, audit.ACTION_LOGOUT, 'logged out')
    user = auth.load_user(current_user.id)
    is_oidc = user.description == 'OpenID Connect' if user is not None else False
    logout_user()
    flash('You have been logged out')
    url = url_for('login')
    if is_oidc and config.OIDC_LOGOUT_URL:
        after_logout_url = quote(f"{request_host_url()}{url}")
        id_token_hint_param = ""  # nosec B105
        if session.get('id_token_hint'):
            id_token_hint = session['id_token_hint']
            id_token_hint_param = f"&id_token_hint={quote(id_token_hint)}"
        redirect_url = f"{config.OIDC_LOGOUT_URL}?post_logout_redirect_uri={after_logout_url}{id_token_hint_param}"
        return redirect(redirect_url)
    else:
        return redirect(url)


def request_host_url() -> str:
    return f"{request.headers.get('X-Forwarded-Proto', 'http')}://{request.host}"


def get_patchserver_url() -> str:
    return config.PATCHSERVER_URL or request_host_url()


@app.context_processor
def inject_eportal_url() -> dict:
    return dict(eportal_url=get_patchserver_url())


@app.context_processor
def inject_perms() -> dict:
    if current_user.is_authenticated:
        return dict(perms=perms, permitted=perms.user_has_permission)
    return dict(perms=perms, permitted=lambda *_args, **__kwargs: False)
