import os
import gzip
import json
from datetime import timedelta, date, datetime
from logging import getLogger
from typing import Union, Dict, List, Optional, Any, cast
from dataclasses import dataclass

from markupsafe import escape
from covador import opt, soft_map, schema, enum, item, split
from covador.flask import query_string, form
from flask import request, redirect, url_for, jsonify, flash, make_response, g, Response
from flask_login import login_required, current_user
from markupsafe import Markup

import config
from eportal import perms, ldaplib, replicated, audit, patches_constants as pc, dynconfig, patchsource, token
from eportal.db import models
from eportal.db.models import Feed, is_product_allowed
from eportal.patches_impl import get_repo
from eportal.utils import get_version, extract_ip
from eportal.web import app, units
from eportal.web.auth import ensure_permission
from eportal.web.views import deployment
from eportal.web.views.admin import admin

logger = getLogger(__name__)

psproduct_t = opt(str, pc.PSProduct.KCE) | pc.PSProduct
stream_t = opt(str, pc.Stream.KERNEL) | pc.Stream


@app.route('/admin/api/kcare/check.plain')
@query_string(server_id=opt(str))
def check_license(server_id: Optional[str]):
    if not server_id:
        return 'success:true\ncode:3\n'
    try:
        kc_server = models.get_server(server_id)
        if not kc_server:
            return 'success:true\ncode:3\n'
        else:
            return 'success:true\ncode:0\n'
    except Exception:
        logger.exception('Raised while checking license stub')
        return 'success:true\ncode:3\n'


def clean(tag):
    return tag.strip().lower()


@app.route('/admin/api/kcare/tag_server.plain')
@query_string(server_id=str, tag=opt(str, u'', dest='raw_tags') | clean)
def tag_server(server_id: str, raw_tags: str):
    try:
        tagged = replicated.tag_server(server_id=server_id, tags=raw_tags)
        if tagged:
            return 'success:true\ncode:0\n'
        return 'success:true\ncode:-3\n'  # nothing tagged
    except models.NoSuchServer:
        return 'success:true\ncode:-1\n'
    except Exception:
        logger.exception('Raised while tagging server')
        return 'success:true\ncode:-2\n'


# there is no trial functionality in ePortal license,
# so here is a trial stub to avoid 404 on client side
@app.route('/admin/api/kcare/trial.plain')
def trial():
    return 'success:na\n'


@app.route('/admin/api/tuxcare/register.shell', methods=['POST'])
@form(key=opt(str, ''), hostname=opt(str, ''))
def tuxcare_register(key: str, hostname: str) -> str:
    def error(msg: str) -> str:
        return '\n'.join([f'echo {msg!r}', 'exit 1', ''])

    try:
        kc_key = models.get_key(key)
        if not kc_key or not kc_key.enabled:
            logger.error('Invalid key: {0}'.format(key))
            return error('Invalid key')

        if kc_key.limit_reached():
            logger.error('Key limit reached: {0}'.format(key))
            return error('Key limit reached')

        if not is_product_allowed(pc.Product.TUXCARE_ESU, kc_key.products_list):
            return error('Key should include TuxCare ESU product')

        ip = extract_ip(request)
        server = replicated.register_server(ip=ip, hostname=hostname, key=key)
        return f'EPORTAL_SERVER_ID={server.server_id!r}\n'
    except Exception:
        logger.exception('Raised while registering server')
        return error('Unhandled error, see ePortal logs')


@app.route('/admin/api/kcare/register_server.plain')
@query_string(key=str, hostname=opt(str, ''))
def register(key: str, hostname: str) -> Union[Response, str]:
    try:
        kc_key = models.get_key(key)
        if not kc_key or not kc_key.enabled:
            logger.error('Invalid key: {0}'.format(key))
            return 'code:2\n'
        if kc_key.limit_reached():
            logger.error('Key limit reached: {0}'.format(key))
            return 'code:3\n'

        ip = extract_ip(request)
        server = replicated.register_server(ip=ip, hostname=hostname, key=key)
        token_data = {
            'unit_id': kc_key.unit_id,
            'server_id': server.server_id,
            'feed': kc_key.feed or models.DEFAULT_FEED,
            'products': kc_key.products_list,
        }
        if config.REGISTRATION_TOKEN_INCLUDES_KEY:
            token_data['key'] = key

        tok = token.create(token_data, ttl=timedelta(hours=48), audience='agent')
        products_str = ",".join(kc_key.products_list or [])
        content = f'code:0\nserver_id:{server.server_id}\nproducts:{products_str}\n'
        resp = make_response(content)
        resp.headers['Kc-Auth-Token'] = tok
        return resp
    except Exception:
        logger.exception('Raised while registering server')
        return 'code:100\n'


@app.route('/admin/api/kcare/unregister_by_key.plain')
@login_required
@query_string(key=str, ip=str)
def unregister_by_key(key: str, ip: str):
    try:
        deleted_servers_cnt = replicated.unregister_server_by_key_and_ip(key=key, ip=ip)
        return Response(str(deleted_servers_cnt), content_type='text/plain')
    except Exception:
        logger.exception('Raised while unregistration by key')
        return Response('-1', content_type='text/plain')


@app.route('/admin/api/kcare/unregister_server.plain')
@query_string(server_id=str)
def unregister(server_id: str):
    try:
        replicated.unregister_server(server_id=server_id)
        return 'success:true'
    except Exception:
        logger.exception('Raised while unregistration by id')
        return 'success:false'


@app.route('/admin/ldap-edit/')
@login_required
def ldap_settings():
    ensure_permission(perms.MANAGE_CONFIG)
    conf = dynconfig.Config()
    return admin.index_view.render('ldap_settings_edit.html', **conf.ldap)


@app.route('/admin/ldap-edit/', methods=['POST'])
@login_required
@form(url=opt(str, ''), bind_template=opt(str, ''), filter_template=opt(str, ''), action=opt(str, ''))
def ldap_settings_save(url: str, bind_template: str, filter_template: str, action: str):
    ensure_permission(perms.MANAGE_CONFIG)
    conf = dynconfig.Config()

    if action == 'reset':
        conf.reset_ldap()
    else:
        try:
            ldaplib.verify(url, bind_template, filter_template)
        except Exception as ex:
            flash(f'Error saving LDAP settings: {ex}', category='error')
            logger.exception(ex)
            return admin.index_view.render(
                'ldap_settings_edit.html', url=url, bind_template=bind_template, filter_template=filter_template
            )

        conf.configure_ldap(url, bind_template, filter_template)

    replicated.audit_log(0, current_user.username, audit.ACTION_MANAGE_LDAP, 'changed ldap configuration')
    flash('LDAP settings updated', category='success')
    return redirect(url_for('ldap_settings'))


@app.route('/admin/patch-source/')
@login_required
def source_edit():
    ensure_permission(perms.MANAGE_CONFIG)
    ps = patchsource.PatchSource()

    if ps.is_configured():
        url = ps.config.patchrepo['url']
        login = ps.config.patchrepo['login']
        password = ps.config.patchrepo['password']
        hide = not config.UI_DEBUG
    else:
        url = 'https://downloads.kernelcare.com/patch-download/'
        login = ''
        # No password leakage, suppress Bandit warning.
        password = ''  # nosec B105
        hide = False

    enabled_distros = ps.config.patchrepo['distros']
    all_distros = sorted(set(ps.get_available_distros()) | set(enabled_distros))
    return admin.index_view.render(
        'patch_source_edit.html',
        url=url,
        login=login,
        password=password,
        all_distros=all_distros,
        enabled_distros=enabled_distros,
        static_patch_source=bool(config.PATCH_SOURCE_USER),
        hide=hide,
    )


@app.route('/admin/patch-source/', methods=['POST'])
@login_required
@form(
    login=opt(str, ''),
    password=opt(str, ''),
    url=opt(str, ''),
    distro=opt(str, [], multi=True),
    all_distros=split(str),
    action=opt(str, ''),
)
def source_save(login: str, password: str, url: str, distro: List[str], all_distros: List[str], action: str):
    ensure_permission(perms.MANAGE_CONFIG)
    ps = patchsource.PatchSource()
    ps.config.configure_enabled_distros(distro)

    if action == 'reset':
        ps.config.reset_patchrepo()
        flash('Credentials have been reset successfully.')
        return redirect(url_for('source_edit'))

    if not config.PATCH_SOURCE_USER and password:
        try:
            patchsource.check_repo(login, password, url)
        except Exception as ex:
            logger.exception('Raised while saving patch source')
            flash(f'Error saving patch source: {ex}', category='error')
            return admin.index_view.render(
                'patch_source_edit.html',
                url=url,
                login=login,
                password=password,
                enabled_distros=distro,
                all_distros=all_distros,
            )

        ps.config.configure_patchrepo(login, password, url)

    replicated.audit_log(
        0,
        current_user.username,
        audit.ACTION_MANAGE_PATCH_SOURCE,
        f'changed patch source to {login}/{url}, distros: {distro}',
    )
    return redirect(url_for('feed_dashboard'))


@app.route('/admin/patch-changelog/<string:name>')
@app.route('/admin/patch-changelog/<feed>/<string:name>')
@login_required
def patch_changelog(name: str, feed: Optional[str] = None):
    try:
        return escape(patchsource.PatchSource().get_changelog_text(name))
    except Exception as ex:
        logger.exception('Raised while patching changelog')
        return escape(f'Unable to get changelog for patchset {name}: {ex}')


@app.route('/patch-download/<string:name>')
def patch_download_local(name: str):
    response = make_response()
    path = os.path.join('/patchsets', 'arch', name)
    response.headers['X-Accel-Redirect'] = path
    return response


@app.route('/admin/api/feed-names')
def get_available_feeds():
    return jsonify({'result': Feed.get_all_names(units.current_unit_id())})


def select_default_ps_product(feed_groups: Dict[pc.PSProduct, Any]) -> pc.PSProduct:
    if pc.PSProduct.KCE in feed_groups:
        return pc.PSProduct.KCE
    elif pc.PSProduct._ALMACARE_CS in feed_groups:
        return pc.PSProduct._ALMACARE_CS
    return pc.PSProduct.KCE


@dataclass
class DashboardContent:
    ps_product: pc.PSProduct
    ps_products: List[pc.PSProduct]
    feeds: List[str]


def get_dashboard_content(unit_id: int, psproduct: Optional[pc.PSProduct]) -> DashboardContent:
    show_main = models.KcKey.find_default_feed(unit_id) or len(Feed.get_all_names(unit_id)) == 1
    feed_groups = Feed.get_feeds_by_product(unit_id)

    if not show_main and len(feed_groups) > 1:  # pragma: no cover, backward compatibility with almacare-cs
        for psproduct in list(feed_groups):
            if feed_groups[psproduct] == [models.DEFAULT_FEED]:
                feed_groups.pop(psproduct)

    psproduct = psproduct or select_default_ps_product(feed_groups)
    feeds = feed_groups.get(psproduct, [])
    ps_products = [it for it in pc.PSProduct if it in feed_groups]
    return DashboardContent(psproduct, ps_products, feeds)


@app.route('/admin/feeds/')
@login_required
@query_string(stream=stream_t, ps_product=psproduct_t)
def feed_dashboard(stream: pc.Stream, ps_product: pc.PSProduct):
    g.stream = stream
    dc = get_dashboard_content(units.current_unit_id(), ps_product)
    ps = patchsource.PatchSource()

    try:
        dashboard = get_repo(units.current_unit_id(), stream, ps, forbidden_ok=False).dashboard(dc.feeds)
    except Exception as e:
        logger.exception('Raised while getting patchset info')
        flash(Markup(f'Unable to get patchset info: {e}'), category='warning')
        dashboard = get_repo(units.current_unit_id(), stream, ps, is_local=True).dashboard(dc.feeds)

    if config.REVERSE_PATCHSETS:
        dashboard['patchsets'] = reversed(dashboard['patchsets'])

    return admin.index_view.render(
        'dashboard.html',
        userspace=stream in pc.USER_STREAMS,
        local=not ps.is_configured(),
        stream=stream,
        deployment_status=deployment.get_status(stream),
        enabled_distros=ps.config.patchrepo['distros'],
        ps_products=dc.ps_products,
        ps_product=dc.ps_product,
        all_feeds=dc.feeds,
        checkin_active_date=str(date.today() - timedelta(days=models.DAY_DELTA)),
        **dashboard,
    )


@app.route('/admin/docs/')
@login_required
def docs():
    return admin.index_view.render('docs.html')


@app.route('/admin/feeds/deploy-settings/')
@login_required
@query_string(stream=stream_t, patchset=str, ps_product=psproduct_t)
def deploy_settings(stream, patchset, ps_product):
    feeds = Feed.get_feeds_by_product(units.current_unit_id()).get(ps_product, [])
    return admin.index_view.render(
        'deploy-settings.html',
        stream=stream,
        feed_names=feeds,
        patchset=patchset,
        ps_product=ps_product.value,
    )


status_t = opt(str) | enum('enable', 'disable', 'enable-upto', 'undeploy-downto')


@app.route('/admin/feeds/deploy-settings/', methods=['POST'])
@login_required
@form(
    soft_map(
        stream=stream_t,
        patchset=str,
        feeds=item(str, multi=True),
        deploy_all=opt(status_t),
        ps_product=psproduct_t,
    )
)
def deploy_settings_apply(
    stream: pc.Stream, patchset: str, feeds: List[str], deploy_all: Optional[str], ps_product: pc.PSProduct, **rest
):
    ensure_permission(perms.DEPLOY)
    deploy_orders = schema({f'deploy_{i}': opt(status_t, dest=fname) for i, fname in enumerate(feeds)})(rest)

    if deploy_all:
        for feed in list(deploy_orders.keys()):
            deploy_orders[feed] = deploy_all

    replicated.deploy_patchset(
        unit_id=units.current_unit_id(),
        stream=stream.value,
        patchset=patchset,
        deploy_orders=deploy_orders,
        user=current_user.username,
    )
    return redirect(url_for('feed_dashboard', stream=stream, ps_product=ps_product.value))


@app.route('/admin/export-import/')
def export_import_data():  # pragma: no cover
    ensure_permission(perms.EXPORT_DATA)
    return admin.index_view.render('export-import.html')


@app.route('/admin/export-import/', methods=['POST'])
@form(action=str | enum('import', 'export'))
def export_import_data_apply(action: str):  # pragma: no cover
    ensure_permission(perms.EXPORT_DATA)
    if action == 'export':
        result = cast(dict, replicated.full_changes())
        result['servers'] = replicated.get_registrations()
        data = gzip.compress(json.dumps(result, ensure_ascii=False).encode())
        dt = datetime.now()  # local host time
        fname = f'eportal-state-{dt.strftime("%Y-%m-%d_%H:%M")}.json.gz'
        return app.response_class(
            data, mimetype='application/gzip', headers={'Content-Disposition': f'attachment; filename={fname}'}
        )
    elif action == 'import':
        file = request.files.get('data')
        if not file or not file.filename:
            flash('Select file (eportal-state-*.json.gz) for import', 'error')
            return redirect(request.url)
        audit.add_record(0, current_user.username, audit.ACTION_IMPORT_DATA, file.filename)
        changes = json.load(gzip.GzipFile(mode='rb', fileobj=file.stream))
        replicated.apply_full_changes(changes, current_user.username)
        replicated.apply_server_registrations(changes['servers'])
        flash('Import done', 'success')
        return redirect(request.url)


@app.context_processor
def inject_data():
    version, _ = get_version()
    result = dict(
        feeds=lambda: Feed.get_all_names(units.current_unit_id()),
        version=version,
        current_unit=units.current_unit,
        unit_slugs=units.slugs,
    )

    if hasattr(g, 'stream'):
        is_kernel = g.stream == pc.Stream.KERNEL
        result.update(
            dict(
                is_kernel=is_kernel,
                has_changelog=is_kernel,
                stream_title=g.stream.title,
            )
        )

    return result


app.add_template_global(enumerate)
