import functools
import hmac
import json
import zlib
from typing import Any, Callable, Optional

from covador import opt
from covador.flask import query_string
from flask import jsonify, request, abort
from flask.wrappers import Response
from werkzeug.exceptions import HTTPException

from eportal.web import app
from eportal import oplog, replicated, utils

import config


def replication_auth(fn: Callable) -> Callable:
    @functools.wraps(fn)
    def inner(*args: Any, **kwargs: Any) -> Any:
        if (
            config.REPLICATION_SHARED_KEY
            and request.authorization
            and request.authorization.password
            and hmac.compare_digest(request.authorization.password, config.REPLICATION_SHARED_KEY)
        ):
            return fn(*args, **kwargs)
        return abort(401)

    return inner


def check_version(version: int) -> None:
    if version < oplog.VERSION:
        resp = jsonify({'error': f'incompatible replication protocol, version {oplog.VERSION} required'})
        resp.status = '409 Incompatible replication protocol'
        raise HTTPException(response=resp)


@app.route('/admin/api/oplog/changes')
@replication_auth
@query_string(last_id=opt(int), node_url=str, action_prefix=opt(str), active=opt(bool, True), version=opt(int, 1))
def oplog_changes(
    last_id: Optional[int], node_url: str, action_prefix: Optional[str], active: bool, version: int
) -> Response:
    node = oplog.node_from_url(node_url)
    if active:
        oplog.update_follower(node, node_url)
    check_version(version)
    changes = oplog.read(last_id, None, node, action_prefix)
    changes['instance_id'] = utils.get_instance_id()
    return jsonify(changes)


@app.route('/admin/api/oplog/server-registrations')
@replication_auth
@query_string(version=opt(int, 1))
def server_registrations(version: int) -> Response:
    check_version(version)
    regs = replicated.get_registrations()
    content = json.dumps(regs, ensure_ascii=False).encode()
    headers = None

    if 'deflate' in request.accept_encodings:  # type: ignore[comparison-overlap]
        content = zlib.compress(content)
        headers = {'Content-Encoding': 'deflate'}

    return app.response_class(content, mimetype='application/json', headers=headers)


@app.route('/admin/api/oplog/full-changes')
@replication_auth
@query_string(version=opt(int, 1))
def full_changes(version: int) -> Response:
    check_version(version)
    result = replicated.full_changes()
    return app.response_class(json.dumps(result, ensure_ascii=False), mimetype='application/json')
