import logging
import os
import re

from hashlib import sha256
from base64 import b16encode
from functools import wraps
from typing import Any, Callable, Dict, Union, Optional

from flask.wrappers import Response
from covador import opt
from covador.flask import args
from flask import request, make_response, abort, send_file, jsonify, render_template
from flask_login import login_required
from flask.helpers import send_from_directory

import config
from eportal import utils, patches_constants as pc, deploy, token
from eportal.db.models import DEFAULT_FEED, get_kc_key_fields_by_server, is_product_allowed, ServerInfo
from eportal.patches_impl import get_store
from eportal.web import app
from eportal.patches import api
from eportal.patches.serve import ReleaseAwareLocation
from werkzeug.datastructures import WWWAuthenticate

logger = logging.getLogger(__name__)

BAD_CLIENTS = ['2.10-1', '2.10-2', '2.10-3', '2.10-4', '2.10-6', '2.10-7', '2.11-1', '2.11-2', '2.11-3']

BAD_REGEX = re.compile(r'({})'.format('|'.join(b16encode(f"'{it}'".encode()).decode() for it in BAD_CLIENTS)))
CACHE_KEY = 'Kc-Cache-Key'


def server_id_required(func: Callable) -> Callable:
    @wraps(func)
    def inner(**kwargs: Any) -> Any:
        auth = request.authorization

        if auth is None:
            response = make_response('', 401)
            response.www_authenticate = WWWAuthenticate("basic")
            return response

        return func(server_id=auth.username, **kwargs)

    return inner


def server_id_opt(func: Callable) -> Callable:
    @wraps(func)
    def inner(**kwargs: Any) -> Any:
        server_id = request.authorization and request.authorization.username
        return func(server_id=server_id, **kwargs)

    return inner


def token_auth(func: Callable) -> Callable:
    @wraps(func)
    def inner(**kwargs: Any) -> Any:
        tok = request.headers.get('Kc-Auth-Token')
        if not tok:
            return func(token_data=None, **kwargs)

        try:
            token_data = token.verify(tok, audience='agent', required_claims=['server_id', 'feed', 'products'])
        except token.TokenError as e:
            if tok:
                logger.debug('bad token auth result: %s', str(e))
            token_data = None
        return func(token_data=token_data, **kwargs)

    return inner


@app.route('/')
@app.route('/index.html')
@utils.wrap_if(lambda: config.AUTH_RESTRICTED_INDEX, login_required)
def index_page() -> Response:
    static_dir = config.jroot('patches')
    return send_from_directory(static_dir, 'index.html')


@app.route('/patches.json')
@app.route('/<feed>/patches.json')
def get_patches_json(feed: str = deploy.DEFAULT_FEED) -> Response:
    data = get_store(0, pc.Stream.KERNEL).get_patches_json(feed)
    return jsonify(data)


@app.route('/userspace-patches.json')
@app.route('/<feed>/userspace-patches.json')
def get_userspace_patches_json(feed: str = deploy.DEFAULT_FEED) -> Response:
    data = get_store(0, pc.Stream.USER).get_userspace_patches_json(feed)
    return jsonify(data)


@app.route('/qemu-patches.json')
@app.route('/<feed>/qemu-patches.json')
def get_qemu_patches_json(feed: str = deploy.DEFAULT_FEED) -> Response:
    data = get_store(0, pc.Stream.QEMU).get_userspace_patches_json(feed)
    return jsonify(data)


@app.route('/db-patches.json')
@app.route('/<feed>/db-patches.json')
def get_db_patches_json(feed: str = deploy.DEFAULT_FEED) -> Response:
    data = get_store(0, pc.Stream.DB).get_userspace_patches_json(feed)
    return jsonify(data)


@app.route('/img/<name>', defaults={"directory": "img"})
@app.route('/css/<name>', defaults={"directory": "css"})
@app.route('/js/<name>', defaults={"directory": "js"})
@app.route('/favicon.ico', defaults={'name': 'favicon.ico'})
@app.route('/license.txt', defaults={'name': 'license.txt'})
def get_static(name: str, directory: str = '') -> Response:
    static_dir = config.jroot('patches')
    return send_from_directory(os.path.join(static_dir, directory), name)


def serve_file(fpath: str, release: Union[str, ReleaseAwareLocation, None] = None) -> Response:
    if config.TEST_SEND_FILE:
        return make_response(fpath + '\n')

    if not config.CACHE_MODE and 'SENDFILE' in os.environ:
        return send_file(fpath)

    fpath = fpath[len(config.PATCHES_DIR) :]
    response = make_response()
    if config.CACHE_MODE:
        fpath = fpath.lstrip('/').partition('/')[2]
        response.headers['X-Accel-Redirect'] = os.path.join('/patchsets-proxy', fpath)
        if release:
            response.headers['KC-Base-Url'] = os.path.join('/patches', release)
    else:
        response.headers['X-Accel-Redirect'] = os.path.join('/patchsets', fpath.lstrip('/'))
        if release:
            response.headers['KC-Base-Url'] = os.path.join('/patches', release)
    return response


# feed: for compatibility reasons. Patchserver handles feed from agent. ePortal forces feed from central place.
@app.route('/<kernel>/<sticky>.<any(latest.v1, latest.v2):latest>')
@app.route('/<feed>/<kernel>/<sticky>.<any(latest.v1, latest.v2):latest>')
@app.route('/<feed>/<kernel>/<any(latest.v1, latest.v2):latest>')
@app.route('/<kernel>/<any(latest.v1, latest.v2):latest>')
@server_id_opt
@token_auth
@args(feed=opt(str, DEFAULT_FEED), kernel=str, latest=str, sticky=opt(api.parse_sticky))
def get_latest(
    server_id: Optional[str],
    feed: str,
    kernel: str,
    latest: str,
    token_data: Optional[dict],
    sticky: Optional[api.StickyT],
) -> Response:
    if BAD_REGEX.search(request.full_path):
        abort(404)
    is_new_format = 'KC-Patch-Version' in request.headers

    sinfo = get_server_info_from_server_id_or_token(server_id, token_data)
    if not sinfo:
        sinfo = ServerInfo(0, '', DEFAULT_FEED, None)

    response: Optional[Response] = None
    if is_new_format:
        try:
            version = int(request.headers['KC-Patch-Version'])
        except ValueError:
            return make_response(jsonify({'error': 'Invalid patch version'}), 400)
        json_level = get_store(sinfo.unit_id, pc.Stream.KERNEL).get_latest_json(sinfo.feed, kernel, version, sticky)
        if json_level:
            response = make_response(jsonify(json_level), 200)
    else:
        int_level = get_store(sinfo.unit_id, pc.Stream.KERNEL).get_latest_content(sinfo.feed, kernel, latest, sticky)
        if int_level is not None:
            response = make_response(str(int_level), 200)
            response.headers['Content-Type'] = 'text/plain'

    if response is None:
        response = make_response('', 404)

    _add_cache_header(response, sinfo.key)
    return response


def _get_kernel_file(unit_id: int, feed: str, kernel: str, level: str, fname: str) -> Response:
    fpath = get_store(unit_id, pc.Stream.KERNEL).get_file_path(feed, kernel, level, fname)
    if not fpath:
        abort(404)

    return serve_file(fpath, release=fpath.release)


def _add_cache_header(resp: Response, key: Optional[str]) -> None:
    if key:
        cache_key = sha256(key.encode()).hexdigest()
        resp.headers[CACHE_KEY] = cache_key


def get_server_info_from_server_id_or_token(
    server_id: Optional[str], token_data: Optional[dict]
) -> Optional[ServerInfo]:
    sinfo = get_kc_key_fields_by_server(server_id)
    if not sinfo and token_data:
        token_server_id = token_data['server_id']
        if server_id != token_server_id:
            abort(401)
        sinfo = ServerInfo(token_data.get('unit_id', 0), 'NONE', token_data['feed'], token_data['products'])
    return sinfo


@app.route('/<kernel>/<fname>')
@app.route('/<kernel>/<int:level>/<fname>')
@app.route('/<feed>/<kernel>/<fname>')
@app.route('/<feed>/<kernel>/<int:level>/<fname>')
@server_id_required
@token_auth
@args(feed=opt(str, DEFAULT_FEED), kernel=str, level=opt(str, ''), fname=str)
def get_kernel_file(
    server_id: str, feed: str, kernel: str, level: str, fname: str, token_data: Optional[Dict[str, Any]]
) -> Response:
    sinfo = get_server_info_from_server_id_or_token(server_id, token_data)
    if not sinfo:
        abort(401)

    if not is_product_allowed(pc.Product.KERNEL, sinfo.products):
        abort(403)

    return _get_kernel_file(sinfo.unit_id, sinfo.feed, kernel, level, fname)


@app.route('/<kernel>/<int:level>/<fname>.html')
@app.route('/<feed>/<kernel>/<int:level>/<fname>.html')
@args(feed=opt(str, deploy.DEFAULT_FEED), kernel=str, level=opt(str, ''), fname=str)
def get_public_kernel_file(feed: str, kernel: str, level: str, fname: str) -> Response:
    return _get_kernel_file(0, feed, kernel, level, fname + '.html')


@app.route('/u/<build_id>/<int:level>/info.html', defaults={"prefix": "u"})
@app.route('/db/<build_id>/<int:level>/info.html', defaults={"prefix": "db"})
@app.route('/qemu/<build_id>/<int:level>/info.html', defaults={"prefix": "qemu"})
def get_userspace_info(prefix: str, build_id: str, level: int) -> str:
    store = get_store(0, pc.USER_PREFIX_STREAM_MAP[prefix])
    info_json = store.get_libcare_info(build_id, level)
    if not info_json:
        abort(404)
    return render_template('package-info.html', info=info_json)


@app.route('/u/<feed>/<lib>/<build_id>/<any(latest.v1):latest>', defaults={"prefix": "u"})
@app.route('/db/<feed>/<lib>/<build_id>/<any(latest.v1):latest>', defaults={"prefix": "db"})
@app.route('/qemu/<feed>/<lib>/<build_id>/<any(latest.v1):latest>', defaults={"prefix": "qemu"})
@server_id_opt
@token_auth
@args(feed=opt(str, DEFAULT_FEED), build_id=str, latest=str, lib=opt(str))
def get_latest_userspace(
    server_id: Optional[str],
    prefix: str,
    feed: str,
    build_id: str,
    latest: str,
    lib: Optional[str],
    token_data: Optional[Dict[str, Any]],
) -> Response:
    sinfo = get_server_info_from_server_id_or_token(server_id, token_data)
    if sinfo:
        server_exists = True
    else:
        sinfo = ServerInfo(0, '', DEFAULT_FEED, [])
        server_exists = False

    stream = pc.USER_PREFIX_STREAM_MAP[prefix]
    store = get_store(sinfo.unit_id, stream)

    latest_dict = store.get_latest_libcare(sinfo.feed, build_id, latest, '/' + prefix)
    if not latest_dict:
        response = make_response('', 404)
    else:
        if not server_exists or not is_product_allowed(stream.product, sinfo.products):
            latest_dict['patch_url'] = 'forbidden'
        else:
            latest_dict['baseurl'] = f'/{prefix}/patches/{latest_dict["release"]}'

        response = make_response(jsonify(latest_dict), 200)

    _add_cache_header(response, sinfo.key)
    return response


# Development endpoint, in production served by nginx
@app.route('/u/patches/<path:path>', defaults={"prefix": "u"})
@app.route('/db/patches/<path:path>', defaults={"prefix": "db"})
@app.route('/qemu/patches/<path:path>', defaults={"prefix": "qemu"})
def serve_userspace_patches(prefix: str, path: str) -> Response:
    store = get_store(0, pc.USER_PREFIX_STREAM_MAP[prefix])
    resource_location = store.get_resource_location(path)
    if not resource_location:
        raise RuntimeError(f'resource {path} not found')  # pragma: no cover
    return serve_file(resource_location)


@app.route('/patches/<path:path>')
def serve_kernelcare_patches(path: str) -> Response:
    resource_location = get_store(0, pc.Stream.KERNEL).get_resource_location(path)
    if not resource_location:
        raise RuntimeError(f'resource {path} not found')  # pragma: no cover
    return serve_file(resource_location)
