import json
import datetime
from logging import getLogger
from typing import Any, List, Optional

from covador import opt, enum, item
from covador.flask import query_string, form, json_body, mergeof, params
from flask import request, make_response
from flask.wrappers import Response
from flask_login import current_user

import config
from eportal import perms, replicated, patches_impl, patches_constants as pc
from eportal.db import models, session
from eportal.dbutils import transaction
from eportal.db.models import get_server_list, NoSuchServer
from eportal.utils import extract_ip, age_t, random_id
from eportal.web import app, units
from eportal.web.auth import ensure_permission, login_required_with_json_error

logger = getLogger(__name__)
anysource = mergeof(query_string, form, json_body)

ERROR_INVALID_FEED = 'invalid-feed-name'


class DatetimeEncoder(json.JSONEncoder):
    def default(self, obj: Any) -> str:
        if isinstance(obj, datetime.datetime):
            as_tz = None if config.USE_LOCAL_TIMEZONE_FOR_API else datetime.timezone.utc
            return obj.replace(tzinfo=datetime.timezone.utc).astimezone(tz=as_tz).isoformat()

        return super().default(obj)  # type: ignore[no-any-return]  # pragma: no cover


def make_json_response(data: Any = None, response_status: int = 200, **kwargs: Any) -> Response:
    """use custom datetime serializer for json responses"""

    content = json.dumps(data or kwargs, cls=DatetimeEncoder)
    resp = make_response(content, response_status)
    resp.headers["Content-Type"] = "application/json"
    return resp


@app.route('/admin/api/servers')
@login_required_with_json_error
@query_string(
    feed=opt(str),
    key=opt(str),
    registered_age=opt(str) | age_t,
    checkin_age=opt(str) | age_t,
    updated_age=opt(str) | age_t,
    is_updated=opt(bool),
    hostname=opt(str),
    ip=opt(str),
    server_id=opt(str),
    limit=opt(int, 10),
    offset=opt(int, 0),
    only_count=opt(bool),
    tag=opt(str),
)
def server_list(**filters: Any) -> Response:
    ensure_permission(perms.VIEW, json=True)
    if unit_id := units.current_unit_id():
        filters['unit_id'] = unit_id
    cnt, result = get_server_list(**filters)
    json_result = {'count': cnt, 'result': result, 'limit': filters['limit'], 'offset': filters['offset']}
    unknown_params = set(request.args) - set(filters)
    if unknown_params:
        json_result['unknown_params'] = list(unknown_params)
    return make_json_response(json_result)


@app.route('/admin/api/register', methods=['POST'])
@form(key=str, hostname=opt(str, ''))
def register_json(key: str, hostname: str) -> Response:
    try:
        kc_key = models.get_key(key)
        if not kc_key or not kc_key.enabled:
            logger.error('Invalid key: {0}'.format(key))
            return make_json_response({'error': 'invalid-key'}, 400)
        if kc_key.limit_reached():
            logger.error('Key limit reached: {0}'.format(key))
            return make_json_response({'error': 'key-limit-reached'}, 400)

        ip = extract_ip(request)
        server = replicated.register_server(ip=ip, hostname=hostname, key=key)
        return make_json_response({'server_id': server.server_id})
    except Exception:
        logger.exception('Raised while registering server')
        return make_json_response({'error': 'internal-error'}, 500)


@app.route('/admin/api/delete_server', methods=['POST'])
@login_required_with_json_error
@query_string(ip=opt(str), hostname=opt(str), server_id=opt(str))
def delete_server(ip: Optional[str], hostname: Optional[str], server_id: Optional[str]) -> Response:
    ensure_permission(perms.MANAGE_SERVERS, json=True)
    ips = hosts = server_ids = None
    if ip:
        ips = [ip]
    if hostname:
        hosts = [hostname]
    if server_id:
        server_ids = [server_id]

    if not any((ips, hosts, server_ids)):
        return make_json_response({'error': 'bad-request', 'details': 'ip, hostname or server_id are required'}, 400)

    count = replicated.delete_servers(ips=ips, hosts=hosts, server_ids=server_ids)
    return make_json_response(result=count)


@app.route('/admin/api/feeds/')
@login_required_with_json_error
def list_feeds() -> Response:
    ensure_permission(perms.VIEW, json=True)
    main = {'name': 'main', 'auto': False, 'deploy_after': 0, 'channel': pc.Channel.DEFAULT.value}
    return make_json_response(
        result=[main] + [it.to_json(public=True) for it in models.get_feeds(units.current_unit_id())]
    )


@app.route('/admin/api/feeds/', methods=['POST'])
@login_required_with_json_error
@anysource(
    name=str,
    auto=opt(bool),
    deploy_after=opt(int, 0),
    channel=opt(str, pc.Channel.DEFAULT.value) | enum(*pc.Channel.items),
)
def create_modify_feed(**data: Any) -> Response:
    ensure_permission(perms.MANAGE_FEEDS, json=True)
    obj = models.ensure_object(models.Feed, {'name': data['name'], 'unit_id': units.current_unit_id()}, **data)
    replicated.sync_model(obj, current_user.username)
    return make_json_response(result=obj.to_json(public=True))


@app.route('/admin/api/feeds/<name>', methods=['DELETE'])
@login_required_with_json_error
def delete_feed(name: str) -> Response:
    ensure_permission(perms.MANAGE_FEEDS, json=True)
    unit_id = units.current_unit_id()
    with transaction(session):
        obj = models.get_feed(unit_id, name)
        deleted = 0
        if obj:
            obj.on_delete()
            session.delete(obj)
            replicated.delete_model(obj, user=current_user.username)
            deleted = 1

    return make_json_response(result=deleted)


@app.route('/admin/api/keys/')
@login_required_with_json_error
def list_keys() -> Response:
    ensure_permission(perms.VIEW, json=True)
    return make_json_response(result=[it.to_json(public=True) for it in models.get_keys(units.current_unit_id())])


@app.route('/admin/api/keys/', methods=['POST'])
@login_required_with_json_error
@anysource(key=opt(str), feed=opt(str), note=opt(str), server_limit=opt(int, 0))
@json_body(products_json=opt(list, src='products'))
@params(products_args=opt(str, multi=True, src='product'))
def create_modify_key(**data: Any) -> Response:
    ensure_permission(perms.MANAGE_KEYS, json=True)

    feed = data['feed']
    if feed and feed not in models.Feed.get_all_names(units.current_unit_id()):
        return make_json_response({'error': ERROR_INVALID_FEED}, 400)

    if not data['key']:
        data['key'] = random_id()

    data['products'] = data.pop('products_args', None) or data.pop('products_json', None)

    try:
        obj = models.ensure_object(models.KcKey, {'key': data['key'], 'unit_id': units.current_unit_id()}, **data)
    except ValueError as e:
        return make_json_response({'error': str(e)}, 400)

    replicated.sync_model(obj, current_user.username)
    return make_json_response(result=obj.to_json(public=True))


@app.route('/admin/api/keys/<key>', methods=['DELETE'])
@login_required_with_json_error
def delete_key(key: str) -> Response:
    ensure_permission(perms.MANAGE_KEYS, json=True)
    with transaction(session):
        obj = models.get_key_mut(units.current_unit_id(), key)
        deleted = 0
        if obj:
            session.delete(obj)
            replicated.delete_model(obj, user=current_user.username)
            deleted = 1

    return make_json_response(result=deleted)


feed_t = opt(str, models.DEFAULT_FEED)
stream_t = opt(str, pc.Stream.KERNEL, src='product') | enum(*pc.Stream.items) | pc.Stream
action_t = item(str) | enum('enable', 'disable', 'enable-upto', 'undeploy-downto')


@app.route('/admin/api/patchsets/')
@login_required_with_json_error
@query_string(feed=feed_t, stream=stream_t)
def list_patchsets(feed: str, stream: pc.Stream) -> Response:
    ensure_permission(perms.VIEW, json=True)
    unit_id = units.current_unit_id()
    if feed not in models.Feed.get_all_names(unit_id):
        return make_json_response({'error': ERROR_INVALID_FEED}, 400)
    repo = patches_impl.get_repo(unit_id, stream)
    result = repo.get_feed_status(feed)
    return make_json_response(result=[{'patchset': r, 'status': s} for r, s in result])


@app.route('/admin/api/patchsets/manage', methods=['POST'])
@login_required_with_json_error
@query_string(patchset=str, feeds=item(str, multi=True, src='feed'), stream=stream_t, action=action_t)
def manage_patchset(patchset: str, feeds: List[str], stream: pc.Stream, action: str) -> Response:
    ensure_permission(perms.DEPLOY, json=True)
    unit_id = units.current_unit_id()

    available_feeds = models.Feed.get_all_names(unit_id)
    if 'all' in feeds:
        feeds = available_feeds

    unknown_feeds = set(feeds) - set(available_feeds)
    if unknown_feeds:
        return make_json_response({'error': ERROR_INVALID_FEED, 'details': list(unknown_feeds)}, 400)

    deploy_orders = {it: action for it in feeds}
    replicated.deploy_patchset(
        unit_id=unit_id,
        stream=stream.value,
        patchset=patchset,
        deploy_orders=deploy_orders,
        user=current_user.username,
    )
    return make_json_response(result='ok')


@app.route('/admin/api/users', methods=['GET'])
@query_string(unit_id=opt(int, 0))
@login_required_with_json_error
def list_users(unit_id: int) -> Response:
    ensure_permission(perms.VIEW, json=True)

    current_unit_id = units.current_unit_id()
    if current_unit_id:
        unit_id = current_unit_id

    if unit_id:
        users = models.get_all_unit_users(unit_id)
    else:
        users = models.get_all_users()

    result = {'result': [user.to_json(public=True) for user in users]}
    return make_json_response(result)


@app.route('/admin/api/set_tags', methods=['POST'])
@login_required_with_json_error
@anysource(tags=opt(str), server_id=str)
def set_tags(tags: Optional[str], server_id: str) -> Response:
    ensure_permission(perms.MANAGE_SERVERS, json=True)
    # TODO: need BU separation
    try:
        replicated.tag_server(server_id=server_id, tags=tags or '')
        return make_json_response(result='ok')
    except NoSuchServer:
        return make_json_response({'error': "NoSuchServer", "details": f"Server '{server_id}' - doesn't exist."}, 400)
