#!/usr/bin/env python3
import os
import re

MIGRATION_LIST = []
NGINX_EPORTAL_CONF = '/etc/nginx/conf.d/eportal.conf'


def migration(title):
    def decorator(fn):
        MIGRATION_LIST.append((title, fn))
        return fn

    return decorator


@migration('Pass Host header in /etc/nginx/conf.d/eportal.conf')
def add_host_into_eportal_conf():
    lines = list(open(NGINX_EPORTAL_CONF))
    result = []

    while True:
        idx = find_line(lines, r'^\s*proxy_pass\s+http://127.0.0.1:8000')
        if idx is None:
            break

        ws = whitespace_prefix(lines[idx])
        result.extend(lines[: idx + 1])
        lines = lines[idx + 1 :]
        stop_mark = r'^\s*(\}|location)'

        remove_line(lines, r'^\s*proxy_set_header Host', stop_mark)
        result.append(f'{ws}proxy_set_header Host $host$port_suffix;\n')

        remove_line(lines, r'^\s*proxy_set_header X-Forwarded-Proto', stop_mark)
        result.append(f'{ws}proxy_set_header X-Forwarded-Proto $orig_scheme;\n')

        remove_line(lines, r'^\s*proxy_set_header X-Forwarded-For', stop_mark)
        result.append(f'{ws}proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;\n')

    result.extend(lines)
    atomic_write(NGINX_EPORTAL_CONF, ''.join(result))


def main():
    import traceback

    print('Applying config migrations:')
    for title, fn in MIGRATION_LIST:
        print('==>', title, '...', end='')
        try:
            fn()
        except:
            print(' Error.')
            traceback.print_exc()
        else:
            print(' Done.')


def atomic_write(fname, content):
    tmp = fname + '.tmp'
    mode = 'wb' if type(content) is bytes else 'w'
    with open(tmp, mode) as f:
        f.write(content)
    os.rename(tmp, fname)


def find_line(lines, regex, stop_mark=None):
    r = re.compile(regex)
    rsm = stop_mark and re.compile(stop_mark)
    for i, l in enumerate(lines):
        if r.search(l):
            return i

        if rsm and rsm.search(l):
            return


def remove_line(lines, regex, stop_mark=None):
    idx = find_line(lines, regex, stop_mark)
    if idx is not None:
        lines.pop(idx)


def whitespace_prefix(line):
    match_ = re.search(r'^\s*', line)
    if not match_:
        raise ValueError("No match for whitespace prefix")
    return match_.group(0)


if __name__ == '__main__':
    main()
