"""Added kc_servers to tag relation table

Revision ID: b90e271fa0fb
Revises: 487a98017b0a
Create Date: 2024-10-23 19:05:22.616118

"""

# revision identifiers, used by Alembic.
revision = 'b90e271fa0fb'
down_revision = '487a98017b0a'
branch_labels = None

import os
from typing import Dict, Iterable, List, Optional, Tuple

import sqlalchemy as sa
from alembic import op
from sqlalchemy.engine import Connection
from sqlalchemy.sql.schema import SchemaItem


def migrate_table(table_name: str, common_columns: Iterable[str], new_schema: Iterable[SchemaItem]) -> None:
    """If you need to change primary key you can't just call 'alter table' because of SQLite limitations.
    In this case the schema is changed by creating a new table and then copying the data."""

    tmp_table_name = f'{table_name}_{os.urandom(8).hex()}'
    op.create_table(tmp_table_name, *new_schema)
    if common_columns:
        column_list = ', '.join(common_columns)
        op.execute(f'INSERT INTO {tmp_table_name} ({column_list}) SELECT {column_list} FROM {table_name}')
    op.drop_table(table_name)
    op.rename_table(tmp_table_name, table_name)


def upgrade_tag_table() -> None:
    unique_columns = ('unit_id', 'name')
    migrate_table(
        'tag',
        common_columns=unique_columns,
        new_schema=(
            sa.Column('tag_id', sa.Integer(), nullable=False, primary_key=True, autoincrement=True),
            sa.Column('unit_id', sa.Integer(), server_default='0', nullable=False),
            sa.Column('name', sa.String(), nullable=False),
        ),
    )
    op.create_index(op.f('ix_tag_unit_id_name'), table_name='tag', columns=unique_columns, unique=True)


def downgrade_tag_table() -> None:
    migrate_table(
        'tag',
        common_columns=('unit_id', 'name'),
        new_schema=(
            sa.Column('unit_id', sa.Integer(), server_default='0', nullable=False, primary_key=True),
            sa.Column('name', sa.String(), nullable=False, primary_key=True),
        ),
    )


def parse_mltag(mltags: str) -> List[Tuple[str, Optional[str]]]:
    def parse_name_value_pair(tag: str) -> Tuple[str, Optional[str]]:
        name, *value = tag.split(':', 1)
        return (name, value[0]) if value else (name, None)

    return [parse_name_value_pair(tag) for tag in mltags.split(';')]


def unit_tag_name_to_id_map(conn: Connection) -> Dict[Tuple[int, str], int]:
    tag_rows = conn.execute(sa.sql.text('SELECT unit_id, name, tag_id FROM tag')).fetchall()
    return {(int(row[0]), row[1]): int(row[2]) for row in tag_rows}


def assign_mltag_via_relation() -> None:
    conn = op.get_bind()
    query = """SELECT s.server_id, s.mltag, k.unit_id FROM kc_servers s
        JOIN kc_keys k ON s.kc_key = k."key" WHERE s.mltag IS NOT NULL"""
    tag_data = conn.execute(sa.sql.text(query))
    if not tag_data.rowcount:
        return
    tags_map = unit_tag_name_to_id_map(conn)
    for server_id, mltag, unit_id in tag_data.fetchall():
        unit_id = int(unit_id)
        params_list = [
            {'server_id': server_id, 'tag_id': tags_map[(unit_id, tag_name)], 'tag_value': tag_value}
            for tag_name, tag_value in parse_mltag(mltag)
            if (unit_id, tag_name) in tags_map
        ]
        statement = sa.sql.text(
            "INSERT INTO kc_servers_tag (server_id, tag_id, tag_value) VALUES (:server_id, :tag_id, :tag_value)"
        )
        for params in params_list:
            conn.execute(statement, params)


def assign_tags_as_mltag() -> None:
    query = "UPDATE kc_servers AS s SET mltag = m.mltag FROM kc_servers_mltag_view AS m WHERE s.server_id = m.server_id"
    op.execute(query)


def upgrade() -> None:
    upgrade_tag_table()
    op.create_table(
        'kc_servers_tag',
        sa.Column(
            'server_id', sa.String(16), sa.ForeignKey('kc_servers.server_id', ondelete="CASCADE"), primary_key=True
        ),
        sa.Column('tag_id', sa.Integer, sa.ForeignKey('tag.tag_id', ondelete="CASCADE"), primary_key=True),
        sa.Column('tag_value', sa.String, nullable=True),
    )
    assign_mltag_via_relation()
    op.drop_column('kc_servers', 'mltag')
    create_view = '''CREATE VIEW kc_servers_mltag_view AS
        SELECT anon_1.server_id, group_concat(anon_1.tag, ';') AS mltag
        FROM (SELECT kc_servers_tag.server_id AS server_id, coalesce(tag.name || ':' || kc_servers_tag.tag_value, tag.name) AS tag
            FROM kc_servers_tag JOIN tag ON tag.tag_id = kc_servers_tag.tag_id ORDER BY tag.name) AS anon_1
        GROUP BY anon_1.server_id'''
    op.execute(create_view)


def downgrade() -> None:
    op.add_column('kc_servers', sa.Column('mltag', sa.String(length=512), nullable=True))
    assign_tags_as_mltag()
    drop_view = 'DROP VIEW kc_servers_mltag_view'
    op.execute(drop_view)
    op.drop_table('kc_servers_tag')
    downgrade_tag_table()
